/*
 *  linux/arch/arm/mach-pxa/fiona.c
 *
 *  Support for the Fiona computer platform
 *
 *  Copyright 2005, 2006 Lab126, Inc.
 *
 *  Ported from gumstix.c
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 */
#include <linux/autoconf.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <asm/hardware.h>
#include <asm/mach-types.h>
#include <asm/mach/arch.h>
#include <asm/arch/udc.h>
#include <asm/arch/pxa-regs.h>
#include <asm/arch/fpow.h>
#include <asm/arch/fiona.h>
#include <asm/arch/audio.h>
#include <asm/irq.h>
#include <asm/mach/irq.h>
#include <asm/arch/mmc.h>
#include <linux/delay.h>
#include <linux/syscalls.h>
#include <linux/suspend.h>
#include <asm/unistd.h>
#include <asm/arch/fiona_eventmap.h>
#include <linux/bootmem.h>
#include <asm/setup.h>
#include <linux/notifier.h>
#include <linux/reboot.h>
#include <linux/zlib.h>
#include <linux/einkfb.h>
#include <linux/vmalloc.h>

#include <asm/arch/boot_globals.h>
#include "generic.h"

#define DO_PS_IGNORE_TEST

#undef FIONA_HOLD_SWITCH_DEBUG
#ifdef FIONA_HOLD_SWITCH_DEBUG
#define hdebug(x...)    printk(x)
#else
#define hdebug(x...)    do {} while (0)
#endif

// Define to handle wan switch locally, undef to pass event up to framework
#undef CONFIG_SYSTEM_SWITCH_CONTROL

// Define to deal with Power switch prior to framework load (i.e. turn off video
// on OFF, turn ON video on ON until framework loads
//
#define HANDLE_POWER_PRIOR_TO_FRAMEWORK

// External prototypes
//
extern void do_peek(int addr, int *value, int size);
extern int eventHandlerActive(void);
extern int calculate_battery_watermarks(void);

#if defined(CONFIG_PM) || defined(CONFIG_FIONA_PM)
extern int fiona_suspend_system(void);
#endif

extern void fiona_set_power_switch_wake_source(int enable);

#ifdef DO_PS_IGNORE_TEST
extern int g_ps_ignore_next_wake;
#endif // DO_PS_IGNORE_TEST

static void fiona_hold_request_bh(unsigned long dummy);
DECLARE_TASKLET_DISABLED(fiona_hold_request_tasklet, fiona_hold_request_bh, 0);


/* Structure to hold our board-rev specific information in */
struct fiona_hw g_fiona_hw;

int g_power_switch_level_int = 1;

#if defined(CONFIG_PM) || defined(CONFIG_FIONA_PM)

#undef FIONA_SLEEP_DEBUG
#ifdef FIONA_SLEEP_DEBUG
#define sdebug(x...)    printk(x)
#else
#define sdebug(x...)
#endif

extern int g_threads_are_stopped;

static int g_did_debounce = 0;

static int  g_last_sleep_request = 0;
static int  g_bounce_count = 0;
static int  g_power_switch_dir = 0;
spinlock_t  g_sleep_lock = SPIN_LOCK_UNLOCKED;

#define POWER_SWITCH_DEBOUNCE_OSCR (20*OSCR_MSECONDS)

static void fiona_sleep_request_bh(unsigned long dummy);
DECLARE_TASKLET_DISABLED(sleep_request_tasklet, fiona_sleep_request_bh, 0);

static void sleep_request_detect_signal(unsigned long dummy);
DECLARE_TASKLET_DISABLED(sleep_tasklet, sleep_request_detect_signal, 0);

#define SLEEP_REQUEST_DETECT_THREAD     "sleepd"
static DECLARE_COMPLETION(sleep_request_detect_thread_exited);

static void fiona_power_switch_bh(unsigned long dummy);
DECLARE_TASKLET_DISABLED(power_switch_tasklet, fiona_power_switch_bh, 0);
#define POWER_SWITCH_DETECT_THREAD     SLEEP_REQUEST_DETECT_THREAD
static DECLARE_COMPLETION(power_switch_detect_thread_exited);

static void __handle_power_switch_event(int switch_status);

static void sleep_switch_debounce_timer_callback(unsigned long dummy);
static struct timer_list sleep_switch_debounce_timer =
{
    function: sleep_switch_debounce_timer_callback
};

static void power_switch_debounce_timer_callback(unsigned long dummy);
static struct timer_list power_switch_debounce_timer =
{
    function: power_switch_debounce_timer_callback
};

typedef struct detect_data {
    pid_t thread_pid;
    wait_queue_head_t event_wq;
    int event_flag;
}detect_data;
struct detect_data sleep_request_detect_data;
struct detect_data power_switch_detect_data;


#endif // CONFIG_PM || CONFIG_FIONA_PM


/*********************************************************
 * Routine:     get_fiona_board_revision
 * Purpose:     Return the Fiona board revision ID
 * Paramters:   none
 * Returns:     FIONA_BOARD_REV2	- REV2 board seen
 *              FIONA_BOARD_DVT1	- DVT1 board seen
 *				FIONA_BOARD_EVT3    - EVT3 board seen
 *              FIONA_BOARD_EVT2    - EVT2 board seen
 *              FIONA_BOARD_EVT1    - EVT1 board seen
 *              FIONA_BOARD_P3_5    - P 3.5 board seen
 *              FIONA_BOARD_P3      - P 3 board seen
 *              FIONA_BOARD_P2      - P 2 board seen
 * Assumptions: system_rev was passed in by bootloader
 *********************************************************/
int get_fiona_board_revision(void)
{
	if (system_rev == FIONA_BOARD_DVT1) {
        unsigned int board_rev;

        do_peek(PHYS_BOARD_REV_ID,&board_rev,4);
        board_rev &= BOARD_REV_ID_MASK;

        //if (BOARD_REV_ID_PVT1 == board_rev)
        //    system_rev = FIONA_BOARD_PVT1;
    }
    return(system_rev);
}

EXPORT_SYMBOL(get_fiona_board_revision);


int fiona_init_board_structure(void)
{
    int version = get_fiona_board_revision();

    // Clear the board structure
    memset(&g_fiona_hw, 0, sizeof(g_fiona_hw));

    // Version check
    switch(version)  {
        case FIONA_BOARD_REV2:
        case FIONA_BOARD_DVT1:
        	// The bootloader only supports DVT1 (REV1) and REV2 boards, and
        	// those watermarks are set up there and stored in the boot globals.
        	// But they are recalcuated in the kernel and stored back to the
        	// boot globals if they change.  This way, on the next reboot, the
        	// bootloader will be using the values as the kernel.
            calculate_battery_watermarks();
        case FIONA_BOARD_EVT3:
        	// This board's watermarks are set below.  But we really no longer
        	// support ANY of the EVTs.
        case FIONA_BOARD_EVT2:
        case FIONA_BOARD_EVT1:
            break;
        case FIONA_BOARD_P2:
        case FIONA_BOARD_P3_5:
        case FIONA_BOARD_P3:
        	// Hmmm...  This is no longer very useful since there's no FIONA_BOARD_EVT1
        	// case below.
            printk("### ERROR: Unsupported Fiona board revision %d seen\n",version);
            printk("### WARNING: Using EVT1 settings for this board (may not work !!)\n");
            version = FIONA_BOARD_EVT1;
            break;
        default:
            printk("### ERROR: Unknown Fiona board revision %d seen\n",get_fiona_board_revision());
            printk("### WARNING: Using EVT3 settings for this board\n");
            version = FIONA_BOARD_EVT3;
            break;
    }

    // Setup board structure based on this board rev
    switch(version)  {
        case FIONA_BOARD_REV2:
            printk("REV2 board detected\n");
            g_fiona_hw.sleep_switch_gpio = GPIO_FIONA_REV2_SLEEP_SWITCH;
            g_fiona_hw.sleep_switch_polarity = FIONA_REV2_SLEEP_SWITCH_SLEEP_POLARITY;
            g_fiona_hw.sleep_switch_md = GPIO_FIONA_REV2_SLEEP_SWITCH_MD;
            g_fiona_hw.sleep_switch_irq = FIONA_REV2_SLEEP_SWITCH_INTR_IRQ;
            g_fiona_hw.ioc_wake_gpio = GPIO_FIONA_REV2_IOC_WAKE;
            g_fiona_hw.ioc_wake_md = GPIO_FIONA_REV2_IOC_WAKE_MD;
            g_fiona_hw.ioc_reset_gpio = GPIO_FIONA_REV2_IOC_RESET;
            g_fiona_hw.ioc_reset_polarity = FIONA_REV2_IOC_RESET_POLARITY;
            g_fiona_hw.pgsr0_reg = CFG_FIONA_REV2_PGSR0;
            g_fiona_hw.pgsr1_reg = CFG_FIONA_REV2_PGSR1;
            g_fiona_hw.pgsr2_reg = CFG_FIONA_REV2_PGSR2;
            g_fiona_hw.wan_switch_gpio = GPIO_FIONA_REV2_WAN_SWITCH;
            g_fiona_hw.wan_switch_polarity = FIONA_REV2_WAN_SWITCH_ON_POLARITY;
            g_fiona_hw.wan_switch_md = GPIO_FIONA_REV2_WAN_SWITCH_MD;
            g_fiona_hw.wan_switch_irq = FIONA_REV2_WAN_SWITCH_INTR_IRQ;
            g_fiona_hw.wake_on_wan_gpio = GPIO_FIONA_REV2_WAKE_ON_WAN;
            g_fiona_hw.wake_on_wan_polarity = FIONA_REV2_WAKE_ON_WAN_ON_POLARITY;
            g_fiona_hw.wake_on_wan_md = GPIO_FIONA_REV2_WAKE_ON_WAN_MD;
            g_fiona_hw.wake_on_wan_irq = FIONA_REV2_WAKE_ON_WAN_INTR_IRQ;
            g_fiona_hw.pwer = CFG_FIONA_REV2_PWER;
            g_fiona_hw.prer = CFG_FIONA_REV2_PRER;
            g_fiona_hw.pfer = CFG_FIONA_REV2_PFER;
            g_fiona_hw.usb_wake_gpio = GPIO_FIONA_REV2_USB_WAKE;
            g_fiona_hw.usb_wake_polarity = GPIO_FIONA_REV2_USB_WAKE_POLARITY;
            break;

        case FIONA_BOARD_DVT1:
            printk("DVT1 board detected\n");
            g_fiona_hw.sleep_switch_gpio = GPIO_FIONA_DVT1_SLEEP_SWITCH;
            g_fiona_hw.sleep_switch_polarity = FIONA_DVT1_SLEEP_SWITCH_SLEEP_POLARITY;
            g_fiona_hw.sleep_switch_md = GPIO_FIONA_DVT1_SLEEP_SWITCH_MD;
            g_fiona_hw.sleep_switch_irq = FIONA_DVT1_SLEEP_SWITCH_INTR_IRQ;
            g_fiona_hw.ioc_wake_gpio = GPIO_FIONA_DVT1_IOC_WAKE;
            g_fiona_hw.ioc_wake_md = GPIO_FIONA_DVT1_IOC_WAKE_MD;
            g_fiona_hw.ioc_reset_gpio = GPIO_FIONA_DVT1_IOC_RESET;
            g_fiona_hw.ioc_reset_polarity = FIONA_DVT1_IOC_RESET_POLARITY;
            g_fiona_hw.pgsr0_reg = CFG_FIONA_DVT1_PGSR0;
            g_fiona_hw.pgsr1_reg = CFG_FIONA_DVT1_PGSR1;
            g_fiona_hw.pgsr2_reg = CFG_FIONA_DVT1_PGSR2;
            g_fiona_hw.wan_switch_gpio = GPIO_FIONA_DVT1_WAN_SWITCH;
            g_fiona_hw.wan_switch_polarity = FIONA_DVT1_WAN_SWITCH_ON_POLARITY;
            g_fiona_hw.wan_switch_md = GPIO_FIONA_DVT1_WAN_SWITCH_MD;
            g_fiona_hw.wan_switch_irq = FIONA_DVT1_WAN_SWITCH_INTR_IRQ;
            g_fiona_hw.wake_on_wan_gpio = GPIO_FIONA_DVT1_WAKE_ON_WAN;
            g_fiona_hw.wake_on_wan_polarity = FIONA_DVT1_WAKE_ON_WAN_ON_POLARITY;
            g_fiona_hw.wake_on_wan_md = GPIO_FIONA_DVT1_WAKE_ON_WAN_MD;
            g_fiona_hw.wake_on_wan_irq = FIONA_DVT1_WAKE_ON_WAN_INTR_IRQ;
            g_fiona_hw.pwer = CFG_FIONA_DVT1_PWER;
            g_fiona_hw.prer = CFG_FIONA_DVT1_PRER;
            g_fiona_hw.pfer = CFG_FIONA_DVT1_PFER;
            g_fiona_hw.usb_wake_gpio = GPIO_FIONA_DVT1_USB_WAKE;
            g_fiona_hw.usb_wake_polarity = GPIO_FIONA_DVT1_USB_WAKE_POLARITY;
            break;

        case FIONA_BOARD_EVT3:
            printk("EVT3 board detected\n");
            g_fiona_hw.sleep_switch_gpio = GPIO_FIONA_EVT3_SLEEP_SWITCH;
            g_fiona_hw.sleep_switch_polarity = FIONA_EVT3_SLEEP_SWITCH_SLEEP_POLARITY;
            g_fiona_hw.sleep_switch_md = GPIO_FIONA_EVT3_SLEEP_SWITCH_MD;
            g_fiona_hw.sleep_switch_irq = FIONA_EVT3_SLEEP_SWITCH_INTR_IRQ;
            g_fiona_hw.hold_switch_gpio = GPIO_FIONA_EVT3_HOLD_SWITCH;
            g_fiona_hw.hold_switch_polarity = FIONA_EVT3_HOLD_SWITCH_ASSERT_POLARITY;
            g_fiona_hw.hold_switch_md = GPIO_FIONA_EVT3_HOLD_SWITCH_MD;
            g_fiona_hw.hold_switch_irq = FIONA_EVT3_HOLD_SWITCH_INTR_IRQ;
            g_fiona_hw.ioc_wake_gpio = GPIO_FIONA_EVT3_IOC_WAKE;
            g_fiona_hw.ioc_wake_md = GPIO_FIONA_EVT3_IOC_WAKE_MD;
            g_fiona_hw.ioc_reset_gpio = GPIO_FIONA_EVT3_IOC_RESET;
            g_fiona_hw.ioc_reset_polarity = FIONA_EVT3_IOC_RESET_POLARITY;
            g_fiona_hw.wan_switch_gpio = GPIO_FIONA_EVT3_WAN_SWITCH;
            g_fiona_hw.wan_switch_polarity = FIONA_EVT3_WAN_SWITCH_ON_POLARITY;
            g_fiona_hw.wan_switch_md = GPIO_FIONA_EVT3_WAN_SWITCH_MD;
            g_fiona_hw.wan_switch_irq = FIONA_EVT3_WAN_SWITCH_INTR_IRQ;
            g_fiona_hw.wake_on_wan_gpio = GPIO_FIONA_EVT3_WAKE_ON_WAN;
            g_fiona_hw.wake_on_wan_polarity = FIONA_EVT3_WAKE_ON_WAN_ON_POLARITY;
            g_fiona_hw.wake_on_wan_md = GPIO_FIONA_EVT3_WAKE_ON_WAN_MD;
            g_fiona_hw.wake_on_wan_irq = FIONA_EVT3_WAKE_ON_WAN_INTR_IRQ;
            g_fiona_hw.pgsr0_reg = CFG_FIONA_EVT3_PGSR0;
            g_fiona_hw.pgsr1_reg = CFG_FIONA_EVT3_PGSR1;
            g_fiona_hw.pgsr2_reg = CFG_FIONA_EVT3_PGSR2;
            g_fiona_hw.wan_watermark = FIONA_WAN_SHUTDOWN_WATERMARK_EVT3;
            g_fiona_hw.low_watermark = FIONA_LOW_WATERMARK_EVT3;
            g_fiona_hw.critical_watermark = FIONA_CRITICAL_WATERMARK_EVT3;
#ifdef CONFIG_SAFETY_NET
            g_fiona_hw.wan_safety_net = FIONA_WAN_SHUTDOWN_WATERMARK_EVT3;
            g_fiona_hw.low_safety_net = FIONA_LOW_WATERMARK_EVT3;
            g_fiona_hw.critical_safety_net = FIONA_CRITICAL_WATERMARK_EVT3;
#endif
            g_fiona_hw.pwer = CFG_FIONA_EVT3_PWER;
            g_fiona_hw.prer = CFG_FIONA_EVT3_PRER;
            g_fiona_hw.pfer = CFG_FIONA_EVT3_PFER;
            g_fiona_hw.usb_wake_gpio = GPIO_FIONA_EVT3_USB_WAKE;
            g_fiona_hw.usb_wake_polarity = GPIO_FIONA_EVT3_USB_WAKE_POLARITY;
            break;

        default:
			printk("\n\n#### ERROR !! DEFAULT CASE Board Detected.\n\n");
            break;
    }
    return(version);
}
        

/****************************************************************************************/

#if (defined(CONFIG_PM) || defined(CONFIG_FIONA_PM)) && defined(CONFIG_SYSTEM_SWITCH_CONTROL)
static void
put_fiona_to_sleep(void)
{
#ifdef CONFIG_PLUG_IN_FPOW
    int rc = 0;
    powdebug("put_fiona_to_sleep(): Calling fiona_sleep_fpow_components()\n");
    rc = fiona_sleep_fpow_components();
    powdebug("put_fiona_to_sleep(): Back from fiona_sleep_fpow_components()\n");
#else
    powdebug("put_fiona_to_sleep(): Calling fiona_suspend_system()\n");
    fiona_suspend_system();
    powdebug("put_fiona_to_sleep(): Back from fiona_suspend_system()\n");
#endif
}
#endif


/*******************************************************************************
 *
 *  Power Switch Handling Code
 *
 *  The following routines are used to support the Power Switch on the back of
 *  Kindle EVT3 and newer units.  This switch is handled differently than the prior
 *  power switch on EVT2 and older, which was a slide-switch as opposed to a toggle.
 *
 *******************************************************************************/

/*******************************************************************
 * Routine:     power_switch
 * Purpose:     Determines if the Power Switch is On or Off
 * Paramters:   none
 * Returns:     0 - Power Switch is OFF
 *              1 - Power Switch is ON
 * Assumptions: none
 *******************************************************************/
int power_switch(void)
{
    if (g_fiona_hw.sleep_switch_polarity)  {
        // Off is in the asserted direction
        return((GPLR(g_fiona_hw.sleep_switch_gpio) & (1<<g_fiona_hw.sleep_switch_gpio)) ? 0 : 1);
    }
    else
        return((GPLR(g_fiona_hw.sleep_switch_gpio) & (1<<g_fiona_hw.sleep_switch_gpio)) ? 1 : 0);
}



/*******************************************************************
 * Routine:     power_switch_detect_signal
 * Purpose:     Dispatch off the signal to the Power Switch daemon
 *              that the switch has transitioned.
 * 
 *              If threads have been stopped, we send the switch status
 *              directly to the event dispatch routine instead of
 *              signalling the daemon because the daemon is sleeping.
 * Paramters:   0 - Power Switch transitioned OFF
 *              1 - Power Switch transitioned ON
 * Returns:     none
 * Assumptions: none
 *******************************************************************/
static inline void power_switch_detect_signal(int switch_status)
{
    if (g_threads_are_stopped) {
        powdebug("POWER: Threads are stopped, calling __handle_power_switch_event() directly.\n");
        __handle_power_switch_event(switch_status);
    }
    else {
        power_switch_detect_data.event_flag = 1;
        wake_up(&power_switch_detect_data.event_wq);
    }
}


/*******************************************************************
 * Routine:     power_switch_debounce_timer_callback
 * Purpose:     Called by the debounce timer, this routine is used to 
 *              debounce the power switch.  When the power switch
 *              transitions, it installs a debounce timer.  When that
 *              timer expires, this routine is called.
 * Paramters:   none
 * Returns:     none
 * Assumptions: none
 *******************************************************************/
static void power_switch_debounce_timer_callback(unsigned long dummy)
{
    int level = 0;

    // Check to see if switch is done bouncing...
    level = power_switch();
        
    // If we're at the target, we should have stopped bouncing (it's been long enough)...
    if (level == g_power_switch_level_int) {
        powdebug("Power switch debounced, calling power_switch_tasklet...\n");
        g_did_debounce = 0;             // Reset indicator for next time...
        enable_irq(g_fiona_hw.sleep_switch_gpio);
        power_switch_detect_signal(level);
    }
    else  {
        powdebug("Sleep switch back in OFF position, debouncing off...\n");
        g_did_debounce = 1;

        // Make sure to handle wrap -njv FIXME
        power_switch_debounce_timer.expires = (jiffies + (20 * HZ / 1000));
        add_timer(&power_switch_debounce_timer);
    }
}



/*******************************************************************
 * Routine:     fiona_power_switch_bh
 * Purpose:     Power switch handler's bottom half.  When a Power Switch
 *              interrupt goes off, the interrupt handler schedules
 *              this task to run.  When we get control, we start a switch
 *              debounce timer that will continue the processing for us
 *              when it goes off, assuring we've debounced the switch.
 * Paramters:   none
 * Returns:     none
 * Assumptions: none
 *******************************************************************/
static void
fiona_power_switch_bh(unsigned long dummy)
{
    powdebug("fiona_power_switch_bh()\n");
    del_timer(&power_switch_debounce_timer);

#ifdef CONFIG_SYSTEM_SWITCH_CONTROL
	// If it's a sleep request, let's go process it...
    if ((GPLR(g_fiona_hw.sleep_switch_gpio) & GPIO_bit(g_fiona_hw.sleep_switch_gpio)) == g_fiona_hw.sleep_switch_polarity) {
        // Send sleep command, when we return, we'll be awake...
        sdebug("Processing power switch transition from BH\n");
        power_switch_detect_signal();
    }
#else
    power_switch_debounce_timer.expires = jiffies + (20 * HZ / 1000);
    add_timer(&power_switch_debounce_timer);
#endif
}


/*******************************************************************
 * Routine:     fiona_power_switch_interrupt
 * Purpose:     Power Switch Interrupt Handler
 *              This handler will disable the Power Switch IRQ, read
 *              and store off the status of the Power Switch, and then
 *              schedule the bottom half to be run.
 * Paramters:   irq - not used here
 *              data - not used here
 *              regs - not used here 
 * Returns:     none
 * Assumptions: The upper-level handling routines must re-enable the IRQ
 *******************************************************************/
static irqreturn_t  
fiona_power_switch_interrupt(int irq, void *data, struct pt_regs *regs)
{
    g_power_switch_level_int = power_switch();
    powdebug("Power Switch Transition Seen %d\n",g_power_switch_level_int);
    tasklet_schedule(&power_switch_tasklet);
    disable_irq(g_fiona_hw.sleep_switch_gpio);
    return IRQ_HANDLED;
}




/*******************************************************************
 * Routine:     __handle_power_switch_event
 * Purpose:     If there's a listener to the event (kbd) driver (the 
 *              framework), then we generate the appropriate POWER_SWITCH
 *              event and pass it up.
 *
 *              If there's not a listener yet (framework hasn't loaded),
 *              then we deal with it ourselves for now until the framework
 *              gets loaded up enough to deal with it the standard way.
 *
 *              If we deal with it, we will clear and disable the screen
 *              on a POWER_SWITCH_OFF, and will enable the screen on a
 *              POWER_SWITCH_ON, but will not stop control of the device.
 *              This allows it to coninue the boot process "under-the-hood",
 *              until the framework loads, at which point it will put the
 *              unit to sleep.  By doing this under-the-hood like this,  we
 *              save the user the bootup time and give them a better 
 *              device power-on user experience when they turn the device
 *              back on (it will be ready, as opposed to having to finish
 *              booting first).
 *
 * Paramters:   switch_status
 *                  0 - Power Switch transitioned to OFF
 *                  1 - Power Switch transitioned to ON
 * Returns:     none
 * Assumptions: none
 * Author:      Nick Vaccaro <nvaccaro@lab126.com>
 *******************************************************************/
static void __handle_power_switch_event(int switch_status)
{
#ifdef CONFIG_SYSTEM_SWITCH_CONTROL
		// Eventually, we'll pass this event off to the upper level
		// so that the framework can decide what to do with it.
		if (switch_status == 0)  {
        	powdebug("power_switch_detect_thread() calling put_fiona_to_sleep()\n");
        	put_fiona_to_sleep();
		}
#else
        g_power_switch_dir = switch_status;

#ifdef HANDLE_POWER_PRIOR_TO_FRAMEWORK
        if (eventHandlerActive()) {
            powdebug("__handle_power_switch_event() calling handle_fiona_event(%s)\n",((switch_status == 0)?"POWER_SW_OFF":"POWER_SW_ON"));
            handle_fiona_event((switch_status == 0)?POWER_SW_OFF_EVENTCODE:POWER_SW_ON_EVENTCODE);
        }
        else  {
            //
            // Framework hasn't started up yet, so deal with power switch manually for now...
            //
            // We could put the system to sleep here, but why not just
            // shut off the light and save the user the rest of the boottime
            // by allowing Fiona to finish booting up into the Framework while the
            // screen is shut off.  Once the framework takes over, it will go into
            // the true "OFF" mode and put things to sleep.  The user shouldn't
            // notice the difference.  IF they power back on before framework loads,
            // we simply turn the lights back on so they can finish watching the boot
            // process.
            if (switch_status == 0)  {
                // Turn off screen and PNLCD
                powdebug("__handle_power_switch_event() POWER OFF - calling fpow_clear_disable_video()\n");
                fpow_clear_disable_video();
            }
            else  {
                // Turn on screen and PNLCD
                powdebug("__handle_power_switch_event() POWER ON - calling fpow_enable_video(0)\n");
                fpow_enable_video(0);
            }
        }
#else
        handle_fiona_event((switch_status == 0)?POWER_SW_OFF_EVENTCODE:POWER_SW_ON_EVENTCODE);
#endif

#endif // CONFIG_SYSTEM_SWITCH_CONTROL
}


static int power_switch_detect_thread(void *arg)
{
    int switch_status = 0;

    daemonize(POWER_SWITCH_DETECT_THREAD);
    allow_signal(SIGKILL);

    while (1) {
        wait_event_interruptible(power_switch_detect_data.event_wq, power_switch_detect_data.event_flag);

#if defined(CONFIG_PM) || defined (CONFIG_FIONA_PM)
        // Check for sleep request
        if (current->flags & PF_FREEZE) {
            refrigerator(PF_FREEZE);
            
            // We need to look at whether Power On was the wake source...
            // If so, we need to NOT continue, and instead fall through to process
            // the switch
    #if 0
            // We just woke up, check to see if the power switch transitioned...
            if (g_power_switch_dir == power_switch())
                continue;
    #else
            if ((POWER_ON_WAKE != fiona_get_wake_source()) && 
                (POWER_OFF_WAKE != fiona_get_wake_source()))
                continue;

        #ifdef DO_PS_IGNORE_TEST
            // If test mode, bypass re-checking switch (code too smart to test paths)
            if (g_ps_ignore_next_wake == 1) {
                g_ps_ignore_next_wake = 0;
                printk("Power Switch Handler: Test IGNORE mode, ignoring this transition.\n"); 
                continue;
            }
        #endif
    #endif
        }
#endif

        if (signal_pending(current)) {
            break;
        }

        power_switch_detect_data.event_flag = 0;

        sdebug("power_switch_detect_thread() active.\n");

        // See if it's an ON or OFF request...
		switch_status = power_switch();

        powdebug("Grabbed Power Switch value of %d == %d (interrupt value)\n",switch_status,g_power_switch_level_int);

        __handle_power_switch_event(switch_status);

    }

    complete_and_exit(&power_switch_detect_thread_exited, 0);

    return 0;
}




static void power_switch_detect_thread_init(void)
{
    pid_t pid;

    memset(&power_switch_detect_data, 0, sizeof(power_switch_detect_data));

    init_waitqueue_head(&power_switch_detect_data.event_wq);

    pid = kernel_thread(power_switch_detect_thread, NULL, CLONE_KERNEL);
    if (pid >= 0) {
        power_switch_detect_data.thread_pid = pid;

    } else {
        power_switch_detect_data.thread_pid = 0;
        printk("power switch detect: thread " POWER_SWITCH_DETECT_THREAD " create failed (%d)\n", pid);

    }
}

static int power_switch_detect_init(irqreturn_t (*fiona_detect_int)(int, void *, struct pt_regs *), void *data)
{
	int err = 0;

    init_timer(&power_switch_debounce_timer);
	power_switch_detect_thread_init();

    // Trigger on switch in both directions
    pxa_gpio_mode(g_fiona_hw.sleep_switch_md);
    set_irq_type(g_fiona_hw.sleep_switch_irq, IRQT_BOTHEDGE);
    err = request_irq(g_fiona_hw.sleep_switch_irq, fiona_detect_int, SA_INTERRUPT, "power switch detect", data);
    if (err) {
        printk(KERN_ERR "power_switch_detect_init: can't request switch IRQ %d\n", g_fiona_hw.sleep_switch_irq);
        err = -1;
		goto exit;
    }

    // Set us up as a wake source
    fiona_set_power_switch_wake_source(1);

    // Enable our tasklet
    tasklet_enable(&power_switch_tasklet);

exit:
	return(err);
}


/*******************************************************************
 * Routine:     startTrackingPowerSwitch
 * Purpose:     Handle the first power switch (generate a transition
 *              event)
 * Paramters:   none
 * Returns:     none
 * Assumptions: none
 *******************************************************************/
static void startTrackingPowerSwitch(void) {
    g_power_switch_dir = power_switch();

    // Switch is in the OFF direction, but we didn't see
    // a transition because it happened before we were completely
    // setup.  We'll issue the OFF transition event now...
    // This should happen well before Framework is listening, so it should
    // cause the System "Power Switch" handler to take care of the situation...
    __handle_power_switch_event(g_power_switch_dir);
}


/*******************************************************************************
 *
 *  Pre-EVT3 Power Switch Handling Code
 *
 *  The following routines are used to support the Power Switch on the back of
 *  Kindle EVT2 and older units.  This switch is handled differently than the 
 *  power switch on EVT3 and newer, which is a toggle switch as opposed to the
 *  pre-EVT3 slide-switch.
 *
 *******************************************************************************/
static int sleep_request_detect_thread(void *arg)
{
    int delay_loops = 1000;

    daemonize(SLEEP_REQUEST_DETECT_THREAD);
    allow_signal(SIGKILL);

    while (1) {
        wait_event_interruptible(sleep_request_detect_data.event_wq, sleep_request_detect_data.event_flag);

#if defined(CONFIG_PM) || defined (CONFIG_FIONA_PM)
        // Check for sleep request
        if (current->flags & PF_FREEZE) {
            refrigerator(PF_FREEZE);
        }
#endif

        if (signal_pending(current)) {
            break;
        }

        sleep_request_detect_data.event_flag = 0;


        // Wait max of 1 second for switch to drop back to neutral position
        while (delay_loops-- && ((GPLR(g_fiona_hw.sleep_switch_gpio) & (1<<g_fiona_hw.sleep_switch_gpio)) == g_fiona_hw.sleep_switch_polarity)) {
            mdelay(1);
        }

        // Re-enable the interrupt
        enable_irq(g_fiona_hw.sleep_switch_gpio);

#ifdef CONFIG_SYSTEM_SWITCH_CONTROL
        sdebug("sleep_request_detect_thread() calling put_fiona_to_sleep()\n");
        // execute user-level sleep script
        put_fiona_to_sleep();
#else
        // We only see the "off" direction, so we only send the off
        // direction.  This will cause problems if framework doesn't
        // handle coherency of switch polarity, but this is only
        // for EVT3 and older, so Shahriar wants to punt this issue. -njv
        //
        handle_fiona_event(POWER_SW_OFF_EVENTCODE);
#endif
    }

    complete_and_exit(&sleep_request_detect_thread_exited, 0);

    return 0;
}

static void sleep_request_detect_thread_init(void)
{
    pid_t pid;

    memset(&sleep_request_detect_data, 0, sizeof(sleep_request_detect_data));

    init_waitqueue_head(&sleep_request_detect_data.event_wq);

    pid = kernel_thread(sleep_request_detect_thread, NULL, CLONE_KERNEL);
    if (pid >= 0) {
        sleep_request_detect_data.thread_pid = pid;

    } else {
        sleep_request_detect_data.thread_pid = 0;
        printk("sleep request detect: thread " SLEEP_REQUEST_DETECT_THREAD " create failed (%d)\n", pid);

    }
}

static int sleep_request_detect_init(irqreturn_t (*fiona_detect_int)(int, void *, struct pt_regs *), void *data)
{
	int err = 0;

    init_timer(&sleep_switch_debounce_timer);

	sleep_request_detect_thread_init();

    // Trigger on switch so we can sleep
    pxa_gpio_mode(g_fiona_hw.sleep_switch_md);
    set_irq_type(g_fiona_hw.sleep_switch_irq, \
        ((0 == g_fiona_hw.sleep_switch_polarity) ? IRQT_FALLING : IRQT_RISING));

    err = request_irq(g_fiona_hw.sleep_switch_irq, fiona_detect_int, SA_INTERRUPT, \
            "sleep request detect", data);
    if (err) {
        printk(KERN_ERR "fiona_sleep_init: can't request switch IRQ %d\n", g_fiona_hw.sleep_switch_irq);
		err = -1;
        goto exit;
    }

    // Set us up as a wake source
    g_fiona_hw.pwer |= PWER_GPIO(g_fiona_hw.sleep_switch_gpio);
    if (g_fiona_hw.sleep_switch_polarity)  {
        // Sleep when signal goes high...  This also means wake when signal goes high,
        // since switch returns to neutral position (low).
        g_fiona_hw.prer |= PRER_GPIO(g_fiona_hw.sleep_switch_gpio);
    } else  {
        g_fiona_hw.pfer |= PFER_GPIO(g_fiona_hw.sleep_switch_gpio);
    }

    // Enable our tasklets
    tasklet_enable(&sleep_request_tasklet);
    tasklet_enable(&sleep_tasklet);
exit:
	return(err);
}



static void sleep_switch_debounce_timer_callback(unsigned long dummy)
{
    // Check to see if switch has transitioned back into off position before
    // acting
    if ((GPLR(g_fiona_hw.sleep_switch_gpio) & GPIO_bit(g_fiona_hw.sleep_switch_gpio)) != g_fiona_hw.sleep_switch_polarity) {
        if (g_did_debounce)  {
            powdebug("Sleep switch back in OFF position, calling sleep_tasklet...\n");
            g_did_debounce = 0;
            enable_irq(g_fiona_hw.sleep_switch_gpio);
            tasklet_schedule(&sleep_tasklet);
        }
        else  {
            powdebug("Sleep switch back in OFF position, debouncing off...\n");
            g_did_debounce = 1;
            sleep_switch_debounce_timer.expires = jiffies + (20 * HZ / 1000);
            add_timer(&sleep_switch_debounce_timer);
        }
    }
    else {
        powdebug("Waiting for switch to drop...\n");
        sleep_switch_debounce_timer.expires = jiffies + (20 * HZ / 1000);
        add_timer(&sleep_switch_debounce_timer);
    }
}

static void sleep_request_detect_signal(unsigned long dummy)
{
    sleep_request_detect_data.event_flag = 1;
    wake_up(&sleep_request_detect_data.event_wq);
}


static void
fiona_sleep_request_bh(unsigned long dummy)
{
    powdebug("fiona_sleep_request_bh()\n");
    del_timer(&sleep_switch_debounce_timer);

    if ((GPLR(g_fiona_hw.sleep_switch_gpio) & GPIO_bit(g_fiona_hw.sleep_switch_gpio)) == g_fiona_hw.sleep_switch_polarity) {
        // Send sleep command, when we return, we'll be awake...
        sdebug("Processing sleep request from BH (BOUNCE COUNT = %d)\n",g_bounce_count);
        sleep_switch_debounce_timer.expires = jiffies + (20 * HZ / 1000);
        add_timer(&sleep_switch_debounce_timer);
    }
    else  {
        sdebug("Switch was in ON position in fiona_sleep_request_bh()\n");
    }
}


static irqreturn_t  fiona_sleep_detect_interrupt(int irq, void *data, struct pt_regs *regs)
{
    int cur = OSCR;

    disable_irq(g_fiona_hw.sleep_switch_gpio);
    powdebug("Sleep request seen\n");

    // Debounce...
    if ((cur - g_last_sleep_request) >= POWER_SWITCH_DEBOUNCE_OSCR)  {
        g_bounce_count = 0;
        tasklet_schedule(&sleep_request_tasklet);
    }
    else  {
        g_bounce_count++;
    }
    g_last_sleep_request = cur;
    return IRQ_HANDLED;
}

extern int g_ioc_in_reset;
static irqreturn_t  fiona_sleep_detect_interrupt_pre_evt3(int irq, void *data, struct pt_regs *regs)
{
    /****************************************************************************
     * EVT2 and older boards have a bug where reseting IOC will cause
     * GPIO1 to transition.  Check to make sure we're not currently
     * resetting IOC before responding to this interrupt. -njv HACK REMOVE at EVT3
     ****************************************************************************/
    if (g_ioc_in_reset == 1)
        return IRQ_HANDLED; 

    if ((GPLR(g_fiona_hw.sleep_switch_gpio) & GPIO_bit(g_fiona_hw.sleep_switch_gpio)) == g_fiona_hw.sleep_switch_polarity) {
        powdebug("Sleep request seen (GPLR0 = %d)\n",GPLR(g_fiona_hw.sleep_switch_gpio) & GPIO_bit(g_fiona_hw.sleep_switch_gpio));
        tasklet_schedule(&sleep_request_tasklet);
    }
    else {
        powdebug("Wake request seen (GPLR0 = %d)\n",GPLR(g_fiona_hw.sleep_switch_gpio) & GPIO_bit(g_fiona_hw.sleep_switch_gpio));
    }

    return IRQ_HANDLED;
}



#if 0
// We currently never un-install the sleep detect handler, so we don't need
// this routine and comment it out to avoid compiler warning.  Should we later
// want to uninstall our handler, we'll want this. -njv

static void sleep_request_detect_done(void)
{
    int ret;

    if (sleep_request_detect_data.thread_pid > 0) {
        ret = kill_proc(sleep_request_detect_data.thread_pid, SIGKILL, 1);
        if (ret == 0) {
            wait_for_completion(&sleep_request_detect_thread_exited);
        }
    }
}

static void fiona_sleep_source_exit(struct device *dev, void *data)
{
    free_irq(g_fiona_hw.sleep_switch_irq, data);
}
#endif


/************************************************************************
 *
 * Power Switch Handling Routines for 2-Position On/Off switch
 *
 ************************************************************************/
static int 
fiona_sleep_source_init(irqreturn_t (*fiona_detect_int)(int, void *, struct pt_regs *), void *data)
{
    int err = 0;
	int	board = get_fiona_board_revision();

    // Initialize our sleep request spinlock
    spin_lock_init(&g_sleep_lock);

    // Setup our sleep request detect thread
	// This is going to be different depending on board type:
	//	FIONA_BOARD_DVT1 and newer have 2-way switch for power, no hold
	//  FIONA_BOARD_EVT3 and older have 3-way momentary on/off and hold
    if (board < FIONA_BOARD_DVT1)
		err = sleep_request_detect_init(fiona_detect_int,data);
	else
		err = power_switch_detect_init(fiona_detect_int,data);

    return(err);
}


/****************************************************************************************/
extern int IOC_Hold_Lockout(int lockout);
static void
fiona_lock_keys(void)
{
    // Call IOC driver to lock out all keys
    hdebug("@@@@ fiona_lock_keys @@@@");
    IOC_Hold_Lockout(1);
}


static void
fiona_unlock_keys(void)
{
    // Call IOC driver to unlock all keys
    hdebug("@@@@ fiona_unlock_keys @@@@");
    IOC_Hold_Lockout(0);
}

static void 
check_and_hold_keys_if_needed(void)
{
    if ((GPLR(g_fiona_hw.hold_switch_gpio) & GPIO_bit(g_fiona_hw.hold_switch_gpio)) == g_fiona_hw.hold_switch_polarity) {
        hdebug("Processing HOLD request from BH - locking out all keys\n");
        fiona_lock_keys();
    }
    else  {
        hdebug("Processing HOLD RELEASE request from BH - unlocking key access\n");
        fiona_unlock_keys();
    }
}

static void
fiona_hold_request_bh(unsigned long dummy)
{
    hdebug("fiona_hold_request_bh\n");

    // If user wants HOLD, lock out all keys
    check_and_hold_keys_if_needed();
}



static irqreturn_t
fiona_hold_detect_interrupt(int irq, void *data, struct pt_regs *regs)
{
    if ((GPLR(g_fiona_hw.hold_switch_gpio) & GPIO_bit(g_fiona_hw.hold_switch_gpio)) == g_fiona_hw.hold_switch_polarity)
        hdebug("HOLD request seen (GPLR = %d)\n",GPLR(g_fiona_hw.hold_switch_gpio) & GPIO_bit(g_fiona_hw.hold_switch_gpio));
    else 
        hdebug("HOLD RELEASE request seen (GPLR = %d)\n",GPLR(g_fiona_hw.hold_switch_gpio) & GPIO_bit(g_fiona_hw.hold_switch_gpio));
    tasklet_schedule(&fiona_hold_request_tasklet);
    return IRQ_HANDLED;
}


static int
fiona_hold_switch_init(irqreturn_t (*fiona_detect_int)(int, void *, struct pt_regs *), void *data)
{
    int err;

    // If user wants HOLD to start with, lock out all keys now...
    //check_and_hold_keys_if_needed();

    // Trigger on HOLD switch in both directions
    set_irq_type(g_fiona_hw.hold_switch_irq, IRQT_BOTHEDGE);
    err = request_irq(g_fiona_hw.hold_switch_irq, fiona_detect_int, SA_INTERRUPT, "HOLD switch detect", data);
    if (err) {
        printk(KERN_ERR "fiona_hold_switch_init: can't request switch IRQ %d\n", g_fiona_hw.hold_switch_irq);
        return -1;
    }

    // Enable our tasklet
    tasklet_enable(&fiona_hold_request_tasklet);

    return 0;
}


/****************************************************************************************/


static int fiona_mci_init(struct device *dev, irqreturn_t (*fiona_detect_int)(int, void *, struct pt_regs *), void *data)
{
    int err;

    set_irq_type(FIONA_P3_SD_CD_IRQ, IRQT_BOTHEDGE);

    err = request_irq(FIONA_P3_SD_CD_IRQ, fiona_detect_int, SA_INTERRUPT, "mmc_detect", data);
    if (err) {
        printk(KERN_ERR "fiona_mci_init: MMC/SD: can't request MMC card detect IRQ %d\n", FIONA_P3_SD_CD_IRQ);
        return -1;
    }

    return 0;
}


static void fiona_mci_setpower(struct device *dev, unsigned int vdd)
{
    //#### to be implemented ####
#ifdef CONFIG_PM_DEBUG
    printk("#### fiona_mci_setpower: vdd=%d\n", vdd);
#endif
}


#define fiona_mci_read_ro()			((GPLR(GPIO9_SD_WP) & GPIO_bit(GPIO9_SD_WP)) != 0)

#define RO_SETTLE_COUNT_MAX			10
#define RO_SETTLE_DELAY				10

static int fiona_mci_get_ro(struct device *dev)
{
	int settle_count = 0, cur_ro = 0, prev_ro = fiona_mci_read_ro();

	while (++settle_count < RO_SETTLE_COUNT_MAX) {
		msleep(RO_SETTLE_DELAY);

		cur_ro = fiona_mci_read_ro(); 

		if (prev_ro != cur_ro) {
			settle_count = 0;

			prev_ro = cur_ro;
		}
	}

	return cur_ro;
}


static void fiona_mci_exit(struct device *dev, void *data)
{
    free_irq(FIONA_P3_SD_CD_IRQ, data);
}


static struct pxamci_platform_data fiona_mci_platform_data = {
    .ocr_mask   = MMC_VDD_32_33|MMC_VDD_33_34,
    .init       = fiona_mci_init,
    .get_ro     = fiona_mci_get_ro,
    .setpower   = fiona_mci_setpower,
    .exit       = fiona_mci_exit,
};


static int fiona_udc_is_connected(void)
{
	return 1; /* always connected on Fiona board (hardwired pull up to D+) */
}


static void fiona_udc_command(int connect_command)
{
#if 0
	if (connect_command == PXA2XX_UDC_CMD_CONNECT) {
		pxa_gpio_mode(GPIO_FIONA_USB_GPIOx_CON_MD);
		GPSR(GPIO_FIONA_USB_GPIOx) = GPIO_bit(GPIO_FIONA_USB_GPIOx);
	}
	if (connect_command == PXA2XX_UDC_CMD_DISCONNECT) {
		GPCR(GPIO_FIONA_USB_GPIOx)  = GPIO_bit(GPIO_FIONA_USB_GPIOx);
		pxa_gpio_mode(GPIO_FIONA_USB_GPIOx_DIS_MD);
	}
#endif
}

static struct pxa2xx_udc_mach_info fiona_udc_info __initdata = {
	.udc_is_connected	= fiona_udc_is_connected,
	.udc_command		= fiona_udc_command,
};


#ifdef CONFIG_PM
static void fiona_set_wakeup_sources(void)
{
#ifdef SYSTEM_CONTROLS_WAKE_SOURCES
    // Prepare power switch, hold switch, and IOC wake source
    //
    pxa_gpio_mode(g_fiona_hw.sleep_switch_md);
    pxa_gpio_mode(g_fiona_hw.hold_switch_md);

    // If GPIO wake source is set, setup the GPIO now..
    if (g_fiona_hw.ioc_wake_gpio) {
        pxa_gpio_mode(g_fiona_hw.ioc_wake_md);
    }

#else
    // The responsibility of setting up as a wake source is in the hands
    // of each individual device.  
    pxa_gpio_mode(g_fiona_hw.hold_switch_md);
#endif
}


static int fiona_pm_prepare(void)
{
    // Setup our sleep source
    int version = get_fiona_board_revision();

    switch (version)  {
        case FIONA_BOARD_EVT2:
        case FIONA_BOARD_EVT1:
        case FIONA_BOARD_P3_5:
        case FIONA_BOARD_P3:
        case FIONA_BOARD_P2:
            fiona_sleep_source_init(fiona_sleep_detect_interrupt_pre_evt3, NULL);
            break;
        case FIONA_BOARD_EVT3:
            fiona_sleep_source_init(fiona_sleep_detect_interrupt, NULL);
            break;
        case FIONA_BOARD_REV2:
		case FIONA_BOARD_DVT1:
            fiona_sleep_source_init(fiona_power_switch_interrupt, NULL);
        default:
            break;
    }

    // Setup our wake sources
    fiona_set_wakeup_sources();

    // Setup our default GPIO Sleep states...
    PGSR0 = g_fiona_hw.pgsr0_reg;
    PGSR1 = g_fiona_hw.pgsr1_reg;
    PGSR2 = g_fiona_hw.pgsr2_reg;
    return 0;
}

#endif  // CONFIG_PM


/****************************************************************************************/

static void _fiona_set_onenand_sync(sync_async_t *sync_settings)
{
	unsigned long msc0;
	
	FIONA_ONENAND_ASYNC_PHYS_REG = sync_settings->onenand_sys_config_1;
	SXCNFG = sync_settings->pxa_sxcnfg;
	MDREFR = sync_settings->pxa_mdrefr;
	
	MSC0 = sync_settings->pxa_msc0;
	msc0 = MSC0;
}

void fiona_set_onenand_sync(void)
{
	_fiona_set_onenand_sync(get_sync_settings());
}

void fiona_set_onenand_async(void)
{
	_fiona_set_onenand_sync(get_async_settings());
}

void fiona_reset_onenand_dbs_bit(void)
{
	if (CLEAR_DBS_TRUE == get_clear_dbs()) {
		FIONA_ONENAND_DBS_REG = FIONA_ONENAND_DBS_CLEAR;
	}
}

void fiona_reset_onenand_dbs_bit_phys(void)
{
	if (CLEAR_DBS_TRUE == get_clear_dbs()) {
		FIONA_ONENAND_DBS_PHYS_REG = FIONA_ONENAND_DBS_CLEAR;
	}
}


/****************************************************************************************/

int fiona_eink_ioctl_stub(unsigned int cmd, unsigned long arg)
{
	int result = 0;
	
	if ( FBIO_EINK_OFF_CLEAR_SCREEN == cmd )
	{
		powdebug("EINK: driver not ready on OFF_CLEAR call, set pending...\n");
		fpow_set_eink_disable_pending();
	}
	
	return ( result );
}

EXPORT_SYMBOL(fiona_eink_ioctl_stub);

int fiona_sys_getpid(void)
{
	return ((int)sys_getpid());
}

EXPORT_SYMBOL(fiona_sys_getpid);

#define HEAD_CRC	2
#define EXTRA_FIELD	4
#define ORIG_NAME	8
#define COMMENT		0x10
#define RESERVED	0xe0

#define DEFLATED	8

#define HEADER_0	0x1f	
#define HEADER_1	0x8b
#define BEST_COMP	2
#define UNIX_OS		3

static void *z_inflate_workspace = NULL;
static void *z_deflate_workspace = NULL;

static int init_z_inflate_workspace(void)
{
	if (NULL == z_inflate_workspace) {
		z_inflate_workspace = vmalloc(zlib_inflate_workspacesize());
	}
	
	return (NULL != z_inflate_workspace);
}

static int init_z_deflate_workspace(void)
{
	if (NULL == z_deflate_workspace) {
		z_deflate_workspace = vmalloc(zlib_deflate_workspacesize());
	}
	
	return (NULL != z_deflate_workspace);
}

/* Table of CRCs of all 8-bit messages. */
static unsigned long crc_table[256];

/* Flag: has the table been computed? Initially false. */
static int crc_table_computed = 0;

/* Make the table for a fast CRC. */
static void make_crc_table(void)
{
  unsigned long c;
  int n, k;

  for (n = 0; n < 256; n++) {
    c = (unsigned long) n;
    for (k = 0; k < 8; k++) {
      if (c & 1) {
        c = 0xedb88320L ^ (c >> 1);
      } else {
        c = c >> 1;
      }
    }
    crc_table[n] = c;
  }
  crc_table_computed = 1;
}

/*
 Update a running crc with the bytes buf[0..len-1] and return
 the updated crc. The crc should be initialized to zero. Pre- and
 post-conditioning (one's complement) is performed within this
 function so it shouldn't be done by the caller. Usage example:

   unsigned long crc = 0L;

   while (read_buffer(buffer, length) != EOF) {
     crc = update_crc(crc, buffer, length);
   }
   if (crc != original_crc) error();
*/
static unsigned long update_crc(unsigned long crc,
                unsigned char *buf, int len)
{
  unsigned long c = crc ^ 0xffffffffL;
  int n;

  if (!crc_table_computed)
    make_crc_table();
  for (n = 0; n < len; n++) {
    c = crc_table[(c ^ buf[n]) & 0xff] ^ (c >> 8);
  }
  return c ^ 0xffffffffL;
}

/* Return the CRC of the bytes buf[0..len-1]. */
static unsigned long crc(unsigned char *buf, int len)
{
  return update_crc(0L, buf, len);
}

int fiona_gunzip(unsigned char *dst, int dstlen, unsigned char *src, unsigned long *lenp)
{
	z_stream s;
	int r, i, flags;

	if (!init_z_inflate_workspace()) {
		printk ("Error: gunzip failed to allocate workspace\n");
		return (-1);
	}
	
	/* skip header */
	i = 10;
	flags = src[3];
	if (src[2] != DEFLATED || (flags & RESERVED) != 0) {
		printk ("Error: Bad gzipped data\n");
		return (-1);
	}
	if ((flags & EXTRA_FIELD) != 0)
		i = 12 + src[10] + (src[11] << 8);
	if ((flags & ORIG_NAME) != 0)
		while (src[i++] != 0)
			;
	if ((flags & COMMENT) != 0)
		while (src[i++] != 0)
			;
	if ((flags & HEAD_CRC) != 0)
		i += 2;
	if (i >= *lenp) {
		printk ("Error: gunzip out of data in header\n");
		return (-1);
	}

	/* Initialize ourself. */
	s.workspace = z_inflate_workspace;
	r = zlib_inflateInit2(&s, -MAX_WBITS);
	if (r != Z_OK) {
		printk ("Error: zlib_inflateInit2() returned %d\n", r);
		return (-1);
	}
	s.next_in = src + i;
	s.avail_in = *lenp - i;
	s.next_out = dst;
	s.avail_out = dstlen;
	r = zlib_inflate(&s, Z_FINISH);
	if (r != Z_OK && r != Z_STREAM_END) {
		printk ("Error: zlib_inflate() returned %d\n", r);
		return (-1);
	}
	*lenp = s.next_out - (unsigned char *) dst;
	zlib_inflateEnd(&s);

	return (0);
}

EXPORT_SYMBOL(fiona_gunzip);

int fiona_gzip(unsigned char *dst, int dstlen, unsigned char *src, unsigned long *lenp)
{
	z_stream s;
	int r, i;
	
	if (!init_z_deflate_workspace()) {
		printk ("Error: gzip failed to allocate workspace\n");
		return (-1);
	}

	/* write header 1 (leave hole for deflate's non-gzip header) */
	for (i = 0; i < 8; i++) {
		dst[i] = 0;
	}
	dst[0] = HEADER_0;
	dst[1] = HEADER_1;
	dst[2] = DEFLATED;
	
	/* Initialize ourself. */
	s.workspace = z_deflate_workspace;
	r = zlib_deflateInit(&s, Z_BEST_COMPRESSION);
	if (r != Z_OK) {
		printk ("Error: zlib_deflateInit() returned %d\n", r);
		return (-1);
	}
	s.next_in = src;
	s.avail_in = *lenp;
	s.next_out = dst + i;
	s.avail_out = dstlen - i;
	r = zlib_deflate(&s, Z_FINISH);
	if (r != Z_OK && r != Z_STREAM_END) {
		printk ("Error: zlib_deflate() returned %d\n", r);
		return (-1);
	}

	/* write header 2 (fill in deflate's header with gzip's)*/
	dst[8] = BEST_COMP;
	dst[9] = UNIX_OS;
	
	/* write trailer (replace adler32 with crc32 & length) */
	i = s.next_out - (unsigned char *) dst - 4;
	s.adler = crc(src, *lenp);
	
	dst[i++] = (unsigned char)(s.adler & 0xff);
	dst[i++] = (unsigned char)((s.adler >> 8) & 0xff);
	dst[i++] = (unsigned char)((s.adler >> 16) & 0xff);
	dst[i++] = (unsigned char)((s.adler >> 24) & 0xff);
	dst[i++] = (unsigned char)(s.total_in & 0xff);
	dst[i++] = (unsigned char)((s.total_in >> 8) & 0xff);
	dst[i++] = (unsigned char)((s.total_in >> 16) & 0xff);
	dst[i++] = (unsigned char)((s.total_in >> 24) & 0xff);

	*lenp = i;
	zlib_deflateEnd(&s);
	
	return (0);
}

EXPORT_SYMBOL(fiona_gzip);

/****************************************************************************************/

#define FIONA_RAM_SIZE_MB			64
#define FIONA_RAM_START				0xA0000000

static unsigned long framebuffer_start = 0;

static unsigned long boot_globals_start = 0;
static boot_globals_t *boot_globals = NULL;


static void init_framebuffer(void)
{
	if (framebuffer_start != 0) {
		reserve_bootmem(framebuffer_start, FRAMEBUFFER_SIZE);
		printk("reserved %lu bytes at %08lx for framebuffer\n", FRAMEBUFFER_SIZE, framebuffer_start);
	}

	set_framebuffer_start(framebuffer_start);
}


boot_globals_t *get_boot_globals(void)
{
	return boot_globals; 
}


static void init_boot_globals(void)
{
	if (boot_globals_start != 0) {
		reserve_bootmem(boot_globals_start, BOOT_GLOBALS_SIZE);
		printk("reserved %lu bytes at %08lx for boot globals\n", BOOT_GLOBALS_SIZE, boot_globals_start);

		boot_globals = (boot_globals_t *)phys_to_virt(boot_globals_start);
		set_warm_restart_flag(BOOT_GLOBALS_COLD_FLAG);
	}
}


static int reboot_boot_globals(struct notifier_block *self, unsigned long u, void *v)
{
    // Say we're "warm" and not (still) hardware-reset.
    //
	set_warm_restart_flag(BOOT_GLOBALS_WARM_FLAG);
	RCSR = RCSR_HWR;

	return NOTIFY_DONE;
}


static struct notifier_block reboot_boot_globals_nb = {
	.notifier_call = reboot_boot_globals,
	.priority = FIONA_REBOOT_BOOT_GLOBALS,
};


static void install_reboot_notifier(void)
{
	register_reboot_notifier(&reboot_boot_globals_nb);
}


static void __init fiona_fixup(struct machine_desc *desc, struct tag *tag,
    char **cmdline, struct meminfo *mi)
{
	SET_BANK(0, FIONA_RAM_START, FIONA_RAM_SIZE_MB * 1024 * 1024);
	mi->nr_banks = 1;
    
	boot_globals_start = mi->bank[mi->nr_banks - 1].start + mi->bank[mi->nr_banks - 1].size - BOOT_GLOBALS_SIZE;
	framebuffer_start  = boot_globals_start - FRAMEBUFFER_SIZE;
}


static void __init fiona_init(void)
{
    // Fill out our g_fiona_hw structure for this board rev 
    fiona_init_board_structure();

    // Setup MMC GPIOs.  (do NOT set up GPIO8 -- it should be left
    // as an input as we are using 1-bit non-SPI mode in Fiona)
    pxa_gpio_mode(GPIO6_MMCCLK_MD);

    pxa_set_udc_info(&fiona_udc_info); 

    // Setup BTUART serial port
    pxa_gpio_mode(GPIO42_BTRXD_MD);
    pxa_gpio_mode(GPIO43_BTTXD_MD);

    if (system_rev == 0x0) {
    	// Power up audio codec (pre-P3.5 )
        GPSR(56) = GPIO_bit(56);
    }

    // Setup MMC/SD detection interface
    pxa_set_mci_info(&fiona_mci_platform_data);

#if defined(CONFIG_PM) || defined(CONFIG_FIONA_PM)
    // Setup for power management
    fiona_pm_prepare();
#endif

    // Start tracking the power switch
    startTrackingPowerSwitch();

    // Initialize support for the hold switch
	// Hold switch only available on EVT3 and older units...
	if (system_rev < FIONA_BOARD_DVT1) {
    	fiona_hold_switch_init(fiona_hold_detect_interrupt, NULL);
    }
}


void __init fiona_init_irq(void)
{
	// Note: We do this in the IRQ init prior to init'ing the IRQs so that
	// interrupts are still disabled prior to our changing the memory
	// layout.
	//
	install_reboot_notifier();
	init_boot_globals();
	init_framebuffer();

	pxa_init_irq();
}


MACHINE_START(FIONA, "Fiona Platform")
	MAINTAINER("Lab126, Inc.")
	BOOT_PARAMS(0xa0000100)
	BOOT_MEM(0xa0000000, 0x40000000, io_p2v(0x40000000))
	.timer = &pxa_timer,
	FIXUP(fiona_fixup)
	MAPIO(pxa_map_io)
	INITIRQ(fiona_init_irq)
	INIT_MACHINE(fiona_init)
MACHINE_END

