/******************************************************************
 *
 *  File:   ioc.c
 *
 *  Author: Nick Vaccaro <nvaccaro@lab126.com>
 *
 *  Date:   12/07/05
 *
 * Copyright 2006, Lab126, Inc.  All rights reserved.
 * Copyright 2005, Lab126, Inc.  All rights reserved.
 *
 *  Description:
 *      Core driver for IOC chip
 *
 ******************************************************************/

#include <linux/autoconf.h>

#include <stdarg.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <asm/errno.h>
#include <asm/string.h>
#include <linux/string.h>
#include <asm/uaccess.h>
#include <asm/arch/hardware.h>
#include <asm/arch/pxa-regs.h>
#include <asm/arch/boot_globals.h>
#include <linux/fs.h>
#include <linux/module.h>
#include <linux/miscdevice.h>
#include <linux/proc_fs.h>
#include <linux/device.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/ioc.h>
#include <linux/pn_lcd.h>
#include "ioc_protocol.h"
#include "ioc_hal.h"
#include "scroll_pm.h"
#include <asm/arch/fpow.h>
#include "ioc_pm.h"



//
// IOC Driver Version
//
#define IOC_DRIVER_VERSION "v1.0"

// Define to fail IOC driver initialization if 
//  wrong IOC FW version seen
#undef  ABORT_ON_BAD_VERSION

//
// Define to query IOC live to check battery
// voltage instead of getting it from sample table
//
#undef CHECK_VOLTAGE_ON_USER_QUERY

//
// Define to prime timer from timer task itself rather than
// from the main task thread.  If defined, this will use
// the "last voltage seen" instead of "this voltage" when
// determining the next voltage check time
//
#undef PRIME_TIMER_ON_LAST_VOLTAGE

//
// Define to keep the voltage check frequency at it's minimum
// setting until the voltage averaging table has filled up.
//
#define PRE_FILL_VOLTAGE_TABLE

//
// Define to disable sliding scale
//
#undef DISABLE_SLIDING_VC_DELAY

//
//  Define if the op amp offset is to be added to the IOC raw value to get correct value,
//  undef if the offset is to be subtracted.
//
#undef POLARITY_OF_OP_AMP_OFFSET_IS_POSITIVE


#define cdebug(x...)    if (g_do_core_debug) printk(x)

typedef struct IOCGlobals {
    int                         signature;
    IOC_HAL_Interface_T         *hal;

    // Rx Client Holding Queue
    wait_queue_head_t rx_client_q; 

    struct fasync_struct        *rx_fasync;

    int                         flags;
}IOCGlobals;

// IOCGlobals.flags field masks
#define AC_ADAPTER_PRESENT_BIT  (1 << 0)

////////////////////////////////////////////////////////
// Globals
///////////////////////////////////////////////////////
static struct IOCGlobals    globs;
static int                  g_user_count = 0;
static int                  g_fw_version = 0;
int                         g_ioc_driver_present = 0;
int                         g_ioc_enabled = 0;
Battery_Info_T              g_battery_info;
int                         g_fw_flash_lock = 1;
int                         g_lowest_amv_seen = 4200;
char                        *g_full_buffer = NULL;
int                         g_full_buffer_index = 0;
int                         g_low_battery_warnings_enabled = 1;
int                         g_keys_locked = 0;
int                         g_board_resistance_set = 0;
int                         g_ioc_allow_sleep_flag = 1;
int                         g_rve_correction_value = UNINITIALIZED_RVE_VALUE;
int                         g_gain_correction_factor = UNINITIALIZED_GAIN_VALUE;

#ifdef SIMULATE_POWER_MODE_TRANSITION_FAILURE
int                         g_ioc_fail_next_transition = 0;
#endif

#ifdef CONFIG_FIONA_PM_IOC
// Semaphore to protect hardware access
DECLARE_MUTEX(g_ioc_hw_lock);

// Static global holding fpow device entry returned from fpow_register()
static struct fpow_component *g_ioc_fpow_component_ptr = NULL;

#endif // CONFIG_FIONA_PM_IOC


#if defined(DEBUG_CORE_IO) || defined(DEBUG_IOC_COMMAND_RECEIVE) || defined(DEBUG_FIFOS)
int                         g_do_core_debug = 1;
#else
int                         g_do_core_debug = 0;
#endif

struct class_simple         *g_ioc_class;


static struct proc_dir_entry *proc_ioc_parent = NULL;
static struct proc_dir_entry *proc_ioc = NULL;
static struct proc_dir_entry *proc_ioc_battery = NULL;
static struct proc_dir_entry *proc_ioc_camv = NULL;
static struct proc_dir_entry *proc_ioc_debug = NULL;
static struct proc_dir_entry *proc_ioc_ac = NULL;
static struct proc_dir_entry *proc_ioc_version = NULL;
static struct proc_dir_entry *proc_ioc_lock = NULL;
static struct proc_dir_entry *proc_ioc_flash = NULL;
static struct proc_dir_entry *proc_ioc_reset_vtbl = NULL;

static struct proc_dir_entry *proc_ioc_min_impedance = NULL;
static struct proc_dir_entry *proc_ioc_max_impedance = NULL;

#ifdef CONFIG_MULTI_RATE_AVERAGING_MONITOR
static struct proc_dir_entry *proc_ioc_battery_vc = NULL;
#endif

#ifdef CONFIG_MULTI_RATE_AVERAGING_MONITOR
static struct proc_dir_entry *proc_ioc_vc_freq = NULL;
static struct proc_dir_entry *proc_ioc_vc_avg_cnt = NULL;
#endif

#ifdef ENABLE_ALLOW_SLEEP_PROC
static struct proc_dir_entry *proc_ioc_allow_sleep = NULL;
#endif
#ifdef SIMULATE_POWER_MODE_TRANSITION_FAILURE
static struct proc_dir_entry *proc_ioc_fail_next_transition = NULL;
#endif
#ifdef ENABLE_PROC_IOC_RESET
static struct proc_dir_entry *proc_ioc_reset = NULL;
#endif


#ifdef CONFIG_IOC_VOLTAGE_CHECKER
static struct timer_list vc_timer;
static void start_ioc_voltage_check_thread(void);
static void stop_ioc_voltage_check_thread(void);
#endif

//////////////////////////////
// External Structs and Prototypes
//////////////////////////////
extern struct Fiona_IOC_HAL_Globals *g_hal_globs;
extern int g_keys_locked;
extern int verify_firmware_protocol_match(void);
extern unsigned long calculate_board_resistance(void);
extern int calculate_battery_watermarks(void);

extern int g_min_board_resistance;
extern int g_max_board_resistance;

#ifdef CONFIG_IOC_VOLTAGE_CHECKER
static void IOC_Voltage_Sampler_Timer(unsigned long parm);
#endif

//////////////////////////////
// Local Prototypes
//////////////////////////////

// Core Driver Probe and Remove Functions
static int IOCDriver_Remove(struct device *dev);

// Top-level Send command and receive response routine
static int Do_IOC_IssueCmd(char *cmd, int cmdSize, char *response, int responseSize, int resetChipOnErr);
#define IOC_IssueCmd(w,x,y,z)   Do_IOC_IssueCmd(w,x,y,z,1)


// /proc interface
static int ioc_proc_init(void);
static void ioc_cleanup_procfs(void);

// Utility Routine Prototypes
//
static void print_battery_info_string(Battery_Info_T *info);
unsigned int ioc_convert_raw_voltage_to_mvolts(unsigned int raw);
signed int ioc_convert_raw_temp_to_degrees_celcius(unsigned int raw);
unsigned int ioc_convert_raw_draw_to_mamps(unsigned int raw);
int IOC_Hold_Lockout(int lockout);
static int IOC_Get_Hold_Lockout(void);

static inline void 
IOC_Enable_Err(char *target, unsigned short enable, int rc, int size)
{
    printk("ERROR: Invalid response %sabling %s (only %d bytes returned (should be %d bytes)\n",(enable == ENABLE)?"en":"dis",target,rc,size);
}

////////////////////////////////////////////////////////////
// Protocol-specific Source Files
// These are included inline like this so that we
// don't have to export globs global
////////////////////////////////////////////////////////////
#include "ioc_cmds.c"

//////////////////////////////
// Core Driver Routines
//////////////////////////////
static int
IOC_Open(struct inode* inNode, struct file* inFile)
{
    PDEBUG("IOC_Open()\n");
    // Inc user count
    if (!g_user_count++)  {

#ifdef ENABLE_IOC_ON_OPEN
        // First, get the lock
        lockdebug("Getting g_ioc_hw_lock in IOC_Open()...\n");
        if (down_interruptible(&g_ioc_hw_lock))
            return(-ERESTARTSYS);
        lockdebug("Got g_ioc_hw_lock in IOC_Open()...\n");

        // Enable the IOC
        if (Do_Enable_IOC(ENABLE)) {
            up(&g_ioc_hw_lock);
            lockdebug("Released g_ioc_hw_lock in IOC_Open...\n");
            return(-EREMOTEIO);  // Error - couldn't open
        }
        up(&g_ioc_hw_lock);
        lockdebug("Released g_ioc_hw_lock in IOC_Open...\n");
#endif
    }

    // Return success
    return 0;
}

static int
IOC_Close(struct inode* inNode, struct file* inFile)
{
    PDEBUG("IOC_Close()\n");
    if (--g_user_count)  {
#ifdef ENABLE_IOC_ON_OPEN

        // First, get the lock
        lockdebug("Getting g_ioc_hw_lock in IOC_Close()...\n");
        if (down_interruptible(&g_ioc_hw_lock))
            return(-ERESTARTSYS);
        lockdebug("Got g_ioc_hw_lock in IOC_Close()...\n");

        // Disable the IOC
        Do_Enable_IOC(DISABLE);

        up(&g_ioc_hw_lock);
        lockdebug("Released g_ioc_hw_lock in IOC_Open...\n");
#endif
    }

    // Return success
    return 0;
}


static ssize_t
IOC_Read(struct file * file, char * buffer, size_t count, loff_t *ppos)
{
    return(count);
}

static ssize_t
IOC_Write(struct file *file, const char __user *buffer, size_t count, loff_t *ppos)
{
    return(count);
}

static int
IOC_Ioctl(struct inode *inNode, struct file *inFile,
            unsigned int inCommand, unsigned long inArg)
{
    IOC_ENABLE_REC              enableRec;
    int                         retVal = 0;
    char                        response[MAX_IOC_RX_PACKET_SIZE+1];
    int                         x = 0;

    PDEBUG("IOC_Ioctl(): inCommand = %x\n", inCommand);

    if (_IOC_TYPE(inCommand) != IOCTL_IOC_MAGIC_NUMBER) return -EINVAL;
    if (_IOC_NR(inCommand) > IOC_IOCTLLast) return -EINVAL;

    // If the IOC is asleep, error out...
    if (0 == ioc_awake())
        return(-FPOW_COMPONENT_NOT_POWERED_ERR);

    // First, get the lock
    lockdebug("Getting g_ioc_hw_lock in IOC_Ioctl()...\n");
    if (down_interruptible(&g_ioc_hw_lock))
        return(-ERESTARTSYS);
    lockdebug("Got g_ioc_hw_lock in IOC_Ioctl()...\n");

    // Get user data
    switch(inCommand)  {
        case IOC_PNLCD_VERT_SEQ_IOCTL:
        case IOC_PNLCD_SEQ_IOCTL:
        {   
            PNLCD_SEQ_SEGMENTS_REC      pnlcdSeqRec;

            NDEBUG("IOC_PNLCD%sSEQ_IOCTL case seen. \n",(inCommand == IOC_PNLCD_SEQ_IOCTL)?"_":"_VERT_");

            // If we were called from inside, don't copy from user space
            if (inNode == (struct inode *)0xDEADBEEF)
                memcpy(&pnlcdSeqRec,(void*)inArg,sizeof(pnlcdSeqRec));
            else 
            if (copy_from_user(&pnlcdSeqRec,(void *)inArg, sizeof(pnlcdSeqRec)))  {
                NDEBUG("IOC_IOCTL() failed copy_from_user !! Kernel Space ??? \n");
                goto errOut;
            }

            // Turn on or off the requested sequential segment range
            NDEBUG("IOC_IOCTL() calling IOC_Seq_Segment() \n");
            retVal = IOC_Sequential_Segment(pnlcdSeqRec.enable, pnlcdSeqRec.start_seg, pnlcdSeqRec.end_seg,(inCommand == IOC_PNLCD_SEQ_IOCTL)?0:1);
            NDEBUG("IOC_IOCTL() back from calling IOC_Seq_Segment(), retVal = %d \n",retVal);
            break;
        }

        case IOC_PNLCD_MISC_IOCTL:
        {
            PNLCD_MISC_SEGMENTS_REC     pnlcdMiscRec;
            char                        segments[PN_LCD_SEGMENT_COUNT];
            int                         user_segments[PN_LCD_SEGMENT_COUNT];
            
            NDEBUG("IOC_PNLCD_MISC_IOCTL case seen. \n");

            // If we were called from inside, don't copy from user space
            if (inNode == (struct inode *) 0xDEADBEEF) {
                memcpy(&pnlcdMiscRec,(void*)inArg,sizeof(pnlcdMiscRec));
                memcpy(user_segments,pnlcdMiscRec.seg_array,MIN((pnlcdMiscRec.num_segs*sizeof(*pnlcdMiscRec.seg_array)),sizeof(user_segments)));
            }
            else  {
                // Copy main parameter block
                if (copy_from_user(&pnlcdMiscRec,(void *)inArg, sizeof(pnlcdMiscRec))) {
                    NDEBUG("IOC_IOCTL() failed to copy user data\n");
                    goto errOut;
                }

                // Copy segment array pointer
                if (copy_from_user(user_segments,(void *)pnlcdMiscRec.seg_array, MIN((pnlcdMiscRec.num_segs*sizeof(*pnlcdMiscRec.seg_array)),sizeof(user_segments)))) {
                    NDEBUG("IOC_IOCTL() failed to copy user segment_array data\n");
                    goto errOut;
                }
            }

            // Now load up our character array pointer
            for (x=0; x<pnlcdMiscRec.num_segs; x++)
                segments[x] = (char) user_segments[x];

            // Turn on or off the requested segments
            retVal = IOC_Multi_Segment(pnlcdMiscRec.enable, pnlcdMiscRec.num_segs, segments);
            break;
        }

        case IOC_PNLCD_VERT_PAIRS_IOCTL:
        case IOC_PNLCD_PAIRS_IOCTL:
        {
            PNLCD_SEGMENT_PAIRS_REC     pnlcdPairRec;
            char                        segments[PN_LCD_SEGMENT_COUNT];
            int                         user_segments[PN_LCD_SEGMENT_COUNT];

            NDEBUG("IOC_PNLCD%sPAIRS_IOCTL case seen. \n",(inCommand == IOC_PNLCD_PAIRS_IOCTL)?"_":"_VERT_");

            // If we were called from inside, don't copy from user space
            if (inNode == (struct inode *) 0xDEADBEEF) {
                memcpy(&pnlcdPairRec,(void*)inArg,sizeof(pnlcdPairRec));
                memcpy(user_segments,pnlcdPairRec.seg_pair_array,MIN((pnlcdPairRec.num_pairs*2*sizeof(*pnlcdPairRec.seg_pair_array)),sizeof(user_segments)));
            }
            else  {
                // Copy main parameter block
                if (copy_from_user(&pnlcdPairRec,(void *)inArg, sizeof(pnlcdPairRec)))
                    goto errOut;

                // Copy segment array pointer
                if (copy_from_user(user_segments,(void *)pnlcdPairRec.seg_pair_array, MIN((pnlcdPairRec.num_pairs*2*sizeof(*pnlcdPairRec.seg_pair_array)),sizeof(user_segments))))
                    goto errOut;
            }

            // Now load up our character array pointer
            for (x=0; x<(pnlcdPairRec.num_pairs*2); x++)
                segments[x] = (char) user_segments[x];

            // Turn on or off the requested segments
            retVal = IOC_Multi_Segment_Pairs(pnlcdPairRec.enable, pnlcdPairRec.num_pairs, segments,(inCommand == IOC_PNLCD_PAIRS_IOCTL)?0:1);
            break;
        }

        case IOC_QUERY_HOLD_LOCKOUT_IOCTL:
            PDEBUG("IOC_QUERY_HOLD_LOCKOUT_IOCTL case seen. \n");
            enableRec.enable = IOC_Get_Hold_Lockout();
            // If we were called from inside, don't copy from user space
            if (inNode == (void*)0xDEADBEEF)
                memcpy((void*)inArg, &enableRec, sizeof(enableRec));
            else  {
                if (copy_to_user((void*) inArg, &enableRec, sizeof(enableRec)))
                    goto errOut;
            }
            break;

        case IOC_HOLD_LOCKOUT_IOCTL:
            PDEBUG("IOC_HOLD_LOCKOUT_IOCTL case seen. \n");

            // Copy main parameter block
            // If we were called from inside, don't copy from user space
            if (inNode == (void*)0xDEADBEEF)
                memcpy(&enableRec,(void*)inArg,sizeof(enableRec));
            else  {
                if (copy_from_user(&enableRec,(void *)inArg, sizeof(enableRec)))
                    goto errOut;
            }

            // Enable or disable ALL keys, buttons, and scrollwheel
            retVal = IOC_Hold_Lockout(enableRec.enable == ENABLE ? 1 : 0);
            break;

        case IOC_KEYBOARD_ENABLE_IOCTL:
            PDEBUG("IOC_KEYBOARD_ENABLE_IOCTL case seen. \n");

            // Copy main parameter block
            // If we were called from inside, don't copy from user space
            if (inNode == (void*)0xDEADBEEF)
                memcpy(&enableRec,(void*)inArg,sizeof(enableRec));
            else  {
                if (copy_from_user(&enableRec,(void *)inArg, sizeof(enableRec)))
                    goto errOut;
            }

            // Enable or disable the keyboard
            retVal = IOC_Enable_Keyboard(enableRec.enable);
            break;

        case IOC_PNLCD_ENABLE_IOCTL:
            PDEBUG("IOC_PNLCD_ENABLE_IOCTL case seen. \n");

            // Copy main parameter block
            // If we were called from inside, don't copy from user space
            if (inNode == (struct inode *)0xDEADBEEF)
                memcpy(&enableRec,(void*)inArg,sizeof(enableRec));
            else  {
                if (copy_from_user(&enableRec,(void *)inArg, sizeof(enableRec)))
                    goto errOut;
            }

            // Enable or disable the pn-lcd 
            retVal = IOC_Enable_PNLCD(enableRec.enable);
            break;

        case IOC_ENABLE_IOC_IOCTL:
            PDEBUG("IOC_ENABLE_IOC_IOCTL case seen. \n");

            // Copy main parameter block
            // If we were called from inside, don't copy from user space
            if (inNode == (struct inode *)0xDEADBEEF)
                memcpy(&enableRec,(void*)inArg,sizeof(enableRec));
            else  {
                if (copy_from_user(&enableRec,(void *)inArg, sizeof(enableRec)))
                    goto errOut;
            }

            // Enable or disable the IOC
            retVal = Do_Enable_IOC(enableRec.enable);
            break;

        case IOC_BATT_LV_WARN_ENABLE_IOCTL:
            {
                IOC_ENABLE_LOW_BATT_WARNINGS_REC    lvwRec;

                PDEBUG("IOC_BATT_LV_WARN_ENABLE_IOCTL case seen. \n");
                BDEBUG("IOC_BATT_LV_WARN_ENABLE_IOCTL case seen. \n");

                // Copy main parameter block
                // If we were called from inside, don't copy from user space
                if (inNode == (struct inode *)0xDEADBEEF)
                    memcpy(&lvwRec,(void*)inArg,sizeof(lvwRec));
                else  {
                    if (copy_from_user(&lvwRec,(void *)inArg, sizeof(lvwRec)))
                        goto errOut;
                }
            
                // Enable or disable the reporting of Battery Low Voltage conditions
                retVal = IOC_Enable_Battery_Voltage_Low_Reporting(lvwRec.enable, lvwRec.level);
            }
            break;

        case IOC_BATT_CAPACITY_ENABLE_IOCTL:
            PDEBUG("IOC_BATT_CAPACITY_ENABLE_IOCTL case seen. \n");
            BDEBUG("IOC_BATT_CAPACITY_ENABLE_IOCTL case seen. \n");

            // Copy main parameter block
            // If we were called from inside, don't copy from user space
            if (inNode == (struct inode *)0xDEADBEEF)
                memcpy(&enableRec,(void*)inArg,sizeof(enableRec));
            else  {
                if (copy_from_user(&enableRec,(void *)inArg, sizeof(enableRec)))
                    goto errOut;
            }
            
            // Enable or disable the reporting of battery capacity
            retVal = IOC_Enable_Battery_Capacity_Reporting(enableRec.enable);
            break;

        case IOC_QUERY_BATT_INSTALLED_IOCTL:
        {
            IOC_BATTERY_INSTALLED_REC rec;

            PDEBUG("IOC_QUERY_BATT_PRESENT_IOCTL case seen. \n");
            BDEBUG("IOC_QUERY_BATT_PRESENT_IOCTL case seen. \n");
            rec.installed = battery_installed();
            copy_to_user((void*) inArg, &rec, sizeof(rec));
            break;
        }

        case IOC_QUERY_BATT_VOLTAGE_IOCTL:
        {
            IOC_BATTERY_VOLTAGE_REC rec;

            PDEBUG("IOC_QUERY_BATT_VOLTAGE_IOCTL case seen. \n");
            BDEBUG("IOC_QUERY_BATT_VOLTAGE_IOCTL case seen. \n");
            rec.mvolts = 0;
            if (!battery_installed()) {
                copy_to_user((void*) inArg, &rec, sizeof(rec));
                break;
            }
            retVal = IOC_Query_Cmd(IOC_QUERY_BATTERY_LEVEL, response, 0);
            if (retVal == 0)  {
                rec.mvolts = ioc_convert_raw_voltage_to_mvolts((unsigned int)response[1]);
                copy_to_user((void*) inArg, &rec, sizeof(rec));
            }
            else {
                copy_to_user((void*) inArg, &rec, sizeof(rec));
                goto errOut;
            }
            break;
        }

        case IOC_QUERY_BATT_CAPACITY_IOCTL:
        {
            IOC_BATTERY_CAPACITY_REC rec;

            PDEBUG("IOC_QUERY_BATT_CAPACITY_IOCTL case seen. \n");
            BDEBUG("IOC_QUERY_BATT_CAPACITY_IOCTL case seen. \n");
            if (!battery_installed()) {
                rec.capacity = 0;
                copy_to_user((void*) inArg,&rec,sizeof(rec));
                break;
            }
            retVal = IOC_Query_Cmd(IOC_QUERY_BATTERY_CAPACITY, response, 0);
            if (retVal == 0)  {
                rec.capacity = response[1];
                copy_to_user((void*) inArg,&rec,sizeof(rec));
            }
            else {
                rec.capacity = 0;
                copy_to_user((void*) inArg,&rec,sizeof(rec));
                goto errOut;
            }
            break;
        }

        case IOC_QUERY_BATT_TEMP_IOCTL:
        {
            IOC_BATTERY_TEMPERATURE_REC rec;

            PDEBUG("IOC_QUERY_BATT_TEMP_IOCTL case seen. \n");
            BDEBUG("IOC_QUERY_BATT_TEMP_IOCTL case seen. \n");
            rec.temperature = 0;
            if (!battery_installed()) {
                copy_to_user((void*) inArg, &rec, sizeof(rec));
                break;
            }
            retVal = IOC_Query_Cmd(IOC_QUERY_BATTERY_TEMPERATURE, response, 0);
            if (retVal == 0)  {
                rec.temperature = ioc_convert_raw_temp_to_degrees_celcius((unsigned int)response[1]);
                copy_to_user((void*) inArg,&rec,sizeof(rec));
            }
            else {
                copy_to_user((void*) inArg, &rec, sizeof(rec));
                goto errOut;
            }
            break;
        }

        case IOC_QUERY_BATT_CHARGING_IOCTL:
        {
            IOC_BATTERY_CHARGE_STATUS_REC rec;
            PDEBUG("IOC_QUERY_BATT_CHARGING_IOCTL case seen. \n");
            BDEBUG("IOC_QUERY_BATT_CHARGING_IOCTL case seen. \n");
            retVal = IOC_Query_Cmd(IOC_QUERY_CURRENT_DRAW, response, 0);
            if (retVal == 0)  {
                // If it's a charge response and not a discharge response,
                // then we're actually drawing "negative" current as we're charging.
                if (response[0] == IOC_BATTERY_CHARGE_RESPONSE) {
                    rec.charging = 1;
                }
                else
                    rec.charging = 0;
                copy_to_user((void*) inArg,&rec,sizeof(rec));
            }
            else {
                rec.charging = 0;
                copy_to_user((void*) inArg,&rec,sizeof(rec));
                goto errOut;
            }
            break;
        }

        case IOC_QUERY_BATT_CURRENT_DRAW_IOCTL:
        {
            IOC_BATTERY_CURRENT_DRAW_REC rec;

            PDEBUG("IOC_QUERY_BATT_CURRENT_DRAW_IOCTL case seen. \n");
            BDEBUG("IOC_QUERY_BATT_CURRENT_DRAW_IOCTL case seen. \n");
            retVal = IOC_Query_Cmd(IOC_QUERY_CURRENT_DRAW, response, 0);
            if (retVal == 0)  {
                rec.draw = ioc_convert_raw_draw_to_mamps((unsigned int) response[1]);

                // If it's a charge response and not a discharge response,
                // then we're actually drawing "negative" current as we're charging.
                if (response[0] == IOC_BATTERY_CHARGE_RESPONSE) {
                    rec.draw *= -1;
                }

                copy_to_user((void*) inArg,&rec,sizeof(rec));
            }
            else {
                rec.draw = 0;
                copy_to_user((void*) inArg,&rec,sizeof(rec));
                goto errOut;
            }
            break;
        }

        case IOC_QUERY_BATT_INFO_IOCTL:
            PDEBUG("IOC_QUERY_BATT_INFO_IOCTL case seen. \n");
            BDEBUG("IOC_QUERY_BATT_INFO_IOCTL case seen. \n");
            IOC_Query_Battery_Info(&g_battery_info);
            copy_to_user((void*) inArg,&g_battery_info,sizeof(g_battery_info));
            break;

        case IOC_QUERY_FW_VERSION_IOCTL:
        {
            IOC_FW_VERSION_REC    rec;

            PDEBUG("IOC_QUERY_FW_VERSION_IOCTL case seen. \n");
            retVal = get_ioc_firmware_version(&rec.version,1);
            if (retVal)  {
                rec.version = g_fw_version;
                copy_to_user((void*) inArg,&rec,sizeof(rec));
                goto errOut;
            }
            copy_to_user((void*) inArg,&rec,sizeof(rec));
            break;
        }

        case IOC_RESET_IOC_IOCTL:
            PDEBUG("IOC_RESET_IOC_IOCTL case seen. \n");

            // Reset the IOC
            retVal = globs.hal->IOC_HAL_ResetChip(globs.hal);
            break;

        case IOC_RESET_FIONA_IOCTL:
            PDEBUG("IOC_RESET_FIONA_IOCTL case seen. \n");

            // Reset the PXA
            IOC_Reset_Fiona();
            break;

        case IOC_SET_TWO_KEY_WAKE_IOCTL:
        {
            IOC_TWO_KEY_WAKE_REC    rec;

            PDEBUG("IOC_SET_TWO_KEY_WAKE_IOCTL case seen. \n");

            // Copy main parameter block
            // If we were called from inside, don't copy from user space
            if (inNode == (void*)0xDEADBEEF)
                memcpy(&rec,(void*)inArg,sizeof(rec));
            else  {
                if (copy_from_user(&rec,(void *)inArg, sizeof(rec)))
                    goto errOut;
            }

            // Set the two keys that are to be pressed to wake up IOC
            // This no longer exists, but we're leaving this hook in
            // because a) it's late in the game, changes propogate upstream
            // too far, and b) we may need it again for follow-on projects.
            // This was depricated with version 56 of IOC firmware.
            //retVal = IOC_Set_Two_Key_Wake(rec.wakeKey1, rec.wakeKey2);
            break;
        }
        case IOC_TWO_KEY_WAKE_ENABLE_IOCTL:
            PDEBUG("IOC_TWO_KEY_WAKE_ENABLE_IOCTL case seen. \n");

            // Copy main parameter block
            // If we were called from inside, don't copy from user space
            if (inNode == (void*)0xDEADBEEF)
                memcpy(&enableRec,(void*)inArg,sizeof(enableRec));
            else  {
                if (copy_from_user(&enableRec,(void *)inArg, sizeof(enableRec)))
                    goto errOut;
            }

            // Enable or disable ALL keys, buttons, and scrollwheel
            retVal = IOC_Two_Key_Wake_Enable(enableRec.enable);
            break;

        case IOC_ALL_KEY_WAKE_ENABLE_IOCTL:
            PDEBUG("IOC_ALL_KEY_WAKE_ENABLE_IOCTL case seen. \n");

            // Copy main parameter block
            // If we were called from inside, don't copy from user space
            if (inNode == (void*)0xDEADBEEF)
                memcpy(&enableRec,(void*)inArg,sizeof(enableRec));
            else  {
                if (copy_from_user(&enableRec,(void *)inArg, sizeof(enableRec)))
                    goto errOut;
            }

            // Enable or disable ALL keys, buttons, and scrollwheel
            retVal = IOC_All_Key_Wake_Enable(enableRec.enable);
            break;

        case IOC_QUERY_IOC_WAKE_CAUSE_IOCTL:
        {
            IOC_WAKE_CAUSE_REC  ioc_wake_rec;
            memclr(&ioc_wake_rec,sizeof(ioc_wake_rec));

            IOC_Get_Wake_Source(&ioc_wake_rec.wake_source);
            printk("IOC_QUERY_WAKE_SOURCE_IOCTL returning %s (%d) wake source\n",fpow_wake_src_to_str(ioc_wake_rec.wake_source), (int)ioc_wake_rec.wake_source);

            // If we were called from inside, don't copy from user space
            if (inNode == (void*)0xDEADBEEF)
                memcpy((void*)inArg, &ioc_wake_rec, sizeof(ioc_wake_rec));
            else  {
                if (copy_to_user((void*) inArg, &ioc_wake_rec, sizeof(ioc_wake_rec)))
                    goto errOut;
            }
            break;
        }

        default:
            printk("Unknown IOCTL case %x seen. \n",inCommand);
            goto errOut;
    }
    up(&g_ioc_hw_lock);
    lockdebug("Released g_ioc_hw_lock leaving IOC_Ioctl...\n");
    return(retVal);

errOut:
    up(&g_ioc_hw_lock);
    lockdebug("Released g_ioc_hw_lock leaving IOC_Ioctl...\n");
    return(-EFAULT);
}


static void
IOCDriver_Platform_Release(
    struct device *device)
{
    // this is called when the reference count goes to zero
}



int fasync_ioc(int fd, struct file *filp, int on)
{
    int retval;

    retval = fasync_helper(fd, filp, on, &g_hal_globs->rx_q.fasync);
    if( retval < 0)
        return retval;
    return 0;
}



//////////////////////////////
// Exported File Operators
//  Init and Destroy Routines
//////////////////////////////

static struct device_driver ioc_driver = {
    .name = "ioc",
    .bus = &platform_bus_type,
    .remove = IOCDriver_Remove,
};

static struct platform_device ioc_device = {
    .name = "ioc",
    .id = 0,
    .dev = {
        .release = IOCDriver_Platform_Release,
    }
};

struct file_operations ioc_fops = {
        owner:          THIS_MODULE,
        read:           IOC_Read,
        write:          IOC_Write,
        ioctl:          IOC_Ioctl,
        open:           IOC_Open,
        release:        IOC_Close,
        fasync:         fasync_ioc
};

static struct miscdevice g_ioc_misc_device = {
    minor:      IOC_MINOR,
    name:       IOC_DRIVER_NAME,
    fops:       &ioc_fops,
};


static int
__init IOCDriver_Init(void)
{
    int     retVal = 0;
    int     version = 0;
    int     offset = 0;

    // Initialize our globals...
    if (globs.signature != IOC_GLOBALS_SIGNATURE)  {
        memclr(&globs,sizeof(globs));
        globs.signature = IOC_GLOBALS_SIGNATURE;

        globs.rx_client_q.lock = SPIN_LOCK_UNLOCKED; 
        globs.rx_client_q.task_list.next = &globs.rx_client_q.task_list;
        globs.rx_client_q.task_list.prev = &globs.rx_client_q.task_list;
        memclr(&g_battery_info,sizeof(g_battery_info));
    }

    // First, get the lock
    lockdebug("Getting g_ioc_hw_lock in IOCDriver_Resume()...\n");
    if (down_interruptible(&g_ioc_hw_lock))
        return(-ERESTARTSYS);
    lockdebug("Got g_ioc_hw_lock in IOCDriver_Resume()...\n");

    // Initialize the HAL....
    globs.hal = IOC_HAL_Init();
    if (globs.hal == NULL)  {
        printk("IOCDriver_Init(): Error initializing IOC HAL !\n");
        goto errExit;
    }

    // Enable the IOC
    retVal = Do_Enable_IOC(ENABLE);
    if (retVal != 0) {
        IOC_HAL_Destroy(globs.hal);
        globs.hal = NULL;
        goto errExit;
    }

    // Get IOC FW Version
    retVal = verify_firmware_protocol_match();
    if (retVal != 0)  {
#ifdef ABORT_ON_BAD_VERSION
        printk("ERROR: IOC NOT initialized !!\n");
        goto exit;
#else
        printk("Loading driver anyway for diagnostics.\n");
        retVal = 0;
#endif
    }


    // Get initial AC adapter status
    update_ac_adapter_status(ac_adapter_installed());
      
    // Set version-specific capabilities
    get_ioc_firmware_version(&version,0);
    g_fiona_hw.ioc_flags |= ((version < IOC_ONE_KEY_WAKE_VERSION) ? 0 : IOC_CAN_WAKE_SINGLE_KEY_BIT);
    g_fiona_hw.ioc_flags |= ((version < IOC_DUAL_KEY_WAKE_VERSION) ? 0 : IOC_CAN_WAKE_DUAL_KEY_BIT);

    // Initialize /proc interface
    ioc_proc_init();

    // Register our driver
    retVal = misc_register(&g_ioc_misc_device);
    if (retVal < 0)  {
        misc_deregister(&g_ioc_misc_device);

        // We couldn't register the driver, but we're going to leave
        // the /proc interface in place as it should still work
        // ioc_cleanup_procfs();
        goto exit;
    }
    retVal = driver_register(&ioc_driver);
    if (!retVal)  {
        retVal = platform_device_register(&ioc_device);
        if (retVal)  {
            driver_unregister(&ioc_driver);
            goto exit;
        }
    }
    else
        goto exit;

    // Report that IOC is alive
    g_ioc_driver_present = 1;

#ifdef CONFIG_FIONA_PM_IOC
    // Register IOC driver with Fiona Power Manager,
    // we're ignoring result here as we'll still allow driver to
    // work even if we can't do efficient power management...
    if (ioc_register_with_fpow((struct device *)&ioc_driver, &g_ioc_fpow_component_ptr))
        printk("IOC: Failed to register with fpow !!\n");
#endif

#ifdef CONFIG_FIONA_PM_SCROLLWHEEL
    // Register IOC scrollwheel with Fiona Power Manager,
    // we're ignoring result here as we'll still allow driver to
    // work even if we can't do efficient power management...
    if (scroll_register_with_fpow(NULL, &g_ioc_fpow_component_ptr))
        printk("SCROLLWHEEL: Failed to register with fpow !!\n");
#endif

    //
    // If GPIO wake source is set, setup the GPIO now..
    // NOTE:  This code assumes that ioc wake can never be GPIO 0 !!
    //
    if (g_fiona_hw.ioc_wake_gpio) {
        pxa_gpio_mode(g_fiona_hw.ioc_wake_md);
        fiona_set_ioc_wake_source(1);
    }

    // Turn on dual or single key wake (HOME key by default)
    if (g_fiona_hw.ioc_flags & IOC_CAN_WAKE_DUAL_KEY_BIT)  {
        IOC_Two_Key_Wake_Enable(1);
    }
    else if (g_fiona_hw.ioc_flags & IOC_CAN_WAKE_SINGLE_KEY_BIT)  {
        IOC_One_Key_Wake_Enable(1);
    }

#ifdef CONFIG_IOC_VOLTAGE_CHECKER
    printk("WATER: Disabling IOC voltage reporting, initing our own timer...\n");
    IOC_Enable_Battery_Voltage_Low_Reporting_At_IOC(DISABLE,0);
    _IOC_Voltage_Sampler();
    start_ioc_voltage_check_thread();
    init_timer(&vc_timer);
    vc_timer.function = IOC_Voltage_Sampler_Timer;
    vc_timer.expires = (jiffies + VC_MIN_TIMER_FREQ);
    add_timer(&vc_timer);
#endif


    /* Get the op amp offset and voltage reference correction values */
    offset = get_op_amp_offset();
    if (offset == UNINITIALIZED_OP_AMP_OFFSET)
        printk("OP Amp Offset : Uninitialized\n");
    else
        printk("OP Amp Offset : %d\n",offset);

    /* Get the gain correction value */
    g_gain_correction_factor = get_gain_value();
    if (g_gain_correction_factor == UNINITIALIZED_GAIN_VALUE) {
        printk("Op Amp Gain Correction : Uninitialized\n");
    }
    else if ((g_gain_correction_factor < MIN_GAIN_CORRECTION_VALUE) || (g_gain_correction_factor > MAX_GAIN_CORRECTION_VALUE)) {
        printk("ERROR !!  Op Amp Gain Correction Value %d is not in the range %d --> %d !!\n",g_gain_correction_factor,MIN_GAIN_CORRECTION_VALUE,MAX_GAIN_CORRECTION_VALUE);
        g_gain_correction_factor = UNINITIALIZED_GAIN_VALUE;
    }
    else
        printk("Op Amp Gain Correction Factor : %d / %d\n",OP_AMP_GAIN_BASE,g_gain_correction_factor);


    /* Get the voltage reference correction values */
    g_rve_correction_value = get_rve_value();
    if (g_rve_correction_value == UNINITIALIZED_RVE_VALUE) {
        printk("Voltage Reference Correction : Uninitialized\n");
    }
    else if (g_rve_correction_value > MAX_VOLTAGE_REFERENCE_VALUE) {
        printk("ERROR !!  Voltage Reference Correction Value %d is GREATER than max of %d !!\n",g_rve_correction_value,MAX_VOLTAGE_REFERENCE_VALUE);

#ifdef CONFIG_MULTI_RATE_AVERAGING_MONITOR
        if (g_rve_correction_value > MAX_TEST_VOLTAGE_REFERENCE_VALUE) {
            printk("ERROR !!  Voltage Reference Correction Value %d is GREATER than TEST max of %d !!\n",g_rve_correction_value,MAX_TEST_VOLTAGE_REFERENCE_VALUE);
            printk("Setting Voltage Reference Correction Value to %d!!\n",UNINITIALIZED_RVE_VALUE);
            g_rve_correction_value = UNINITIALIZED_RVE_VALUE;
        }
        else
            printk("Voltage Reference Correction Factor : %d / %d\n",g_rve_correction_value,VOLTAGE_REFERENCE_BASE);
#else
        printk("Setting Voltage Reference Correction Value to %d!!\n",UNINITIALIZED_RVE_VALUE);
        g_rve_correction_value = UNINITIALIZED_RVE_VALUE;
#endif
    }
    else if (g_rve_correction_value < MIN_VOLTAGE_REFERENCE_VALUE) {
        printk("ERROR !!  Voltage Reference Correction Value %d is LESS than min of %d !!\n",g_rve_correction_value,MIN_VOLTAGE_REFERENCE_VALUE);
#ifdef CONFIG_MULTI_RATE_AVERAGING_MONITOR
        if (g_rve_correction_value < MIN_TEST_VOLTAGE_REFERENCE_VALUE) {
            printk("ERROR !!  Voltage Reference Correction Value %d is LESS than TEST min of %d !!\n",g_rve_correction_value,MIN_TEST_VOLTAGE_REFERENCE_VALUE);
            printk("Setting Voltage Reference Correction Value to %d!!\n",UNINITIALIZED_RVE_VALUE);
            g_rve_correction_value = UNINITIALIZED_RVE_VALUE;
        }
        else
            printk("Voltage Reference Correction Factor : %d / %d\n",g_rve_correction_value,VOLTAGE_REFERENCE_BASE);
#else
        printk("Setting Voltage Reference Correction Value to %d!!\n",UNINITIALIZED_RVE_VALUE);
        g_rve_correction_value = UNINITIALIZED_RVE_VALUE;
#endif
    }
    else
        printk("Voltage Reference Correction Factor : %d / %d\n",g_rve_correction_value,VOLTAGE_REFERENCE_BASE);


    /* We're up and alive, inform the user */
    printk("IOC: Fiona IOC Driver %s, IOC FW version #%d\n",IOC_DRIVER_VERSION, g_fw_version);

exit:
    // Get initial battery condition information and inform user
    IOC_Query_Battery_Info(&g_battery_info);

#ifdef USE_DEFAULT_BOARD_RESISTANCE
    g_board_resistance_set = 1;
#else
    // Calculate board resistance and re-calculate watermarks
    if (calculate_board_resistance())  {
        g_board_resistance_set = 1;
        // Board resistance just changed, recalculate watermarks
        calculate_battery_watermarks();
    }
#endif

    up(&g_ioc_hw_lock);
    lockdebug("Released g_ioc_hw_lock leaving IOCDriver_Init()...\n");

    print_battery_info_string(&g_battery_info);

    return(retVal);

errExit:
    up(&g_ioc_hw_lock);
    lockdebug("Released g_ioc_hw_lock leaving IOCDriver_Init()...\n");
    return(-EIO);
}

static int 
IOCDriver_Remove(struct device *dev)
{
#ifdef CONFIG_IOC_VOLTAGE_CHECKER
    VC_TIMEROFF();
    stop_ioc_voltage_check_thread();
#endif

    ioc_cleanup_procfs();
#ifdef CONFIG_FIONA_PM_SCROLLWHEEL
    // Unregister IOC scrollwheel driver with Fiona Power Manager
    scroll_unregister_with_fpow(g_scroll_fpow_component_ptr);
    g_scroll_fpow_component_ptr  = NULL;
#endif

#ifdef CONFIG_FIONA_PM_IOC
    // Unregister IOC driver with Fiona Power Manager
    ioc_unregister_with_fpow(g_ioc_fpow_component_ptr);
    g_ioc_fpow_component_ptr  = NULL;
#endif

    return(0);
}

static void
__exit IOCDriver_Exit(void)
{
    ioc_cleanup_procfs();
    misc_deregister(&g_ioc_misc_device);
    devfs_remove(IOC_DRIVER_NAME);
    unregister_chrdev(IOC_MAJOR,IOC_DRIVER_NAME);
    class_simple_destroy(g_ioc_class);
    IOC_HAL_Destroy(globs.hal);
    globs.hal = 0;
}

#ifdef CONFIG_IOC_VOLTAGE_CHECKER
static pid_t ioc_voltage_check_thread_pid = 0;
static DECLARE_COMPLETION(ioc_voltage_check_thread_exited);

static wait_queue_head_t voltage_checker_wq;
static int voltage_check_pending = 0;


//
// Sliding Voltage Check Delay Scale
//
#define VC_DELAY_TABLE_MV_INDEX     0
#define VC_DELAY_TABLE_DELAY_INDEX  1
#define MAX_VC_DELAY_LEVELS         2
static int vc_delay_table[MAX_VC_DELAY_LEVELS][2] = {   {3800,  (10*HZ) },
                                                        {0,     (1*HZ)  } };

/***************************************************************************
 * Routine:     get_next_vc_delay
 * Purpose:     Determine where we're at battery-level wise.  Depending on 
 *              our battery level, we adjust check frequency accordingly 
 *              (more often as we get lower due to rapid decrease characteristic
 *              of power curve at the lower part of the curve (past the knee)
 * Paramters:   none
 * Returns:     int - delay to use to reprime timer based on voltage level
 * Assumptions: none
 ***************************************************************************/
int get_next_vc_delay(void)
{
    int x = 0;
    int delay = VC_TIMER_FREQ;  // Default value...
    int mv = 0;

#ifdef CONFIG_MULTI_RATE_AVERAGING_MONITOR
    delay = VC_MIN_TIMER_FREQ;
    goto exit;
#endif

#ifndef DISABLE_SLIDING_VC_DELAY
  #ifdef PRE_FILL_VOLTAGE_TABLE
    // If we don't have a full table yet, keep frequency at a minimum until we
    // do...
    if (g_voltage_sample_count < VOLTAGE_AVERAGE_SAMPLE_COUNT)  {
        delay = VC_MIN_TIMER_FREQ;
        goto exit;
    }
  #endif
#else 
    if (g_voltage_sample_count < MIN_SAMPLES_FOR_WATERMARK_AVERAGE)  {
        delay = VC_MIN_TIMER_FREQ;
        goto exit;
    }
#endif

    get_battery_info_from_samples(&g_battery_info);
    mv = g_battery_info.mvolts;

    while ((x < MAX_VC_DELAY_LEVELS) && (mv < vc_delay_table[x][VC_DELAY_TABLE_MV_INDEX]))
        x++;
    if (x < MAX_VC_DELAY_LEVELS)
        delay = vc_delay_table[x][VC_DELAY_TABLE_DELAY_INDEX];

#if defined(DISABLE_SLIDING_VC_DELAY) || defined(PRE_FILL_VOLTAGE_TABLE)
exit:
#endif

    powdebug("VC DELAY:  %d MS     for VOLTAGE: %d mV\n",(delay*1000/HZ),mv);
    return(delay);
}

static void IOC_Voltage_Sampler_Timer(unsigned long parm)
{
    voltage_check_pending = 1;
    wake_up(&voltage_checker_wq);
#ifdef PRIME_TIMER_ON_LAST_VOLTAGE
    VC_TIMERON(get_next_vc_delay());
#endif
}

static int
ioc_voltage_check_thread(
    void *unused)
{
    daemonize("voltd");
    allow_signal(SIGKILL);

    while (1) {
        wait_event_interruptible(voltage_checker_wq, voltage_check_pending);

#ifdef CONFIG_PM
        // Check for sleep request
        if (current->flags & PF_FREEZE) {
            // Shut off the voltage checker timer
            VC_TIMEROFF();

            // Put task to sleep
            refrigerator(PF_FREEZE);

            // Make sure to populate the table with at least one value upon 
            // waking up so if a caller calls in before the next timer goes off,
            // we have some data to give them.  This keeps the "Not enough power
            // for WAN dialog that would come up if the Java world asks for a battery
            // level upon waking up (table is cleared on wakeup), but prior to the
            // next scheduled check for voltage.
            IOC_Voltage_Sampler();

            // Fire back up the timer
            VC_TIMERCONT();
            VC_TIMERON(get_next_vc_delay());

            continue;
        }
#endif

        if (signal_pending(current)) {
            break;
        }

        if (voltage_check_pending)  {
            voltage_check_pending = 0;
            IOC_Voltage_Sampler();
#ifndef PRIME_TIMER_ON_LAST_VOLTAGE
            VC_TIMERON(get_next_vc_delay());
#endif
        }
    }

    complete_and_exit(&ioc_voltage_check_thread_exited, 0);
}

static void
start_ioc_voltage_check_thread(
    void)
{
    pid_t pid;

    init_waitqueue_head(&voltage_checker_wq);

    pid = kernel_thread(ioc_voltage_check_thread, NULL, CLONE_KERNEL);
    if (pid >= 0) {
        ioc_voltage_check_thread_pid = pid;

    } else {
        ioc_voltage_check_thread_pid = 0;

    }
}


static void
stop_ioc_voltage_check_thread(
    void)
{
    int ret;

    if (ioc_voltage_check_thread_pid > 0) {
        ret = kill_proc(ioc_voltage_check_thread_pid , SIGKILL, 1);
        if (ret == 0) {
            wait_for_completion(&ioc_voltage_check_thread_exited);
        }
    }
}

#endif



static int
format_battery_info_string(char *str, Battery_Info_T *info)
{
    int len = 0;

    // If the battery is installed, report the status
    if (info->installed)  {
	    len = sprintf(str, "Battery: %d.%03d Volts, %d Degrees C, %d mA Draw, %sHARGING, %d%%\n", \
            (info->mvolts/1000), (info->mvolts%1000),info->temperature,info->draw, \
            (info->charging)?"C":"DISC",info->capacity);
    }
    else  {
	    len = sprintf(str, "Battery: NOT INSTALLED, %d mA Draw, A/C Charger is %s\n", \
            info->draw,(info->charging)?"IN":"OUT");
    }
    powdebug("BATTERY MVOLTS: %d\n",info->mvolts);
	return len;
}

int 
IOC_Hold_Lockout(int lockout)
{
    char    cmd;
    char    response[MAX_BYTES_IOC_RESPONSE];

    if (lockout)  {
#ifdef DISABLE_KEYBOARD_ON_HOLD_LOCKOUT
        // Disable keyboard input
        cmd = IOC_CMD_DISABLE_KEYBOARD;
        IOC_IssueCmd(&cmd,IOC_CMD_DISABLE_KEYBOARD_CMD_BYTES,response,IOC_CMD_DISABLE_KEYBOARD_RESPONSE_BYTES);
#endif

        // Disable buttons
        cmd = IOC_CMD_DISABLE_BUTTONS;
        IOC_IssueCmd(&cmd,IOC_CMD_DISABLE_BUTTONS_CMD_BYTES,response,IOC_CMD_DISABLE_BUTTONS_RESPONSE_BYTES);

        // Disable scroll wheel
        cmd = IOC_CMD_DISABLE_SCROLLWHEEL;
        IOC_IssueCmd(&cmd,IOC_CMD_DISABLE_SCROLLWHEEL_CMD_BYTES,response,IOC_CMD_DISABLE_SCROLLWHEEL_RESPONSE_BYTES);

        // Set our flag
        g_keys_locked = ENABLE;
    }
    else  {
#ifdef DISABLE_KEYBOARD_ON_HOLD_LOCKOUT
        // Enable keyboard input
        cmd = IOC_CMD_ENABLE_KEYBOARD;
        IOC_IssueCmd(&cmd,IOC_CMD_ENABLE_KEYBOARD_CMD_BYTES,response,IOC_CMD_ENABLE_KEYBOARD_RESPONSE_BYTES);
#endif

        // Enable buttons
        cmd = IOC_CMD_ENABLE_BUTTONS;
        IOC_IssueCmd(&cmd,IOC_CMD_ENABLE_BUTTONS_CMD_BYTES,response,IOC_CMD_ENABLE_BUTTONS_RESPONSE_BYTES);

        // Enable scroll wheel
        cmd = IOC_CMD_ENABLE_SCROLLWHEEL;
        IOC_IssueCmd(&cmd,IOC_CMD_ENABLE_SCROLLWHEEL_CMD_BYTES,response,IOC_CMD_ENABLE_SCROLLWHEEL_RESPONSE_BYTES);

        // Set our flag
        g_keys_locked = DISABLE;
    }
    return(0);
}

static int
IOC_Get_Hold_Lockout(void)
{
    return(g_keys_locked);
}


// Returns number of bytes read in response, or 0 on failure
// IF NOT expecting a response, return value of 0 indicates success
static int
Do_IOC_IssueCmd(char *cmd, int cmdSize, char *response, int responseSize, int resetChipOnErr)
{
    int rc = 0;
    int x = 0;

    // Send the command, get the response
    rc = (globs.hal->IOC_HAL_SendRcv(globs.hal,cmd,cmdSize,response,responseSize));
    if ((rc != responseSize) && !((rc == 1) && (response[0] == IOC_NO_BATTERY_RESPONSE))) {

        // we weren't able to send all data, try and
        // correct for it now by resetting chip ??
        printk("ERROR:  IOC_IssueCmd() failed (err %d) issuing %d byte cmd [ ",rc, cmdSize);
        for (x=0;x<cmdSize;x++)
            printk("%02x ",cmd[x]);
        printk("]\n");
        cdebug("  rc:%d, respSize:%d, res[0] = %x\n",rc,responseSize,response[0]);

        if (resetChipOnErr)  {
            globs.hal->IOC_HAL_ResetChip(globs.hal);
            if (IOC_Reset_Last_Wake_Condition()) {
                printk("Do_IOC_IssueCmd(): After ResetChip, call to IOC_Reset_Last_Wake_Condition() failed !!\n");
            }
        }
    
        printk("\n"); // Indicate end of dealing with "this error"

        // Return an error condition.
        // If responseSize != 0, then 0 indicates an error.
        // Otherwise, any value != 0 would indicate error.
        //
        rc = (responseSize) ? 0 : 1;     // Indicate an error 
        goto exit; 
    }
exit:
    return(rc);
}


////////////////////////////////////////////////////////
//
// Utility Routines
//
////////////////////////////////////////////////////////

unsigned int
ioc_convert_rve_voltage_to_adj_mvolts(unsigned int rve_mvolts, unsigned int draw) {
    unsigned int amv = rve_mvolts +((draw*g_min_board_resistance)/1000);
    return(amv);
}


signed int
ioc_convert_raw_temp_to_degrees_celcius(unsigned int raw)
{
    // 0x43 = 80 degrees Celcius, 0x64 = 60, 0xae = 25
    // ~2 degrees per tick
    // Thus, 0x00 = 214 degrees celcius
    return((signed int)raw);
}


unsigned int
ioc_convert_mvolts_to_raw_voltage(unsigned int mvolts)
{
    //
    // Range 0x00-0xFF is 0 - 4.2 Volts
    // Volts = raw/255 * 4.2 * (10/10) = (42*raw)/2550
    // mVolts = ((42*raw)/2550)*1000
    // raw = (mV*2550/42000)
    // printk ("IOC: ioc_convert_mvolts_to_raw_voltage(%d) = 0x%x\n",mvolts,((mvolts*255)/4200));
    return((mvolts*255)/4200);
}


unsigned int
ioc_convert_raw_voltage_to_mvolts(unsigned int raw)
{
#ifdef CONFIG_VOLTAGE_REFERENCE_CORRECTION
    //
    // Range 0x00-0xFF is 0 - 4.2 Volts
    // Volts = raw/255 * 4.2 * (10/10) = (42*raw)/2550
    // mVolts = ((42*raw)/2550)*1000
    // mVolts = mVolts * vr_correction_factor
    if (g_rve_correction_value && (g_rve_correction_value != UNINITIALIZED_RVE_VALUE))  {
        powdebug("Voltage: ((%d * 42) * 1000)/2550 = %d\n",raw,(((raw*42)*1000)/2550));
        powdebug("Voltage: %d * ( %d / %d ) = %d\n",(((raw*42)*1000)/2550),g_rve_correction_value,VOLTAGE_REFERENCE_BASE,((((raw*42)*1000)/2550)*g_rve_correction_value/VOLTAGE_REFERENCE_BASE));
        return((((raw*42)*1000)/2550)*g_rve_correction_value/VOLTAGE_REFERENCE_BASE);
    }
#endif
    //
    // Range 0x00-0xFF is 0 - 4.2 Volts
    // Volts = raw/255 * 4.2 * (10/10) = (42*raw)/2550
    // mVolts = ((42*raw)/2550)*1000
    return(((raw*42)*1000)/2550);
}


unsigned int
ioc_convert_raw_draw_to_mamps(unsigned int raw)
{
    unsigned int value = (raw * 2100 / 255);
    int offset = 0;
    int result = 0;

#ifdef CONFIG_OP_AMP_OFFSET
    offset = get_op_amp_offset();
    if (offset == UNINITIALIZED_OP_AMP_OFFSET) {
        offset = 0;     // It's not set, so use 0
    }

  #ifndef POLARITY_OF_OP_AMP_OFFSET_IS_POSITIVE
    offset *= -1;
  #endif
#endif

    //
    // We need to add in the op amp offset as the Fiona HW can vary w/ respects 
    // to how it reports current draw, so we compensate for it here by adding in 
    // the offset to our read value to determine actual value before storing it off..
    //

#ifdef CONFIG_VOLTAGE_REFERENCE_CORRECTION
    // If a voltage reference correction value exists, use it....
    if (g_rve_correction_value && (g_rve_correction_value != UNINITIALIZED_RVE_VALUE))  {
        result = (value*g_rve_correction_value/VOLTAGE_REFERENCE_BASE);
        powdebug("convert_amps rve: %d = (%d * (%d / %d))\n",result,value,g_rve_correction_value,VOLTAGE_REFERENCE_BASE);
    }
#endif


#ifdef CONFIG_OP_AMP_OFFSET
    // Apply op amp offset adjustment
    powdebug("convert_amps offset: %d = (%d + %d)\n",(value + offset),value,offset);
    result = (value + offset);
#endif


#ifdef CONFIG_AMPERAGE_GAIN_CORRECTION
    // If a gain adjustment value is given for this board, apply it to finish adjusting current
    if ((g_gain_correction_factor != 0) && (g_gain_correction_factor != UNINITIALIZED_GAIN_VALUE)) {
        powdebug("convert_amps gain: %d = ( %d * (%d/%d)\n",(result*OP_AMP_GAIN_BASE/g_gain_correction_factor),result,OP_AMP_GAIN_BASE,g_gain_correction_factor);
        result *= (OP_AMP_GAIN_BASE/g_gain_correction_factor);
    }
#endif

    // Make sure value is "sane" (i.e. a valid current that we would see, and that a
    // bad offset setting isn't screwing us up...
    if ((result < MIN_MAMP_VALUE) || (result > MAX_MAMP_VALUE)) {
        powdebug("Value %d out-of-range for current, sending back MIN current value %d\n",result,MIN_MAMP_VALUE);
        return(MIN_MAMP_VALUE);
    }
    return(result);
}


static void
print_battery_info_string(Battery_Info_T *info)
{
    int len = 0;
    char str[BATTERY_INFO_STRING_SIZE];

    memclr(str,BATTERY_INFO_STRING_SIZE);
    len = format_battery_info_string(str,info);
    if (len)
        printk(str);
}


////////////////////////////////////////////////////////
//
// /proc Interface Routines
//
////////////////////////////////////////////////////////


/******************************************************************************
 *
 * /proc/ioc/flash_fw Support
 *
 * Reading from /proc/flash_fw will return the contents of the IOC fw (4k bytes).
 *  
 * Writing data to /proc/ioc/flash_fw will flash that data to the IOC flash
 *
 ******************************************************************************/
static int proc_read_ioc_flash(char *page, char **start,
			    off_t off, int count, 
			    int *eof, void *data)
{
	int len = 0;
    int version = 0;
    int retVal = 0;
	if (count < 2)
		return -EINVAL;

    // If it's the "second" call, ignore it...
    if (off > 0)
        goto exit;

    // First, get the lock
    lockdebug("Getting g_ioc_hw_lock in proc_read_ioc_flash()...\n");
    if (down_interruptible(&g_ioc_hw_lock))
        return(-ERESTARTSYS);
    lockdebug("Got g_ioc_hw_lock in proc_read_ioc_flash()...\n");

    retVal = get_ioc_firmware_version(&version,1);

    up(&g_ioc_hw_lock);
    lockdebug("Released g_ioc_hw_lock in proc_read_ioc_flash...\n");

	len = sprintf(page, "IOC Version #%d\n",version);
exit:
    *eof = 1;
	return len;
}

/******************************************************************************/

static int proc_write_ioc_flash(struct file *file,
			     const char *buffer,
			     unsigned long orig_count, 
			     void *data)
{
    unsigned long count = orig_count;
    int rc = 0;
    int version = 0;

    if (g_fw_flash_lock) {
        printk( "IOC FW is locked from flashing.\n"
                "Echo 0 to lock to unlock for flashing.\n");
        return -EFAULT;
    }

    if (count > IOC_FW_FLASH_SIZE)
        goto usage;

    // Allocate local buffer
    if (g_full_buffer == NULL) {
        g_full_buffer_index = 0;
        g_full_buffer = kmalloc(IOC_MAX_SIZE_FW_TEXT, GFP_KERNEL);
        if (g_full_buffer == NULL) {
            return -ENOMEM;
        }
        memclr(g_full_buffer,IOC_MAX_SIZE_FW_TEXT);
    }

	if (copy_from_user(&g_full_buffer[g_full_buffer_index], buffer, count)) {
        kfree(g_full_buffer);
        g_full_buffer_index = 0;
		return -EFAULT;
	}

    // Update our index, we've got more data in the buffer
    g_full_buffer_index += count;

#ifdef DEBUG
    printk("Last 6 characters were %x %x %x %x %x %x\n",g_full_buffer[g_full_buffer_index-6],g_full_buffer[g_full_buffer_index-5],g_full_buffer[g_full_buffer_index-4],g_full_buffer[g_full_buffer_index-3],g_full_buffer[g_full_buffer_index-2],g_full_buffer[g_full_buffer_index-1]);
#endif

    // If this is the last of the data, write it out...
    if (count == 4096)
        return(orig_count);

    // Just allow one pass
    g_fw_flash_lock = 1;

    // First, get the lock
    lockdebug("Getting g_ioc_hw_lock in proc_write_ioc_flash()...\n");
    if (down_interruptible(&g_ioc_hw_lock))
        return(-ERESTARTSYS);
    lockdebug("Got g_ioc_hw_lock in proc_write_ioc_flash()...\n");

    if (get_ioc_firmware_version(&version,0) == 0)
        printk("\nFlash updating version #%d of IOC FW with %d bytes of data...\n",version, g_full_buffer_index);
    else
        printk("\nFlash updating IOC FW with %d bytes of data...\n",g_full_buffer_index);

    // Write the flash
    rc = IOC_Flash_Write(g_full_buffer,g_full_buffer_index);   // count-1 to not include proc included '\n'
    if (rc != g_full_buffer_index) {
	    printk("IOC_Flash_Write returned error %d writing %d bytes to IOC\n",rc,g_full_buffer_index);
        kfree(g_full_buffer);
        g_full_buffer = NULL;
        g_full_buffer_index = 0;
        up(&g_ioc_hw_lock);
        lockdebug("Released g_ioc_hw_lock in proc_write_ioc_flash...\n");
        return -EFAULT;
    }

    kfree(g_full_buffer);
    g_full_buffer = NULL;
    g_full_buffer_index = 0;

    // Reset the IOC and we're done....
    globs.hal->IOC_HAL_ResetChip(globs.hal);
 
    // Report results
    if (get_ioc_firmware_version(&version,1) == 0)  {
	    printk("\nIOC FW successfully flash updated to Version #%d.\n",version);
        if (IOC_Reset_Last_Wake_Condition()) {
            printk("IOC_Flash_Write(): After ResetChip, call to IOC_Reset_Last_Wake_Condition() failed !!\n");
        }
    }
    else
	    printk("\nERROR: IOC FW failed to flash update %d bytes!\n",rc);

    up(&g_ioc_hw_lock);
    lockdebug("Released g_ioc_hw_lock in proc_write_ioc_flash...\n");

	return orig_count;

usage:
    printk("Cat the S_HEX FW file to this proc entry to update IOC firmware.\n");
	return count;   // Don't error or we'll get called again
}


/******************************************************************************
 *
 * /proc/ioc/min_impedance Support
 *
 * Reading from /proc/ioc/min_impedance returns current minimum board resitance
 * value used by the system
 *
 * Writing a non-zero value to /proc/ioc/min_impedance will set the minimum
 * impedance used by the system (runtime calculation of voltage level)
 *
 ******************************************************************************/
static int proc_read_min_impedance(char *page, char **start,
			    off_t off, int count, 
			    int *eof, void *data)
{
	int len = 0;
	if (count < 2)
		return -EINVAL;

    // Skip second round calls
    if (off)
        goto exit;

	len = sprintf(page, "%d\n",g_min_board_resistance);
exit:
    *eof = 1;
	return len;
}

static int proc_write_min_impedance(struct file *file,
			     const char *buffer,
			     unsigned long count, 
			     void *data)
{
    char valueStr[32];
    int rc = 0;
    int value = 0;
    char    *new_line;

	if ((count < 2) || (count >= sizeof(valueStr)))
		goto usage;

    // Pre-clear the string storage
    memset(valueStr,0,sizeof(valueStr));

	if (copy_from_user(valueStr, buffer, count)) {
		return -EFAULT;
	}

    // Strip the new line
    new_line = strchr(valueStr,'\n');
    if (new_line)
        *new_line = '\0';

    rc = sscanf(valueStr,"%d",&value);
    if ((rc == 0) || (value == 0))
        goto usage;

    g_min_board_resistance = value;
    printk("Set minimum board resistance to %d\n",g_min_board_resistance);

	return count;

usage:
    printk("You must write a non-zero value to /proc/ioc/min_impedance to set minimum board resistance used by the system.\n");
	return count;   // Don't error or we'll get called again
}


/******************************************************************************
 *
 * /proc/ioc/max_impedance Support
 *
 * Reading from /proc/ioc/max_impedance returns current maximum board resitance
 * value used by the system
 *
 * Writing a non-zero value to /proc/ioc/max_impedance will set the maximum
 * impedance used by the system (for determining watermark levels)
 *
 ******************************************************************************/
static int proc_read_max_impedance(char *page, char **start,
			    off_t off, int count, 
			    int *eof, void *data)
{
	int len = 0;
	if (count < 2)
		return -EINVAL;

    // Skip second round calls
    if (off)
        goto exit;

	len = sprintf(page, "%d\n",g_max_board_resistance);
exit:
    *eof = 1;
	return len;
}

static int proc_write_max_impedance(struct file *file,
			     const char *buffer,
			     unsigned long count, 
			     void *data)
{
    char valueStr[32];
    int rc = 0;
    int value = 0;
    char    *new_line;

	if ((count < 2) || (count >= sizeof(valueStr)))
		goto usage;

    // Pre-clear the string storage
    memset(valueStr,0,sizeof(valueStr));

	if (copy_from_user(valueStr, buffer, count)) {
		return -EFAULT;
	}

    // Strip the new line
    new_line = strchr(valueStr,'\n');
    if (new_line)
        *new_line = '\0';

    rc = sscanf(valueStr,"%d",&value);
    if ((rc == 0) || (value == 0))
        goto usage;

    g_max_board_resistance = value;
    printk("Set maximum board resistance to %d\n",g_max_board_resistance);

    // We just changed max board resistance, so we need to recalculate watermarks
    calculate_battery_watermarks();

	return count;

usage:
    printk("You must write a non-zero value to /proc/ioc/max_impedance to set maximum board resistance used by the system.\n");
	return count;   // Don't error or we'll get called again
}



#ifdef SIMULATE_POWER_MODE_TRANSITION_FAILURE
/******************************************************************************
 *
 * /proc/ioc/fail_next_transition Support
 *
 * Reading from /proc/ioc/fail_next_transition returns
 *      0 - Test failure mode not enabled
 *      1 - IOC will purposely fail the next power mode transition call.
 *
 * Writing to /proc/ioc/fail_next_transition supports the following values:
 *      0 - Test failure mode is disabled
 *      1 - IOC will purposely fail the next power mode call, resetting this
 *          flag back to zero to assure a one-time failure mechanism.
 *
 ******************************************************************************/
static int proc_read_ioc_fail_next_transition(char *page, char **start,
			    off_t off, int count, 
			    int *eof, void *data)
{
    int len = 0;

    // Skip second round calls
    if (off)
        goto exit;
	if (count < 2)
		return -EINVAL;
	len = sprintf(page, "%d\n",g_ioc_fail_next_transition);
exit:
    *eof = 1;
	return len;
}

/******************************************************************************/

static int proc_write_ioc_fail_next_transition(struct file *file,
			     const char *buffer,
			     unsigned long count, 
			     void *data)
{
    char valueStr[2];
    int rc = 0;
    int value;
    char *new_line;

    if (count != 2)
        goto usage;

    // Pre-clear the string storage
    memset(valueStr,0,sizeof(valueStr));

	if (copy_from_user(valueStr, buffer, count)) {
		return -EFAULT;
	}

    // Strip the new line
    new_line = strchr(valueStr,'\n');
    if (new_line)
        *new_line = '\0';

    rc = sscanf(valueStr,"%d",&value);
    if ((value != 1) && (value != 0))
        goto usage;

    // Set the flag
    g_ioc_fail_next_transition = value;

	return count;

usage:
    printk("You must write a '0' or a '1' to /proc/ioc/fail_next_transition.\n");
	return count;   // Don't error or we'll get called again
}
#endif // SIMULATE_POWER_MODE_TRANSITION_FAILURE


#ifdef ENABLE_PROC_IOC_RESET
/******************************************************************************
 *
 * /proc/ioc/reset Support
 *
 * Reading from /proc/ioc/reset is not supported
 * Writing "1" to /proc/ioc/reset will reset the IOC via the HW reset line.
 *
 ******************************************************************************/
static int proc_read_ioc_reset(char *page, char **start,
			    off_t off, int count, 
			    int *eof, void *data)
{
    // Skip second round calls
    if (off)
        goto exit;

	printk("Reading /proc/ioc/reset is not supported.\n");
exit:
    *eof = 1;
	return 0;
}

static int proc_write_ioc_reset(struct file *file,
			     const char *buffer,
			     unsigned long count, 
			     void *data)
{
    char valueStr[2];
    int rc = 0;
    int value;
    char    *new_line;

    if (count != 2)
        goto usage;

    // Pre-clear the string storage
    memset(valueStr,0,sizeof(valueStr));

	if (copy_from_user(valueStr, buffer, count)) {
		return -EFAULT;
	}

    // Strip the new line
    new_line = strchr(valueStr,'\n');
    if (new_line)
        *new_line = '\0';

    rc = sscanf(valueStr,"%d",&value);
    if (value != 1)
        goto usage;

    // First, get the lock
    lockdebug("Getting g_ioc_hw_lock in proc_write_ioc_reset()...\n");
    if (down_interruptible(&g_ioc_hw_lock))
        return(-ERESTARTSYS);
    lockdebug("Got g_ioc_hw_lock in proc_write_ioc_reset()...\n");

    // Reset the IOC
    rc = globs.hal->IOC_HAL_ResetChip(globs.hal);

    up(&g_ioc_hw_lock);
    lockdebug("Released g_ioc_hw_lock in proc_write_ioc_reset...\n");

	return count;

usage:
    printk("You must write a '1' to /proc/ioc/reset to reset the IOC.\n");
	return count;   // Don't error or we'll get called again
}
#endif


/******************************************************************************
 *
 * /proc/ioc/lock Support
 *
 * Reading from /proc/ioc/lock will return the state of the g_fw_flash_lock flag,
 *  which controls the ability to flash the IOC FW.
 *  
 * Writing "0" to /proc/ioc/lock will unlock the IOC FW to ALLOW
 *      flashing via /proc interface
 * Writing "1" to /proc/ioc/lock will lock the IOC FW and NOT ALLOW
 *      flashing via /proc interface
 *
 ******************************************************************************/
static int proc_read_ioc_flash_lock(char *page, char **start,
			    off_t off, int count, 
			    int *eof, void *data)
{
	int len = 0;
	if (count < 2)
		return -EINVAL;

    // Skip second round calls
    if (off)
        goto exit;

	len = sprintf(page, "%d\n",g_fw_flash_lock);
exit:
    *eof = 1;
	return len;
}

/******************************************************************************/
#ifdef CONFIG_MULTI_RATE_AVERAGING_MONITOR
static int proc_write_ioc_vc_freq(struct file *file,
			     const char *buffer,
			     unsigned long count, 
			     void *data)
{
    char valueStr[32];
    int rc = 0;
    int value;
    int x = 0;
    char *new_line = NULL;

    if (count >= 4)
        goto usage;

    // Pre-clear the string storage
    memset(valueStr,0,sizeof(valueStr));

	if (copy_from_user(valueStr, buffer, count)) {
		return -EFAULT;
	}

    // Strip the new line
    new_line = strchr(valueStr,'\n');
    if (new_line)
        *new_line = '\0';

    rc = sscanf(valueStr,"%d",&value);
    if ((value <= 0) || (value > 200))
        goto usage;

    if (value) {
	    powdebug("proc_write_ioc_vc_freq(): Resetting voltage check frequency to %d\n",value);
        for (x=0; x<MAX_VC_DELAY_LEVELS; x++)
            vc_delay_table[x][VC_DELAY_TABLE_DELAY_INDEX] = (value*HZ);
    }

	return count;

usage:
    printk("You must write a value between '1' and '200' to /proc/ioc/vc_freq\n");
	return count;   // Don't error or we'll get called again
}


static int proc_read_ioc_vc_freq(char *page, char **start,
			    off_t off, int count, 
			    int *eof, void *data)
{
	int len = 0;

	if (count < 2)
		return -EINVAL;

    // Skip second round calls
    if (off)
        goto exit;

    printk("Voltage Check Frequency is %d seconds\n",(int)(vc_delay_table[0][VC_DELAY_TABLE_DELAY_INDEX]/HZ));
	len = sprintf(page, "%d\n",(int)(vc_delay_table[0][VC_DELAY_TABLE_DELAY_INDEX]/HZ));
exit:
    *eof = 1;
	return len;
}

static int proc_read_ioc_vc_avg_cnt(char *page, char **start,
			    off_t off, int count, 
			    int *eof, void *data)
{
	int len = 0;

	if (count < 2)
		return -EINVAL;

    // Skip second round calls
    if (off)
        goto exit;

    printk("Voltage Average Sample count is %d samples\n",(int)g_vc_sample_count);
	len = sprintf(page, "%d\n",(int)g_vc_sample_count);
exit:
    *eof = 1;
	return len;
}

static int proc_write_ioc_vc_avg_cnt(struct file *file,
			     const char *buffer,
			     unsigned long count, 
			     void *data)
{
    char valueStr[32];
    int rc = 0;
    int value;
    char *new_line = NULL;

    if (count >= 4)
        goto usage;

    // Pre-clear the string storage
    memset(valueStr,0,sizeof(valueStr));

	if (copy_from_user(valueStr, buffer, count)) {
		return -EFAULT;
	}

    // Strip the new line
    new_line = strchr(valueStr,'\n');
    if (new_line)
        *new_line = '\0';

    rc = sscanf(valueStr,"%d",&value);
    if ((value <= 0) || (value > VOLTAGE_AVERAGE_SAMPLE_COUNT))
        goto usage;

    if (value) {
	    printk("proc_write_ioc_vc_avg_cnt(): Changed voltage average sample count to %d\n",value);
        g_vc_sample_count = value;
    }

	return count;

usage:
    printk("You must write a value between '1' and '%d' to /proc/ioc/vc_avg_cnt\n",VOLTAGE_AVERAGE_SAMPLE_COUNT);
	return count;   // Don't error or we'll get called again
}

#endif // CONFIG_MULTI_RATE_AVERAGING_MONITOR


static int proc_write_ioc_reset_vtbl(struct file *file,
			     const char *buffer,
			     unsigned long count, 
			     void *data)
{
    char valueStr[2];
    int rc = 0;
    int value;
    char    *new_line;

    if (count != 2)
        goto usage;

    // Pre-clear the string storage
    memset(valueStr,0,sizeof(valueStr));

	if (copy_from_user(valueStr, buffer, count)) {
		return -EFAULT;
	}

    // Strip the new line
    new_line = strchr(valueStr,'\n');
    if (new_line)
        *new_line = '\0';

    rc = sscanf(valueStr,"%d",&value);
    if ((value != 0) && (value != 1))
        goto usage;

    if (value) {
	    powdebug("proc_write_ioc_reset_vtbl(): Resetting voltage tables...\n");
        ioc_reset_voltage_averages();
    }

	return count;

usage:
    printk("You must write a '1' or a '0' to /proc/ioc/reset_vtbl\n");
	return count;   // Don't error or we'll get called again
}


static int proc_write_ioc_flash_lock(struct file *file,
			     const char *buffer,
			     unsigned long count, 
			     void *data)
{
    char valueStr[2];
    int rc = 0;
    int value;
    char    *new_line;

    if (count != 2)
        goto usage;

    // Pre-clear the string storage
    memset(valueStr,0,sizeof(valueStr));

	if (copy_from_user(valueStr, buffer, count)) {
		return -EFAULT;
	}

    // Strip the new line
    new_line = strchr(valueStr,'\n');
    if (new_line)
        *new_line = '\0';

    rc = sscanf(valueStr,"%d",&value);
    if ((value != 0) && (value != 1))
        goto usage;

    if (value) {
	    printk("Locked IOC FW to prevent flashing.\n");
        g_fw_flash_lock = 1;
    }
    else {
	    printk("Unlocked IOC FW for flashing.\n");
        g_fw_flash_lock = 0;
    }

	return count;

usage:
    printk("You must write a '1' or a '0' to /proc/ioc/lock.\n");
	return count;   // Don't error or we'll get called again
}


/******************************************************************************
 *
 * /proc/ioc/version Support
 *
 * Reading from /proc/ioc/version will return the version of the IOC firmware
 * Writing anything to /proc/ioc/version is invalid and not supported.
 *
 ******************************************************************************/
static int proc_read_ioc_version(char *page, char **start,
			    off_t off, int count, 
			    int *eof, void *data)
{
	int len = 0;
    int version = 0;
    int retVal = 0;

	if (count < 2)
		return -EINVAL;

    // Skip second round calls
    if (off)
        goto exit;

    // First, get the lock
    lockdebug("Getting g_ioc_hw_lock in proc_write_ioc_reset()...\n");
    if (down_interruptible(&g_ioc_hw_lock))
        return(-ERESTARTSYS);
    lockdebug("Got g_ioc_hw_lock in proc_write_ioc_reset()...\n");

    retVal = get_ioc_firmware_version(&version,1);
	len = sprintf(page, "%d\n",version);

    up(&g_ioc_hw_lock);
    lockdebug("Released g_ioc_hw_lock in proc_write_ioc_reset...\n");
exit:
    *eof = 1;
	return len;
}

/******************************************************************************/

static int proc_write_ioc_version(struct file *file,
			     const char *buffer,
			     unsigned long count, 
			     void *data)
{
	printk("Writing to /proc/ioc/version is not supported.\n");
	return count;   // Don't error or we'll get called again
}




/******************************************************************************
 *
 * /proc/ioc/ac debug Support
 *
 * Reading from /proc/ioc/ac will return the state of the ac adapter,
 *  0 - Adapter NOT present
 *  1 - Adapter present
 *  
 * Writing to /proc/ioc/ac is not supported.
 *
 ******************************************************************************/
static int proc_read_ioc_ac(char *page, char **start,
			    off_t off, int count, 
			    int *eof, void *data)
{
	int len = 0;

    // If it's the "second" call, ignore it...
    if (off > 0)
        goto exit;

	if (count < 2)
		return -EINVAL;
	len = sprintf(page, "%d\n",(globs.flags & AC_ADAPTER_PRESENT_BIT)?1:0);
exit:
    *eof = 1;
	return len;
}


/******************************************************************************
 *
 * /proc/ioc_debug debug Support
 *
 * Reading from /proc/ioc_debug will return the state of the g_do_core_debug flag,
 *  which turns on or off core IOC I/O logging.
 *  
 * Writing "1" to /proc/ioc_debug will enable IOC core I/O logging
 * Writing "0" to /proc/ioc_debug will disable IOC core I/O logging
 *
 ******************************************************************************/
static int proc_read_ioc_debug(char *page, char **start,
			    off_t off, int count, 
			    int *eof, void *data)
{
	int len = 0;

    // If it's the "second" call, ignore it...
    if (off > 0)
        goto exit;

	if (count < 2)
		return -EINVAL;
	len = sprintf(page, "%d\n",g_do_core_debug);
exit:
    *eof = 1;
	return len;
}

/******************************************************************************/

static int proc_write_ioc_debug(struct file *file,
			     const char *buffer,
			     unsigned long count, 
			     void *data)
{
    char valueStr[2];
    int rc = 0;
    int value;
    char    *new_line;

    if (count != 2)
        goto usage;

    // Pre-clear the string storage
    memset(valueStr,0,sizeof(valueStr));

	if (copy_from_user(valueStr, buffer, count)) {
		return -EFAULT;
	}

    // Strip the new line
    new_line = strchr(valueStr,'\n');
    if (new_line)
        *new_line = '\0';

    rc = sscanf(valueStr,"%d",&value);
    if ((value != 0) && (value != 1))
        goto usage;

    if (value) {
	    printk("Enabled IOC core logging.\n");
        g_do_core_debug = 1;
    }
    else {
	    printk("Disabled IOC core logging.\n");
        g_do_core_debug = 0;
    }

	return count;

usage:
    printk("You must write a '1' or a '0' to /proc/ioc_debug.\n");
	return count;   // Don't error or we'll get called again
}



#ifdef CONFIG_MULTI_RATE_AVERAGING_MONITOR
/******************************************************************************
 *
 * /proc/battery Support
 *
 * Reading from /proc/battery will result in the following formatted line:
 *
 * Battery: V 0xVV, T 0xTT, DD mA Draw, CHARGING, PP%
 *  
 * where:
 *  VV = Battery voltage (0x00~0xFF = 0 Volt~5 Volt)
 *  TT = Battery temperature ( 0x43=80C, 0x64=60C, 0xAE=25C)
 *  DD = Battery current draw in decimal (mA)
 *  CHARGING ( or DISCHARGING) = current charge/discharge state
 *  PP = Actual percentge of battery capacity left
 *
 * Writing to /proc/battery is not supported.
 *
 ******************************************************************************/
int g_test_toggle = 0;
static int proc_read_battery_test(char *page, char **start,
			    off_t off, int count, 
			    int *eof, void *data)
{
	int len = 0;

	if (count < 50) {
		return -EINVAL;
	}

    // If it's the "second" call, ignore it...
    if (off > 0)
        goto exit;

    // First, get the lock
    lockdebug("Getting g_ioc_hw_lock in proc_read_battery()...\n");
    if (down_interruptible(&g_ioc_hw_lock))
        return(-ERESTARTSYS);
    lockdebug("Got g_ioc_hw_lock in proc_read_battery()...\n");

    // Get current battery condition information and report it to the user
    pxa_gpio_mode(GPIO_FIONA_VC_TEST_MD);
    g_test_toggle = 1;
    _IOC_Voltage_Sampler();      // Read another sample, average it in....
    g_fiona_voltage_debug = 0;

    get_battery_info_from_samples(&g_battery_info);
    powdebug("VOLTAGE: %d mV\n",g_battery_info.mvolts);

    up(&g_ioc_hw_lock);
    lockdebug("Released g_ioc_hw_lock in proc_read_battery()...\n");

    len = format_battery_info_string(page,&g_battery_info);
exit:
    *eof = 1;
	return len;
}
#endif


static int proc_read_battery(char *page, char **start,
			    off_t off, int count, 
			    int *eof, void *data)
{
	int len = 0;
    int pass = 1;

	if (count < 50) {
		return -EINVAL;
	}

    // If it's the "second" call, ignore it...
    if (off > 0)
        goto exit;

    // First, get the lock
    lockdebug("Getting g_ioc_hw_lock in proc_read_battery()...\n");
    if (down_interruptible(&g_ioc_hw_lock))
        return(-ERESTARTSYS);
    lockdebug("Got g_ioc_hw_lock in proc_read_battery()...\n");

    // Get current battery condition information and report it to the user
#ifdef CHECK_VOLTAGE_ON_USER_QUERY
    _IOC_Voltage_Sampler();      // Read another sample, average it in....
#endif
retry:
    get_battery_info_from_samples(&g_battery_info);
    powdebug("VOLTAGE: %d mV\n",g_battery_info.mvolts);

    //
    // If the table was cleared recently, but the voltage check timer hasn't
    // gone off yet, the table could be empty, in which case a zero will be
    // returned for mvolts.  If this happens, we are going to force a read of
    // voltage so that we can send back an actual value.  This keeps the upper level
    // from having to ask again (or display a dialog that it shouldn't have), and also
    // assures that we always have an answer when asked what the voltage is.
    // There's code in place where the table gets cleared to pre-load with a value,
    // but this is a second safeguard to GUARANTEE that it always gets a valid reading.
    //
    // If we don't have a valid reading ready after first pass, retry a voltage reading ONCE.
    //
    if ((g_battery_info.mvolts == 0) && (pass == 1))  {
        // There were no battery samples to read
        // read a sample and try again
        pass = 2; 
        _IOC_Voltage_Sampler();      // Read another sample, average it in....
        goto retry;
    }
        
    up(&g_ioc_hw_lock);
    lockdebug("Released g_ioc_hw_lock in proc_read_battery()...\n");

    len = format_battery_info_string(page,&g_battery_info);
exit:
    *eof = 1;
	return len;
}

static int proc_write_lowest_amv(struct file *file,
			     const char *buffer,
			     unsigned long count, 
			     void *data)
{
    g_lowest_amv_seen = 4200;
	return count;
}



/***************************************************************************
 * Routine:     proc_write_battery
 * Purpose:     Writing a value to the battery proc entry will set the logging
 *              report frequency (seconds between reporting battery voltage
 *              level of system)
 * Paramters:   buffer - character buffer holding report frequency
 * Returns:     none
 * Assumptions: report frequency is in seconds and can be expressed in a
 *              signed integer (although it will be treated as unsigned).
 ***************************************************************************/
static int proc_write_battery(struct file *file,
			     const char *buffer,
			     unsigned long count, 
			     void *data)
{
    char valueStr[32];
    int rc = 0;
    int value;
    char    *new_line;

    if (count > 32)
        goto usage;

    memset(valueStr,0,sizeof(valueStr));

	if (copy_from_user(valueStr, buffer, count)) {
		return -EFAULT;
	}

    // Strip the new line
    new_line = strchr(valueStr,'\n');
    if (new_line)
        *new_line = '\0';

    rc = sscanf(valueStr,"%d",&value);
    if (value < 0)
        goto usage;

	printk("Set voltage report freq to once every %d seconds (was %d).\n",value,(int)voltage_level_log_freq);

    voltage_level_log_freq = value;
    next_voltage_level_log_time = 0;    // Force check to restart using new frequency

	return count;

usage:
    printk("To set voltage report frequency, you must write a value between 0 and 2147483647 to /proc/ioc/%s.\n",IOC_PROC_BATTERY_ENTRY_NAME);
	return count;   // Don't error or we'll get called again
}


/******************************************************************************
 *
 * /proc/ioc support
 *
 * Reading from /proc/ioc will result in:
 *  Nothing (N/A)
 *
 * Writing to /proc/ioc will result in:
 *
 * Writing 'E' -> Enable HW Feature
 * Writing 'D' -> Disable HW Feature
 * Writing 'Q' -> Execute query command, display result
 *  + Q_NUM     -> Query command number
 * Writing 'R' -> Reset IOC Chip
 * Writing 'T' -> Test IOC (via Echo cmd and response)
 *
 ******************************************************************************/

static int proc_read_ioc(char *page, char **start,
			    off_t off, int count, 
			    int *eof, void *data)
{
	int len = 0;

	if (count < 23) {
		return -EINVAL;
	}

    // Skip second round calls
    if (off)
        goto exit;

	len = sprintf(page, "Not Implemented Yet\n");

exit:
    *eof = 1;
	return len;
}


/******************************************************************************/

static void print_proc_ioc_usage(const char *cmd)
{
    printk( "Invalid format '%s' for /proc/ioc/ioc.\n\n"
        "Usage: \n"
        "   SendByteCount:Byte1:...:ByteN:ResponseByteCount\n"
        "       SendByteCount: Hex number representing number of bytes to send to IOC.\n"
        "           Max value for SendByteCount is 4.  To send a PN-LCD message, use\n"
        "           /proc/ioc/lcd instead.\n"
        "       Byte1:  First byte to send to IOC (represented in hex)\n"
        "       ByteN:  Last byte to send to IOC (represented in hex)\n"
        "       ResponseByteCount: Number of bytes to expect as a response from IOC.\n"
        "\n"   
        "Example Usages: \n"
        "  Get firmware version of IOC:\n"
        "        echo '2:F3:3:1' >> /proc/ioc/ioc\n\n"
        "  The above command says to write 2 bytes of data to the IOC and to expect\n"
        "  1 byte of response data back from the IOC.  The two bytes it will send is \n"
        "  0xF3 followed 0x03, which is the command to request the IOC firmware \n"
        "  version from the IOC.\n\n",cmd);
}


/******************************************************************************/


static int 
parse_proc_write_format(const char *cmd, char *cmd_buff, int *cmd_byte_count, int *result_count)
{
    int bytes_to_write = 0;
    int buff[32];
    int     x = 0;
    int     rc = 0;

    if (cmd == NULL)
        goto bad_usage_error;

    memclr(buff,32);

#ifdef DEBUG_IOC_COMMAND_SEND
    printk("Parsing ioc command = %s\n",cmd);
#endif  // DEBUG_IOC_COMMAND_SEND

    rc = sscanf(cmd,"%d:%x:%x:%x:%x:%x:%x:%x:%x"
                    "%x:%x:%x:%x:%x:%x:%x:%x"
                    "%x:%x:%x:%x:%x:%x:%x:%x"
                    "%x:%x:%x:%x:%x:%x:%x:%x"
                    "\n",&bytes_to_write, 
            &buff[0], &buff[1], &buff[2], &buff[3], &buff[4], &buff[5], &buff[6], &buff[7],
            &buff[8], &buff[9], &buff[10], &buff[11], &buff[12], &buff[13], &buff[14], &buff[15],
            &buff[16], &buff[17], &buff[18], &buff[19], &buff[20], &buff[21], &buff[22], &buff[23],
            &buff[24], &buff[25], &buff[26], &buff[27], &buff[28], &buff[29], &buff[30], &buff[31]); 

    // If it wasn't a valid format, error
    if (rc == 0)
        goto bad_usage_error;

    if (bytes_to_write > 32)  {
            printk("Sorry, cannot write more than 32 bytes to IOC.\n");
            return(0);
    }
    
    // Create our command string
    for (x=0;x<bytes_to_write;x++)  {
        cmd_buff[x] = (char) buff[x];
    }

    // Copy converted data back to buffer
#ifdef DEBUG_IOC_COMMAND_SEND
    printk("Converted bytes: ");
    for (x=0;x<bytes_to_write;x++)  {
        printk("0x%02x ",buff[x]);
    }
    printk("\n");
#endif // DEBUG_IOC_COMMAND_SEND

    // The next item at buff[x] is the result count
    *result_count = buff[x];

    // Pass back the total bytes to write for the command
    *cmd_byte_count = bytes_to_write;
    
    return(0);

bad_usage_error:
    print_proc_ioc_usage(cmd);
    *result_count = 0;
    return(1);
}

static int proc_write_ioc(struct file *file,
			     const char *buffer,
			     unsigned long count, 
			     void *data)
{
	char	cmd[256];
    int     x = 0;
    int     rc = 0;
    char    cmd_bytes[5];
    int     result_count = 0;
    char    result[256];
    int     cmd_byte_count = 0;
    char    *new_line = NULL;

	if (count > 256) {
		return -EINVAL;
	}

    // Pre-clear the string storage
    memset(cmd,0,sizeof(cmd));

	if (copy_from_user(cmd, buffer, count)) {
		return -EFAULT;
	}

    // Strip the new line
    new_line = strchr(cmd,'\n');
    if (new_line)
        *new_line = '\0';

    rc = parse_proc_write_format(cmd,cmd_bytes,&cmd_byte_count, &result_count);
    if (rc != 0) 
		return count;   // Return no-error so it doesn't try to issue write again

#ifdef DEBUG_IOC_COMMAND_SEND
    else {
        printk("proc_write_ioc (cmd_byte_count = 0x%02x):",cmd_byte_count);
        for (x=0; x<cmd_byte_count; x++)  {
            printk(" 0x%02x",cmd_bytes[x]);
        }
        printk("\n");
    }
#endif  // DEBUG_IOC_COMMAND_SEND

    // First, get the lock
    lockdebug("Getting g_ioc_hw_lock in proc_write_ioc()...\n");
    if (down_interruptible(&g_ioc_hw_lock))
        return(-ERESTARTSYS);
    lockdebug("Got g_ioc_hw_lock in proc_write_ioc()...\n");

    rc = IOC_IssueCmd(cmd_bytes,cmd_byte_count,result,result_count);

    up(&g_ioc_hw_lock);
    lockdebug("Released g_ioc_hw_lock in proc_write_ioc()...\n");

    if (rc != result_count)
        goto exit; // don't return error or we'll keep getting called !!

    printk("IOC Result :");
    for (x=0; x<result_count; x++)  {
        printk(" 0x%02x",result[x]);
    }
    printk("\n");
exit:

	return count;
}

#ifdef ENABLE_ALLOW_SLEEP_PROC

/******************************************************************************
 *
 * /proc/ioc/allow_sleep Support
 *
 * Reading from /proc/ioc/allow_sleep will return whether or not the IOC
 * will sleep when the System sleeps (i.e. the state of the g_ioc_allow_sleep flag).
 *
 *
 * Writing "1" to /proc/ioc/allow_sleep will cause IOC to sleep when
 *              system sleeps (default).
 * Writing "0" to /proc/ioc/allow_sleep will disable IOC core I/O logging
 *
 ******************************************************************************/
static int proc_read_ioc_allow_sleep(char *page, char **start,
                           off_t off, int count,
                           int *eof, void *data)
{
       int len = 0;

       if (count < 2)
               return -EINVAL;

    // If it's the "second" call, ignore it...
    if (off > 0)
        goto exit;

       len = sprintf(page, "%d\n",g_ioc_allow_sleep_flag);
exit:
    *eof = 1;
       return len;
}

/******************************************************************************/

static int proc_write_ioc_allow_sleep(struct file *file,
                            const char *buffer,
                            unsigned long count,
                            void *data)
{
    char valueStr[2];
    int rc = 0;
    int value;
    char    *new_line;

    if (count != 2)
        goto usage;

    // Pre-clear the string storage
    memset(valueStr,0,sizeof(valueStr));

    if (copy_from_user(valueStr, buffer, count)) {
        return -EFAULT;
    }

    // Strip the new line
    new_line = strchr(valueStr,'\n');
    if (new_line)
        *new_line = '\0';

    rc = sscanf(valueStr,"%d",&value);
    if ((value != 0) && (value != 1))
        goto usage;

    if (value) {
        printk("IOC instructed to sleep on system sleep.\n");
        g_ioc_allow_sleep_flag = 1;
    }
    else {
        printk("IOC instructed NOT to sleep on system sleep.\n");
        g_ioc_allow_sleep_flag = 0;
    }

    return count;
usage:
    printk("You must write a '1' or a '0' to /proc/ioc/allow_sleep.\n");
    return count;
}

#endif


/******************************************************************************/

static int ioc_proc_init(void)
{
	int rc = 0;

    // 
	// Create /proc/ioc directory
    //
    proc_ioc_parent = create_proc_entry(IOC_PROC_PARENT_NAME, S_IFDIR | S_IRUGO | S_IXUGO, NULL);
    if (!proc_ioc_parent) {
		rc = -ENOMEM;
		goto exit;
    }

    // 
	// Create /proc/ioc/ioc entry 
    //
	proc_ioc = create_proc_entry(IOC_PROC_ENTRY_NAME, 0666, proc_ioc_parent);
	if (!proc_ioc ) {
		rc = -ENOMEM;
		goto exit;
	}
	proc_ioc->owner	= THIS_MODULE;
	proc_ioc->data	= NULL;
	proc_ioc->read_proc	= proc_read_ioc;
	proc_ioc->write_proc	= proc_write_ioc;

	/* All's good */
	PDEBUG("/proc/ioc/%s initialized.\n",IOC_PROC_ENTRY_NAME);
		

    // 
	// Create /proc/ioc/camv entry 
    //
	proc_ioc_camv = create_proc_entry(IOC_PROC_CLEAR_AMV_ENTRY_NAME, 0666, proc_ioc_parent);
	if (!proc_ioc_camv ) {
		rc = -ENOMEM;
		goto exit;
	}
	proc_ioc_camv->owner	= THIS_MODULE;
	proc_ioc_camv->data	= NULL;
	proc_ioc_camv->read_proc	= NULL;
	proc_ioc_camv->write_proc = proc_write_lowest_amv;

	/* All's good */
	PDEBUG("/proc/ioc/%s initialized.\n",IOC_PROC_CLEAR_AMV_ENTRY_NAME);


    // 
	// Create /proc/ioc/battery entry 
    //
	proc_ioc_battery = create_proc_entry(IOC_PROC_BATTERY_ENTRY_NAME, 0666, proc_ioc_parent);
	if (!proc_ioc_battery) {
		rc = -ENOMEM;
		goto exit;
	}
	proc_ioc_battery->owner	= THIS_MODULE;
	proc_ioc_battery->data	= NULL;
	proc_ioc_battery->read_proc	= proc_read_battery;
	proc_ioc_battery->write_proc = proc_write_battery;

	/* All's good */
	PDEBUG("/proc/ioc/%s initialized.\n",IOC_PROC_BATTERY_ENTRY_NAME);

#ifdef ENABLE_PROC_IOC_RESET
    // 
	// Create /proc/ioc/reset entry 
    //
	proc_ioc_reset = create_proc_entry(IOC_PROC_RESET_ENTRY_NAME, 0666, proc_ioc_parent);
	if (!proc_ioc_reset) {
		rc = -ENOMEM;
		goto exit;
	}
	proc_ioc_reset->owner	= THIS_MODULE;
	proc_ioc_reset->data	= NULL;
	proc_ioc_reset->read_proc	= proc_read_ioc_reset;
	proc_ioc_reset->write_proc = proc_write_ioc_reset;

	/* All's good */
	PDEBUG("/proc/ioc/%s initialized.\n",IOC_PROC_RESET_ENTRY_NAME);
#endif

#ifdef SIMULATE_POWER_MODE_TRANSITION_FAILURE
    // 
	// Create /proc/ioc/fail_next_transition entry 
    //
	proc_ioc_fail_next_transition = create_proc_entry(IOC_PROC_FAIL_TRANSITION_ENTRY_NAME, 0666, proc_ioc_parent);
	if (!proc_ioc_fail_next_transition) {
		rc = -ENOMEM;
		goto exit;
	}
	proc_ioc_fail_next_transition->owner	= THIS_MODULE;
	proc_ioc_fail_next_transition->data	= NULL;
	proc_ioc_fail_next_transition->read_proc	= proc_read_ioc_fail_next_transition;
	proc_ioc_fail_next_transition->write_proc = proc_write_ioc_fail_next_transition;

	/* All's good */
	PDEBUG("/proc/ioc/%s initialized.\n",IOC_PROC_FAIL_TRANSITION_ENTRY_NAME);
#endif // SIMULATE_POWER_MODE_TRANSITION_FAILURE


#ifdef ENABLE_ALLOW_SLEEP_PROC
    //
    // Create /proc/ioc/allow_sleep entry
    //
    proc_ioc_allow_sleep = create_proc_entry(IOC_PROC_ALLOW_SLEEP_ENTRY_NAME, 0666, proc_ioc_parent);
    if (!proc_ioc_allow_sleep) {
        rc = -ENOMEM;
        goto exit;
    }
    proc_ioc_allow_sleep->owner     = THIS_MODULE;
    proc_ioc_allow_sleep->data      = NULL;
    proc_ioc_allow_sleep->read_proc = proc_read_ioc_allow_sleep;
    proc_ioc_allow_sleep->write_proc = proc_write_ioc_allow_sleep;

    /* All's good */
    PDEBUG("/proc/ioc/%s initialized.\n",IOC_PROC_DEBUG_ENTRY_NAME);
#endif

    // 
	// Create /proc/ioc/debug entry 
    //
	proc_ioc_debug = create_proc_entry(IOC_PROC_DEBUG_ENTRY_NAME, 0666, proc_ioc_parent);
	if (!proc_ioc_debug) {
		rc = -ENOMEM;
		goto exit;
	}
	proc_ioc_debug->owner	= THIS_MODULE;
	proc_ioc_debug->data	= NULL;
	proc_ioc_debug->read_proc	= proc_read_ioc_debug;
	proc_ioc_debug->write_proc = proc_write_ioc_debug;


	/* All's good */
	PDEBUG("/proc/ioc/%s initialized.\n",IOC_PROC_DEBUG_ENTRY_NAME);


    // 
	// Create /proc/ioc/ac entry 
    //
	proc_ioc_ac = create_proc_entry(IOC_PROC_AC_ENTRY_NAME, 0666, proc_ioc_parent);
	if (!proc_ioc_ac) {
		rc = -ENOMEM;
		goto exit;
	}
	proc_ioc_ac->owner	= THIS_MODULE;
	proc_ioc_ac->data	= NULL;
	proc_ioc_ac->read_proc	= proc_read_ioc_ac;
	proc_ioc_ac->write_proc = NULL;


	/* All's good */
	PDEBUG("/proc/ioc/%s initialized.\n",IOC_PROC_DEBUG_ENTRY_NAME);


    // 
	// Create /proc/ioc/version entry 
    //
	proc_ioc_version = create_proc_entry(IOC_PROC_VERSION_ENTRY_NAME, 0666, proc_ioc_parent);
	if (!proc_ioc_version) {
		rc = -ENOMEM;
		goto exit;
	}
	proc_ioc_version->owner	= THIS_MODULE;
	proc_ioc_version->data	= NULL;
	proc_ioc_version->read_proc	= proc_read_ioc_version;
	proc_ioc_version->write_proc = proc_write_ioc_version;


	/* All's good */
	PDEBUG("/proc/ioc/%s initialized.\n",IOC_PROC_VERSION_ENTRY_NAME);

    // 
	// Create /proc/ioc/lock entry 
    //
	proc_ioc_lock = create_proc_entry(IOC_PROC_LOCK_ENTRY_NAME, 0666, proc_ioc_parent);
	if (!proc_ioc_lock) {
		rc = -ENOMEM;
		goto exit;
	}
	proc_ioc_lock->owner	= THIS_MODULE;
	proc_ioc_lock->data	= NULL;
	proc_ioc_lock->read_proc	= proc_read_ioc_flash_lock;
	proc_ioc_lock->write_proc = proc_write_ioc_flash_lock;


	/* All's good */
	PDEBUG("/proc/ioc/%s initialized.\n",IOC_PROC_LOCK_ENTRY_NAME);
		
    // 
	// Create /proc/ioc/reset_vtbl entry 
    //
	proc_ioc_reset_vtbl = create_proc_entry(IOC_PROC_RESET_VTBL_ENTRY_NAME, 0666, proc_ioc_parent);
	if (!proc_ioc_reset_vtbl) {
		rc = -ENOMEM;
		goto exit;
	}
	proc_ioc_reset_vtbl->owner	= THIS_MODULE;
	proc_ioc_reset_vtbl->data	= NULL;
	proc_ioc_reset_vtbl->write_proc = proc_write_ioc_reset_vtbl;
	proc_ioc_reset_vtbl->read_proc = NULL;


	/* All's good */
	PDEBUG("/proc/ioc/%s initialized.\n",IOC_PROC_RESET_VTBL_ENTRY_NAME);

#ifdef CONFIG_MULTI_RATE_AVERAGING_MONITOR
    // 
	// Create /proc/ioc/vc entry 
    //
	proc_ioc_battery_vc = create_proc_entry(IOC_PROC_VC_ENTRY_NAME, 0666, proc_ioc_parent);
	if (!proc_ioc_battery_vc) {
		rc = -ENOMEM;
		goto exit;
	}
	proc_ioc_battery_vc->owner	= THIS_MODULE;
	proc_ioc_battery_vc->data	= NULL;
	proc_ioc_battery_vc->read_proc	= proc_read_battery_test;
	proc_ioc_battery_vc->write_proc = NULL;

	/* All's good */
	PDEBUG("/proc/ioc/%s initialized.\n",IOC_PROC_VC_ENTRY_NAME);


    // 
	// Create /proc/ioc/vc_frequency entry 
    //
	proc_ioc_vc_freq = create_proc_entry(IOC_PROC_VC_FREQ_ENTRY_NAME, 0666, proc_ioc_parent);
	if (!proc_ioc_vc_freq) {
		rc = -ENOMEM;
		goto exit;
	}
	proc_ioc_vc_freq->owner	= THIS_MODULE;
	proc_ioc_vc_freq->data	= NULL;
	proc_ioc_vc_freq->write_proc = proc_write_ioc_vc_freq;
	proc_ioc_vc_freq->read_proc = proc_read_ioc_vc_freq;


	/* All's good */
	PDEBUG("/proc/ioc/%s initialized.\n",IOC_PROC_VC_FREQ_ENTRY_NAME);


    // 
	// Create /proc/ioc/vc_sample_cnt entry 
    //
	proc_ioc_vc_avg_cnt = create_proc_entry(IOC_PROC_VC_AVG_CNT_ENTRY_NAME, 0666, proc_ioc_parent);
	if (!proc_ioc_vc_avg_cnt) {
		rc = -ENOMEM;
		goto exit;
	}
	proc_ioc_vc_avg_cnt->owner	= THIS_MODULE;
	proc_ioc_vc_avg_cnt->data	= NULL;
	proc_ioc_vc_avg_cnt->write_proc = proc_write_ioc_vc_avg_cnt;
	proc_ioc_vc_avg_cnt->read_proc = proc_read_ioc_vc_avg_cnt;


	/* All's good */
	PDEBUG("/proc/ioc/%s initialized.\n",IOC_PROC_VC_AVG_CNT_ENTRY_NAME);

#endif // CONFIG_MULTI_RATE_AVERAGING_MONITOR

    // 
	// Create /proc/ioc/flash_fw entry 
    //
	proc_ioc_flash = create_proc_entry(IOC_PROC_FLASH_FW_ENTRY_NAME, 0666, proc_ioc_parent);
	if (!proc_ioc_flash) {
		rc = -ENOMEM;
		goto exit;
	}
	proc_ioc_flash->owner	= THIS_MODULE;
	proc_ioc_flash->data	= NULL;
	proc_ioc_flash->read_proc	= proc_read_ioc_flash;
	proc_ioc_flash->write_proc = proc_write_ioc_flash;


    //
    // Create /proc/ioc/min_impedance entry
    //
    proc_ioc_min_impedance = create_proc_entry(IOC_PROC_MIN_IMPEDANCE_ENTRY_NAME, 0666, proc_ioc_parent);
    if (!proc_ioc_min_impedance) {
        rc = -ENOMEM;
        goto exit;
    }
    proc_ioc_min_impedance->owner   = THIS_MODULE;
    proc_ioc_min_impedance->data    = NULL;
    proc_ioc_min_impedance->read_proc   = proc_read_min_impedance;
    proc_ioc_min_impedance->write_proc = proc_write_min_impedance;


    //
    // Create /proc/ioc/max_impedance entry
    //
    proc_ioc_max_impedance = create_proc_entry(IOC_PROC_MAX_IMPEDANCE_ENTRY_NAME, 0666, proc_ioc_parent);
    if (!proc_ioc_max_impedance) {
        rc = -ENOMEM;
        goto exit;
    }
    proc_ioc_max_impedance->owner   = THIS_MODULE;
    proc_ioc_max_impedance->data    = NULL;
    proc_ioc_max_impedance->read_proc   = proc_read_max_impedance;
    proc_ioc_max_impedance->write_proc = proc_write_max_impedance;

exit:
	return(rc);
}


static void ioc_cleanup_procfs(void)
{
    if (proc_ioc_debug) {
	    remove_proc_entry(IOC_PROC_DEBUG_ENTRY_NAME, proc_ioc_parent);
	    PDEBUG("/proc/ioc/%s file entry removed.\n",IOC_PROC_DEBUG_ENTRY_NAME);
        proc_ioc_debug = NULL;
    }
    if (proc_ioc_ac) {
	    remove_proc_entry(IOC_PROC_AC_ENTRY_NAME, proc_ioc_parent);
	    PDEBUG("/proc/ioc/%s file entry removed.\n",IOC_PROC_AC_ENTRY_NAME);
        proc_ioc_ac = NULL;
    }
#ifdef ENABLE_PROC_IOC_RESET
    if (proc_ioc_reset) {
	    remove_proc_entry(IOC_PROC_RESET_ENTRY_NAME, proc_ioc_parent);
	    PDEBUG("/proc/ioc/%s file entry removed.\n",IOC_PROC_RESET_ENTRY_NAME);
        proc_ioc_reset = NULL;
    }
#endif

#ifdef ENABLE_ALLOW_SLEEP_PROC
if (proc_ioc_allow_sleep) {
    remove_proc_entry(IOC_PROC_ALLOW_SLEEP_ENTRY_NAME, proc_ioc_parent);
    PDEBUG("/proc/ioc/%s file entry removed.\n",IOC_PROC_ALLOW_SLEEP_ENTRY_NAME);
    proc_ioc_allow_sleep= NULL;
}
#endif

#ifdef SIMULATE_POWER_MODE_TRANSITION_FAILURE
    if (proc_ioc_fail_next_transition) {
	    remove_proc_entry(IOC_PROC_FAIL_TRANSITION_ENTRY_NAME, proc_ioc_parent);
	    PDEBUG("/proc/ioc/%s file entry removed.\n",IOC_PROC_FAIL_TRANSITION_ENTRY_NAME);
        proc_ioc_fail_next_transition = NULL;
    }
#endif // SIMULATE_POWER_MODE_TRANSITION_FAILURE
    if (proc_ioc_version) {
	    remove_proc_entry(IOC_PROC_VERSION_ENTRY_NAME, proc_ioc_parent);
	    PDEBUG("/proc/ioc/%s file entry removed.\n",IOC_PROC_VERSION_ENTRY_NAME);
        proc_ioc_version = NULL;
    }
    if (proc_ioc) {
	    remove_proc_entry(IOC_PROC_ENTRY_NAME, proc_ioc_parent);
	    PDEBUG("/proc/ioc/%s file entry removed.\n",IOC_PROC_ENTRY_NAME);
        proc_ioc = NULL;
    }
#ifdef CONFIG_MULTI_RATE_AVERAGING_MONITOR
    if (proc_ioc_vc_freq)  {
	    remove_proc_entry(IOC_PROC_VC_FREQ_ENTRY_NAME, proc_ioc_parent);
	    PDEBUG("/proc/ioc/%s file entry removed.\n",IOC_PROC_VC_FREQ_ENTRY_NAME);
        proc_ioc_vc_freq = NULL;
    }
    if (proc_ioc_vc_avg_cnt)  {
	    remove_proc_entry(IOC_PROC_VC_AVG_CNT_ENTRY_NAME, proc_ioc_parent);
	    PDEBUG("/proc/ioc/%s file entry removed.\n",IOC_PROC_VC_AVG_CNT_ENTRY_NAME);
        proc_ioc_vc_avg_cnt = NULL;
    }
#endif // CONFIG_MULTI_RATE_AVERAGING_MONITOR
    if (proc_ioc_reset_vtbl)  {
	    remove_proc_entry(IOC_PROC_RESET_VTBL_ENTRY_NAME, proc_ioc_parent);
	    PDEBUG("/proc/ioc/%s file entry removed.\n",IOC_PROC_RESET_VTBL_ENTRY_NAME);
        proc_ioc_reset_vtbl = NULL;
    }
    if (proc_ioc_flash)  {
	    remove_proc_entry(IOC_PROC_FLASH_FW_ENTRY_NAME, proc_ioc_parent);
	    PDEBUG("/proc/ioc/%s file entry removed.\n",IOC_PROC_FLASH_FW_ENTRY_NAME);
        proc_ioc_flash = NULL;
    }
    if (proc_ioc_min_impedance)  {
        remove_proc_entry(IOC_PROC_MIN_IMPEDANCE_ENTRY_NAME, proc_ioc_parent);
        PDEBUG("/proc/ioc/%s file entry removed.\n",IOC_PROC_MIN_IMPEDANCE_ENTRY_NAME);
        proc_ioc_min_impedance = NULL;
    }
    if (proc_ioc_max_impedance)  {
        remove_proc_entry(IOC_PROC_MAX_IMPEDANCE_ENTRY_NAME, proc_ioc_parent);
        PDEBUG("/proc/ioc/%s file entry removed.\n",IOC_PROC_MAX_IMPEDANCE_ENTRY_NAME);
        proc_ioc_max_impedance = NULL;
    }
#ifdef CONFIG_MULTI_RATE_AVERAGING_MONITOR
    if (proc_ioc_battery_vc)  {
	    remove_proc_entry(IOC_PROC_VC_ENTRY_NAME, proc_ioc_parent);
	    PDEBUG("/proc/ioc/%s file entry removed.\n",IOC_PROC_VC_ENTRY_NAME);
        proc_ioc_battery_vc = NULL;
    }
#endif
    if (proc_ioc_battery)  {
	    remove_proc_entry(IOC_PROC_BATTERY_ENTRY_NAME, proc_ioc_parent);
	    PDEBUG("/proc/ioc/%s file entry removed.\n",IOC_PROC_BATTERY_ENTRY_NAME);
        proc_ioc_battery = NULL;
    }
    if (proc_ioc_camv)  {
	    remove_proc_entry(IOC_PROC_CLEAR_AMV_ENTRY_NAME, proc_ioc_parent);
	    PDEBUG("/proc/ioc/%s file entry removed.\n",IOC_PROC_CLEAR_AMV_ENTRY_NAME);
        proc_ioc_camv = NULL;
    }
    if (proc_ioc_parent)  {
	    remove_proc_entry(IOC_PROC_PARENT_NAME, NULL);
	    PDEBUG("/proc/%s directory removed.\n",IOC_PROC_PARENT_NAME);
        proc_ioc_parent = NULL;
    }
}

module_init(IOCDriver_Init);
module_exit(IOCDriver_Exit);

MODULE_AUTHOR("Nick Vaccaro");
MODULE_LICENSE("GPL");
