/*
 * Sample code for a driver supporting the
 * Fiona Power Management System (fpow)
 * 
 * Copyright (C) 2006 Lab126, Inc.  All rights reserved. <nvaccaro@lab126.com>
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/completion.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/tty.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/proc_fs.h>
#include <linux/syscalls.h>
#include <asm/uaccess.h>
#include <linux/fb.h>
#include <linux/init.h>
#include <linux/suspend.h>
#include <asm/system.h>
#include <asm/arch/system.h>
#include <asm/arch/fiona.h>
#include <asm/arch/fpow.h>
#include <linux/ioc.h>
#include "ioc_cmds.h"
#include "fiona_ioc_hal.h"

// global holding current FPOW power state of the driver
FPOW_POWER_MODE  g_ioc_power_mode = FPOW_MODE_ON;

// holds current IOC sleeping status
int g_ioc_sleeping = 0;

// Hardware access semaphore
extern struct semaphore g_ioc_hw_lock;

// If set, forces IOC to NOT allow sleep requests (user set via /proc)
extern int g_ioc_allow_sleep_flag;

// Set if low battery warnings are disabled at IOC
extern int g_low_battery_warnings_enabled;

// Set if IOC is currently enabled
extern int g_ioc_enabled;

#ifdef SIMULATE_POWER_MODE_TRANSITION_FAILURE
extern int g_ioc_fail_next_transition;
#endif  // SIMULATE_POWER_MODE_TRANSITION_FAILURE

// Static global holding fpow device entry returned from fpow_register()
struct fpow_component *g_ioc_fpow_component_ptr = NULL;

// External Prototypes
extern int fiona_get_wake_source(void);
extern void ioc_set_sleep_flag(void);
extern void get_battery_info_from_samples(Battery_Info_T *info);
extern int  IOC_Voltage_Sampler(void);

// Local prototypes
static int IOCDriver_Suspend(u32 mode);
static int IOCDriver_Resume(u32 mode);


/*********************************************************
 * Routine:     ioc_fpow_setmode
 *
 * Purpose:     Handled Fiona Power Management callbacks
 *              for doing IOC power management.
 *
 * Paramters:   private - value passed at registration - ignored
 *              state   - ignored
 *              mode    - Power mode to switch to
 *                        Currently supported Modes:
 *                          FPOW_MODE_ON
 *                          FPOW_MODE_SLEEP
 *                          FPOW_MODE_OFF
 *
 * Returns:     
 *  FPOW_NO_ERR                     - Transitioned to mode with no errors.
 *  -FPOW_NOT_ENOUGH_POWER_ERR      - not enough power to transition
 *  -FPOW_COMPONENT_TRANSITION_ERR  - Failed to transition to mode
 *  -EINVAL                         - Invalid mode
 *
 * Assumptions: none
 *********************************************************/

static FPOW_ERR
ioc_fpow_setmode(void *private, u32 state, u32 mode)
{
	int	retVal = FPOW_NO_ERR;
    Battery_Info_T info;

    // If we're in the right mode, we don't need to do anything
    if (g_ioc_power_mode == mode)  {
        powdebug("ioc_fpow_setmode(%s): driver already in %s mode, ignoring call.\n",fpow_mode_to_str(mode),fpow_mode_to_str(g_ioc_power_mode));
        goto exit;
    }
        
    powdebug("ioc_fpow_setmode(%s)\n",fpow_mode_to_str(mode));

#ifdef SIMULATE_POWER_MODE_TRANSITION_FAILURE
    if (g_ioc_fail_next_transition)  {
        retVal = -FPOW_NOT_ENOUGH_POWER_ERR;
        g_ioc_fail_next_transition = 0;
        powdebug("IOC: Purposely failing transition request, returning %d\n",retVal);
        return(retVal);
    }
#endif  // SIMULATE_POWER_MODE_TRANSITION_FAILURE

        
	// Switch to the appropriate power level
	switch (mode)  {
		case FPOW_MODE_ON:
			retVal = IOCDriver_Resume(mode); 

            // Read a sample, get our average...
            IOC_Voltage_Sampler();
            get_battery_info_from_samples(&info);
            printk("IOC: Woke up with voltage = %d.\n",info.mvolts);
			break;

		case FPOW_MODE_SLEEP:
		case FPOW_MODE_OFF:
			retVal = IOCDriver_Suspend(mode); 
			break;

		default:
			retVal = -EINVAL;
			break;
	}

	powdebug("IOC: driver now in %s mode (retVal=%d)\n",fpow_mode_to_str(g_ioc_power_mode),retVal);
exit:
    return(retVal);
}

static int
ioc_fpow_getmode(void *private)
{
    extpowdebug("ioc_fpow_getmode(0x%x) returning %s\n",(int)private,fpow_mode_to_str(g_ioc_power_mode));
	return (g_ioc_power_mode);
}


int
ioc_awake(void)
{
    switch(ioc_fpow_getmode(NULL))  {
        case FPOW_MODE_ON:
            return(1);
        case FPOW_MODE_SLEEP:
        case FPOW_MODE_OFF:
        default:
            return(0);
    }
}
//
// Power transition utility routines
//
static int
Do_Sleep_IOC(void)
{
    int retVal = 0;
#ifdef ALLOW_LOW_BATTERY_TO_WAKE_US
    int mvolts = 0;
#endif

    // If we're not supposed to sleep, don't
    if (0 == g_ioc_allow_sleep_flag)
        return(1);  // Return error - not sleeping

    // If we're not already sleeping, go to sleep now...
    if (g_ioc_power_mode != FPOW_MODE_ON) {
        extpowdebug("Do_Sleep_IOC() entered, but IOC already sleeping, so doing nothing...(mode = %s)\n",fpow_mode_to_str(g_ioc_power_mode));
        return(0);  // Return no error - already sleeping
    }

    // Sleep IOC
    extpowdebug("Do_Sleep_IOC() entered.....\n");

    // At this point, we need to check battery condition.
    // IF battery is lower than low-voltage threshold, disable the warning
    // so the IOC doesn't keep waking us up every second.
    if (g_low_battery_warnings_enabled)  {
#ifndef ALLOW_LOW_BATTERY_TO_WAKE_US
        retVal = IOC_Enable_Battery_Voltage_Low_Reporting(DISABLE, 0);
        if (retVal == 0) {
            g_low_battery_warnings_enabled = 0;
            powdebug("Disabled IOC low-battery warnings\n");
        }
        else
            powdebug("IOC ERROR: Failed to disabled IOC low-battery warnings (err=%d)\n",retVal);
#else
        mvolts = get_battery_mili_voltage();
        powdebug("Leaving low-battery warnings enabled at %d mVolts\n",mvolts);
#endif
    }
    else extpowdebug("g_low_battery_warnings_enabled FALSE coming into Do_Sleep_IOC\n");

    // If we're on an EVT3 AND have power plugged in, don't allow IOC to sleep
    if ((FIONA_BOARD_EVT3 == get_fiona_board_revision()) && \
        ac_adapter_installed())  {
        powdebug("Do_Sleep_IOC(): Can't sleep IOC on EVT3 when AC inserted, so ignoring request.\n");
        return(1);  // Return error - not sleeping
    }
    else {
        IOC_Enable_PowerSaving(ENABLE);
        ioc_set_sleep_flag();
        g_ioc_sleeping = 1;
    }
    extpowdebug("Do_Sleep_IOC() leaving.\n");
    return(0);
}

static int
wasIOCWakeSource(void)
{
    int source = fiona_get_wake_source();
#if 0
    // determine actual IOC wake cause
    IOC_WAKE_CAUSE_REC  ioc_wake_rec;
    memclr(&ioc_wake_rec,sizeof(ioc_wake_rec));
    IOC_Get_Wake_Source(&ioc_wake_rec.wake_source);
#endif
    return(source == IOC_WAKE);
}

static void
Do_Wake_IOC(void)
{
    // If we didn't sleep, don't wake...
    if (0 == g_ioc_allow_sleep_flag) {
        wakelog("Do_Wake_IOC() entered, but IOC not allowed to sleep/wake.\n");
        return;
    }

    if (g_ioc_power_mode == FPOW_MODE_ON) {
        wakelog("Do_Wake_IOC() entered, but IOC already awake - exiting.\n");
        return;
    }

    // Wake IOC
    wakelog("Do_Wake_IOC() entered.....\n");
    if (g_ioc_sleeping)  {
        // First, make sure it's not already awake...
        if (!wasIOCWakeSource())  {
            wakelog("Do_Wake_IOC() WAKING IOC.....\n");
            IOC_Enable_PowerSaving(DISABLE);
            mdelay(100); // Give IOC a bit of wakeup time
        }
        else
            wakelog("Do_Wake_IOC() NOT WAKING IOC, already awake !!\n");
        g_ioc_sleeping = 0;
    }

#ifdef ALLOW_LOW_BATTERY_TO_WAKE_US
    // If low-battery warnings are disabled, enable them now...
    if (0 == g_low_battery_warnings_enabled)  {
        int retVal = 0;
        powdebug("Enabled battery warnings inside Do_Wake_IOC\n");
        retVal = IOC_Enable_Battery_Voltage_Low_Reporting(ENABLE, get_current_lowbatt_mvolt_level());
        if (retVal == 0) {
            g_low_battery_warnings_enabled = 1;
            powdebug("g_low_battery_warnings_enabled set TRUE inside Do_Wake_IOC\n");
        }
        else  {
            printk("IOC:  IOC_Enable_Battery_Voltage_Low_Reporting() failed inside Do_Wake_IOC!!\n");
        }
    }
#endif // ALLOW_LOW_BATTERY_TO_WAKE_US

    wakelog("Do_Wake_IOC() leaving.\n");
}

static int
IOCDriver_Suspend(u32 mode)
{
    if ((mode != FPOW_MODE_OFF) && (mode != FPOW_MODE_SLEEP)) {
        powdebug("IOCDriver_Suspend() - unsupported mode %s\n",fpow_mode_to_str(mode));
        return(1);  // Error, we only support OFF and SLEEP
    }

    // If IOC is currently enabled, put it to sleep...
    extpowdebug("IOCDriver_Suspend(%s)\n",fpow_mode_to_str(mode));
    if (g_ioc_enabled)  {
        // First, get the lock
        lockdebug("Getting g_ioc_hw_lock in IOCDriver_Suspend()...\n");
        if (down_interruptible(&g_ioc_hw_lock))
            return(-ERESTARTSYS);
        lockdebug("Got g_ioc_hw_lock in IOCDriver_Suspend()...\n");

        Do_Sleep_IOC();
        g_ioc_power_mode = FPOW_MODE_OFF;

        up(&g_ioc_hw_lock);
        lockdebug("Released g_ioc_hw_lock leaving IOCDriver_Suspend()...\n");
    }
    return(0);
}

static int
IOCDriver_Resume(u32 mode)
{
    if (mode != FPOW_MODE_ON) {
        powdebug("IOCDriver_Resume(%s) - unsupported mode!\n",fpow_mode_to_str(mode));
        return(1);  // Error, we only support ON
    }

    // First, get the lock
    lockdebug("Getting g_ioc_hw_lock in IOCDriver_Resume()...\n");
    if (down_interruptible(&g_ioc_hw_lock))
        return(-ERESTARTSYS);
    lockdebug("Got g_ioc_hw_lock in IOCDriver_Resume()...\n");

    // Fully reset the UART
    __do_full_init_uart();

    extpowdebug("IOCDriver_Resume(%s): ioc_enabled = %d\n",fpow_mode_to_str(mode),g_ioc_enabled);
    // We actually don't need to do this if g_ioc_wake_state is DISABLE,
    if (g_ioc_enabled) {
        Do_Wake_IOC();

        // test to verify IOC woke up....
        wakelog("IOC:  WOKE IOC, now calling IOC_Echo()\n");
        if (IOC_Echo())
            printk("IOC:  IOC failed to wakeup in IOCDriver_Resume()\n");
        else  {
            g_ioc_power_mode = FPOW_MODE_ON;
        }
    }
    up(&g_ioc_hw_lock);
    lockdebug("Released g_ioc_hw_lock leaving IOCDriver_Resume()...\n");
    return(0);
}


int 
ioc_register_with_fpow(struct device *dev, struct fpow_component **fpow_dev)
{
    struct fpow_registration_rec   reg;
    int rc = 0;

	powdebug("ioc_register_with_fpow(0x%x, 0x%x)\n",(int)dev,(int)fpow_dev);
    memclr(&reg,sizeof(reg));

    strcpy(reg.name,"ioc");
    reg.device_class = FPOW_CLASS_IOC;
    reg.supported_modes = ( FPOW_MODE_ON_SUPPORTED | \
                            FPOW_MODE_SLEEP_SUPPORTED | \
                            FPOW_MODE_OFF_SUPPORTED);
    reg.private = (void*) dev;
    reg.getmode = ioc_fpow_getmode;
    reg.setmode = ioc_fpow_setmode;
    *fpow_dev = fpow_register(&reg);
    if (*fpow_dev == NULL) {
        printk("IOC: Failed to register with fpow !!\n");
        rc = -1;
    }
    return(rc);
}

void
ioc_unregister_with_fpow(struct fpow_component *fpow_dev)
{
	fpow_unregister(fpow_dev);
}

