/*
 * bootdata.c
 *
 * Copyright (C) 2006-2007 Lab126, Inc.  All rights reserved. 
 *
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/proc_fs.h>
#include <asm/uaccess.h>
#include <asm/arch/boot_globals.h>

// Templates for reading/writing numeric boot-global data.
//
typedef unsigned long (*get_boot_data_t)(void);
typedef void (*set_boot_data_t)(unsigned long boot_data);

#define PROC_GET_BOOT_DATA(f) return proc_get_boot_data(page, start, off, count, eof, data, (get_boot_data_t)f)
#define PROC_SET_BOOT_DATA(f) return proc_set_boot_data(file, buf, count, data, (set_boot_data_t)f)

static int
proc_get_boot_data(
    char *page,
    char **start,
    off_t off,
    int count,
    int *eof,
    void *data,
    get_boot_data_t get_boot_data)
{
    int len = sprintf(page, "%08X\n", (unsigned int)(*get_boot_data)());

    *eof = 1;

    return len;
}

static int
proc_set_boot_data(
    struct file *file,
    const char __user *buf,
    unsigned long count,
    void *data,
    set_boot_data_t set_boot_data)
{
    char lbuf[16];

    memset(lbuf, 0, 16);

    if (copy_from_user(lbuf, buf, 8)) {
        return -EFAULT;
    }

    (*set_boot_data)(simple_strtol(lbuf, NULL, 16)); 

    return count;
}

// read/write update_flag
//
static int
proc_update_flag_read(
    char *page,
    char **start,
    off_t off,
    int count,
    int *eof,
    void *data)
{
    PROC_GET_BOOT_DATA(get_update_flag);
}

static int
proc_update_flag_write(
    struct file *file,
    const char __user *buf,
    unsigned long count,
    void *data)
{
    PROC_SET_BOOT_DATA(set_update_flag);
}

// read/write update_data
//
static int
proc_update_data_read(
    char *page,
    char **start,
    off_t off,
    int count,
    int *eof,
    void *data)
{
    PROC_GET_BOOT_DATA(get_update_data);
}

static int
proc_update_data_write(
    struct file *file,
    const char __user *buf,
    unsigned long count,
    void *data)
{
    PROC_SET_BOOT_DATA(set_update_data);
}

// read/write drivemode_screen_ready
//
static int
proc_drivemode_screen_ready_read(
    char *page,
    char **start,
    off_t off,
    int count,
    int *eof,
    void *data)
{
    PROC_GET_BOOT_DATA(get_drivemode_screen_ready);
}

static int
proc_drivemode_screen_ready_write(
    struct file *file,
    const char __user *buf,
    unsigned long count,
    void *data)
{
    PROC_SET_BOOT_DATA(set_drivemode_screen_ready);
}

// read/write framework_started, framework_running
//
static int
proc_framework_started_read(
    char *page,
    char **start,
    off_t off,
    int count,
    int *eof,
    void *data)
{
    PROC_GET_BOOT_DATA(get_framework_started);
}

static int
proc_framework_started_write(
    struct file *file,
    const char __user *buf,
    unsigned long count,
    void *data)
{
    PROC_SET_BOOT_DATA(set_framework_started);
}

static int
proc_framework_running_read(
    char *page,
    char **start,
    off_t off,
    int count,
    int *eof,
    void *data)
{
    PROC_GET_BOOT_DATA(get_framework_running);
}

static int
proc_framework_running_write(
    struct file *file,
    const char __user *buf,
    unsigned long count,
    void *data)
{
    PROC_SET_BOOT_DATA(set_framework_running);
}

// read/write env_script
//
static int
proc_env_script_read(
    char *page,
    char **start,
    off_t off,
    int count,
    int *eof,
    void *data)
{
    int len = sprintf(page, "%s", get_env_script());

    *eof = 1;

    return len;
}

static int
proc_env_script_write(
    struct file *file,
    const char __user *buf,
    unsigned long count,
    void *data)
{
    char lbuf[ENV_SCRIPT_SIZE];
    
    if (count > ENV_SCRIPT_SIZE) {
        return -EINVAL;
    }

    memset(lbuf, 0, count);

    if (copy_from_user(lbuf, buf, count)) {
        return -EFAULT;
    }

    lbuf[count] = '\0';
    set_env_script(lbuf);

    return count;
}

// read/write dirty_boot_flag
//
static int
proc_dirty_boot_read(
    char *page,
    char **start,
    off_t off,
    int count,
    int *eof,
    void *data)
{
    PROC_GET_BOOT_DATA(get_dirty_boot_flag);
}

static int
proc_dirty_boot_write(
    struct file *file,
    const char __user *buf,
    unsigned long count,
    void *data)
{
    PROC_SET_BOOT_DATA(set_dirty_boot_flag);
}

// read calibration_info
//
static int
proc_calibration_read(
    char *page,
    char **start,
    off_t off,
    int count,
    int *eof,
    void *data)
{
    int len = sprintf(page, "%d, %ld, %ld\n", get_op_amp_offset(), get_gain_value(), get_rve_value());

    *eof = 1;

    return len;
}

// Entry/Exit
//
#define BD_PROC_MODE_PARENT    (S_IFDIR | S_IRUGO | S_IXUGO)
#define BD_PROC_MODE_CHILD_RW  (S_IWUGO | S_IRUGO)
#define BD_PROC_MODE_CHILD_R   (S_IRUGO)

#define BD_PROC_PARENT         "bd"
#define BD_PROC_UPDATE_FLAG    "update_flag"
#define BD_PROC_UPDATE_DATA    "update_data"
#define BD_PROC_DRIVEMODE      "drivemode_screen_ready"
#define BD_PROC_FRAMEWORK1     "framework_started"
#define BD_PROC_FRAMEWORK2     "framework_running"
#define BD_PROC_ENV_SCRIPT      "env_script"
#define BD_PROC_DIRTY_BOOT     "dirty_boot_flag"
#define BD_PROC_CALIBRATION    "calibration_info"

static struct proc_dir_entry *proc_bd_parent    = NULL;
static struct proc_dir_entry *proc_update_flag  = NULL;
static struct proc_dir_entry *proc_update_data  = NULL;
static struct proc_dir_entry *proc_drivemode    = NULL;
static struct proc_dir_entry *proc_framework1   = NULL;
static struct proc_dir_entry *proc_framework2   = NULL;
static struct proc_dir_entry *proc_env_script   = NULL;
static struct proc_dir_entry *proc_dirty_boot   = NULL;
static struct proc_dir_entry *proc_calibration  = NULL;

static struct proc_dir_entry *create_bd_proc_entry(const char *name, mode_t mode, struct proc_dir_entry *parent,
    read_proc_t *read_proc, write_proc_t *write_proc)
{
    struct proc_dir_entry *bd_proc_entry = create_proc_entry(name, mode, parent);
    
    if (bd_proc_entry) {
        bd_proc_entry->owner      = THIS_MODULE;
        bd_proc_entry->data       = NULL;
        bd_proc_entry->read_proc  = read_proc;
        bd_proc_entry->write_proc = write_proc;
    }
    
    return bd_proc_entry;
}

#define remove_bd_proc_entry(name, entry, parent)   \
    do                                              \
    if ( entry )                                    \
    {                                               \
        remove_proc_entry(name, parent);            \
        entry = NULL;                               \
    }                                               \
    while ( 0 )

static void
bootdata_cleanup(
	void)
{
	if ( proc_bd_parent )
	{
	    remove_bd_proc_entry(BD_PROC_UPDATE_FLAG,   proc_update_flag,   proc_bd_parent);
	    remove_bd_proc_entry(BD_PROC_UPDATE_DATA,   proc_update_data,   proc_bd_parent);
	    remove_bd_proc_entry(BD_PROC_DRIVEMODE,     proc_drivemode,     proc_bd_parent);
	    remove_bd_proc_entry(BD_PROC_FRAMEWORK1,    proc_framework1,    proc_bd_parent);
	    remove_bd_proc_entry(BD_PROC_FRAMEWORK2,    proc_framework2,    proc_bd_parent);
	    remove_bd_proc_entry(BD_PROC_ENV_SCRIPT,    proc_env_script,    proc_bd_parent);
	    remove_bd_proc_entry(BD_PROC_DIRTY_BOOT,    proc_dirty_boot,    proc_bd_parent);
	    remove_bd_proc_entry(BD_PROC_CALIBRATION,   proc_calibration,   proc_bd_parent);
	    remove_bd_proc_entry(BD_PROC_PARENT,        proc_bd_parent,     NULL);
	}
}

static int __init
bootdata_init(
	void)
{
    int result = -ENOMEM;

    // Parent:  /proc/bd.
    //
    proc_bd_parent = create_proc_entry(BD_PROC_PARENT, BD_PROC_MODE_PARENT, NULL);
    
    if (proc_bd_parent)
    {
        int null_check = -1;
        
        // Child:   /proc/bd/update_flag.
        //
        proc_update_flag = create_bd_proc_entry(BD_PROC_UPDATE_FLAG, BD_PROC_MODE_CHILD_RW, proc_bd_parent,
            proc_update_flag_read, proc_update_flag_write);
        
        null_check &= (int)proc_update_flag;

        // Child:   /proc/bd/update_data.
        //
        proc_update_data = create_bd_proc_entry(BD_PROC_UPDATE_DATA, BD_PROC_MODE_CHILD_RW, proc_bd_parent,
            proc_update_data_read, proc_update_data_write);
            
        null_check &= (int)proc_update_data;

        // Child:   /proc/bd/drivemode_screen_ready.
        //
        proc_drivemode = create_bd_proc_entry(BD_PROC_DRIVEMODE, BD_PROC_MODE_CHILD_RW, proc_bd_parent,
            proc_drivemode_screen_ready_read, proc_drivemode_screen_ready_write);
            
        null_check &= (int)proc_drivemode;
        
        // Child:   /proc/bd/framework_started.
        //
        proc_framework1 = create_bd_proc_entry(BD_PROC_FRAMEWORK1, BD_PROC_MODE_CHILD_RW, proc_bd_parent,
            proc_framework_started_read, proc_framework_started_write);
            
        null_check &= (int)proc_framework1;
        
        // Child:   /proc/bd/framework_running.
        //
        proc_framework2 = create_bd_proc_entry(BD_PROC_FRAMEWORK2, BD_PROC_MODE_CHILD_RW, proc_bd_parent,
            proc_framework_running_read, proc_framework_running_write);
            
        null_check &= (int)proc_framework2;
        
        // Child:   /proc/bd/env_script.
        //
        proc_env_script = create_bd_proc_entry(BD_PROC_ENV_SCRIPT, BD_PROC_MODE_CHILD_RW, proc_bd_parent,
            proc_env_script_read, proc_env_script_write);
            
        // Child:   /proc/bd/dirty_boot_flag.
        //
        proc_dirty_boot = create_bd_proc_entry(BD_PROC_DIRTY_BOOT, BD_PROC_MODE_CHILD_RW, proc_bd_parent,
            proc_dirty_boot_read, proc_dirty_boot_write);

        // Child:   /proc/bd/calibration_info.
        //
        proc_calibration = create_bd_proc_entry(BD_PROC_CALIBRATION, BD_PROC_MODE_CHILD_R, proc_bd_parent,
            proc_calibration_read, NULL);
        
        null_check &= (int)proc_calibration;
        
        // Success? 
        //
        if ( 0 == null_check )
            bootdata_cleanup();
        else
            result = 0;
    }

    // Done.
    //
	return ( result );
}

static void __exit
bootdata_exit(
	void)
{
    bootdata_cleanup();
}

module_init(bootdata_init);
module_exit(bootdata_exit);

MODULE_DESCRIPTION("bootdata driver");
MODULE_AUTHOR("Lab126");
MODULE_LICENSE("Proprietary");


