/*
 *  linux/drivers/mtd/onenand/fiona-onenand.c
 *
 *  Copyright (c) 2005 Samsung Electronics
 *  Kyungmin Park <kyungmin.park@samsung.com>
 *
 *  Derived from linux/drivers/mtd/nand/fiona-nand-flash.c
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 *  Overview:
 *   This is a device driver for the OneNAND flash device for FIONA boards.
 */

#include <linux/autoconf.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/onenand.h>
#include <linux/mtd/partitions.h>
#include <linux/mtd/concat.h>
#include <asm/arch/fiona.h>

#include <asm/io.h>
#include <asm/arch/hardware.h>
#include <asm/arch/pxa-regs.h>
#include <asm/sizes.h>

#define FIONA_ONENAND_FLASH_START1	PXA_CS2_PHYS
#define FIONA_ONENAND_FLASH_START1_ALT	PXA_CS0_PHYS
#define FIONA_ONENAND_FLASH_START2	PXA_CS1_PHYS

#define FIONA_ONENAND_SUPPORT_EVT1      0

extern int get_fiona_board_revision(void); /* In fiona.c, eliminates warning. */

/*
 * MTD structure for FIONA board
 */
static struct mtd_info *fiona_onenand_mtd  = NULL;
static struct mtd_info *fiona_onenand_mtd1 = NULL;

#if FIONA_ONENAND_SUPPORT_EVT1
static struct mtd_info *fiona_onenand_mtd2 = NULL;
#endif

/*
 * Define partitions for flash devices
 */
#ifdef CONFIG_MTD_PARTITIONS
static struct mtd_partition static_partition_evt1[] = {
	{
		.name		= "flash",		// [0] Full Device
		.offset		= 0,
		.size		= 0,
		.mask_flags	= MTD_CAP_ROM
 	},
	
	{
		.name		= "boot",		// [1] Boot Area
		.offset		= 0,
		.size		= SZ_4M,
		.mask_flags	= MTD_WRITEABLE
 	},
	
	{
		.name		= "system",		// [2] Filesystem
		.offset		= MTDPART_OFS_APPEND,
		.size		= SZ_16M
	},
	
	{
		.name		= "recovery",		// [3] Recovery
		.offset		= MTDPART_OFS_APPEND,
		.size		= SZ_16M
	},
	
	{
		.name		= "user",		// [4] Userstore
		.offset		= MTDPART_OFS_APPEND,
		.size		= (SZ_1M * 86)
	},
	
	{
		.name		= "reserved",		// [5] Reserved
		.offset		= MTDPART_OFS_APPEND,
		.size		= (SZ_1M * 6)
	}
};

static struct mtd_partition static_partition[] = {
	{
		.name		= "flash",		// [0] Full Device
		.offset		= 0,
		.size		= 0,
		.mask_flags	= MTD_CAP_ROM
 	},
	
	{
		.name		= "boot",		// [1] Boot Area
		.offset		= 0,
		.size		= SZ_4M,
		.mask_flags	= MTD_WRITEABLE
 	},
	
	{
		.name		= "system",		// [2] Filesystem
		.offset		= MTDPART_OFS_APPEND,
		.size		= SZ_16M
	},
	
	{
		.name		= "recovery",		// [3] Recovery
		.offset		= MTDPART_OFS_APPEND,
		.size		= SZ_16M
	},
	
	{
		.name		= "user",		// [4] Userstore
		.offset		= MTDPART_OFS_APPEND,
		.size		= (SZ_1M * 214)
	},
	
	{
		.name		= "reserved",		// [5] Reserved
		.offset		= MTDPART_OFS_APPEND,
		.size		= (SZ_1M * 6)
	}
};

const char *part_probes[] = {  NULL  };

#endif

/* Scan to find existance of the device at base.
   This also allocates oob and data internal buffers */
static char onenand_name1[] = "onenand1";
#if FIONA_ONENAND_SUPPORT_EVT1
static char onenand_name2[] = "onenand2";
static char onenand_name[]  = "onenand";
#endif

/*
 * Main initialization routine
 */
static int __init fiona_onenand_init (void)
{
	struct onenand_chip *this1;
#if FIONA_ONENAND_SUPPORT_EVT1
	struct onenand_chip *this2;
        struct mtd_info *subdev[2];
#endif
	struct mtd_partition *dynamic_partition = 0;
	int err = 0, fiona_board_revision = get_fiona_board_revision();

    /* Make sure we're on new enough hardware.
     * Don't load on a board older than EVT1
     */
    if ((fiona_board_revision == FIONA_BOARD_P3_5) || \
        (fiona_board_revision < FIONA_BOARD_EVT1)) {
		printk (KERN_WARNING "Aborting OneNAND intialization.  Board revision is too old.\n");
		err = -EIO;
		goto out;
	}
        
        
	/* Allocate memory for MTD device structure and private data */
	fiona_onenand_mtd1 = kmalloc (sizeof(struct mtd_info) + sizeof (struct onenand_chip),
				GFP_KERNEL);
	if (!fiona_onenand_mtd1) {
		printk (KERN_WARNING "Unable to allocate OneNAND MTD device structure.\n");
		err = -ENOMEM;
		goto out;
	}

#if FIONA_ONENAND_SUPPORT_EVT1
	fiona_onenand_mtd2 = kmalloc (sizeof(struct mtd_info) + sizeof (struct onenand_chip),
				GFP_KERNEL);
	if (!fiona_onenand_mtd2) {
		printk (KERN_WARNING "Unable to allocate OneNAND MTD device structure.\n");
		err = -ENOMEM;
		kfree(fiona_onenand_mtd1);
		goto out;
	}
#endif

	/* Get pointer to private data */
	this1 = (struct onenand_chip *) (&fiona_onenand_mtd1[1]);
#if FIONA_ONENAND_SUPPORT_EVT1
	this2 = (struct onenand_chip *) (&fiona_onenand_mtd2[1]);
#endif

	/* Initialize structures */
	memset((char *) fiona_onenand_mtd1, 0, sizeof(struct mtd_info) + sizeof(struct onenand_chip));
#if FIONA_ONENAND_SUPPORT_EVT1
	memset((char *) fiona_onenand_mtd2, 0, sizeof(struct mtd_info) + sizeof(struct onenand_chip));
#endif

	/* Link the private data with the MTD structure */
	fiona_onenand_mtd1->priv = this1;
#if FIONA_ONENAND_SUPPORT_EVT1
	fiona_onenand_mtd2->priv = this2;
#endif

#if 0 && !defined(OBSOLETE)
	configure_bus_cs2();
	configure_bus_cs1();
#endif

	fiona_onenand_mtd1->name = onenand_name1;
#if FIONA_ONENAND_SUPPORT_EVT1
	fiona_onenand_mtd2->name = onenand_name2;
#endif

	this1->base = ioremap_nocache(FIONA_ONENAND_FLASH_START1, SZ_128K);
	if (!this1->base) {
		printk (KERN_WARNING "Unabled to ioremap 1st OneNAND device at 0x%08X.\n",
			FIONA_ONENAND_FLASH_START1);
	}

	if (onenand_scan(fiona_onenand_mtd1, 1)) {
	  //printk(KERN_NOTICE "onenand_scan failed for 1st OneNAND device.\n");
	  iounmap(this1->base);

	  /* Check for OneNAND at 0.  (Normal boot from OneNAND case.)	*/
	  
	  this1->base = ioremap_nocache(FIONA_ONENAND_FLASH_START1_ALT, SZ_128K);
	  if (!this1->base) {
	      printk(KERN_WARNING "Unable to ioremap 1st OneNAND device at 0x%08X.\n",
	      		FIONA_ONENAND_FLASH_START1_ALT);
	      err = -ENXIO;
	      goto out_mtd;
	  }

	  if (onenand_scan(fiona_onenand_mtd1, 1)) {
	      //printk(KERN_NOTICE "onenand_scan failed for 1st OneNAND device at 0.\n");
	      iounmap(this1->base);
	      err = -ENXIO;
	      goto out_mtd;
	  }
	}

#if FIONA_ONENAND_SUPPORT_EVT1
	this2->base = ioremap_nocache(FIONA_ONENAND_FLASH_START2, SZ_128K);
	if (!this2->base) {
		printk (KERN_WARNING "Unabled to ioremap 2nd OneNAND device at 0x%08X.\n",
			FIONA_ONENAND_FLASH_START2);
	}

	if (onenand_scan(fiona_onenand_mtd2, 1)) {
	  //printk(KERN_NOTICE "onenand_scan failed for 2nd OneNAND device.\n");
	  iounmap(this1->base);
	  iounmap(this2->base);
	  err = -ENXIO;
	  goto out_mtd;
	}
#endif

#if 0
	if (onenand_scan(fiona_onenand_mtd, 1)){
		/* try the second address */
		iounmap(this->base);
		this->base = ioremap_nocache(FIONA_ONENAND_FLASH_START2, SZ_128K);
		if (onenand_scan(fiona_onenand_mtd, 1)) {
			iounmap(this->base);
                        err = -ENXIO;
                        goto out_mtd;
		}
	}
#endif

#if FIONA_ONENAND_SUPPORT_EVT1
	subdev[0] = fiona_onenand_mtd1;
	subdev[1] = fiona_onenand_mtd2;

	fiona_onenand_mtd = mtd_concat_create(subdev, 2, onenand_name);

	if(!fiona_onenand_mtd) {
	  err = -ENXIO;
	  goto out_con;
	}
#endif

	/* Register the partitions */

        if (NULL == fiona_onenand_mtd)
                fiona_onenand_mtd = fiona_onenand_mtd1;

	switch (fiona_onenand_mtd->size) {
	case SZ_256M:
	case SZ_128M:
	case SZ_64M:
	case SZ_32M:
#ifdef CONFIG_MTD_PARTITIONS
		err = parse_mtd_partitions(fiona_onenand_mtd, part_probes,
					&dynamic_partition, 0);
		if (err > 0) {
			printk(KERN_NOTICE "Trying dynamic partition map...\n");
			err = add_mtd_partitions(fiona_onenand_mtd,
					dynamic_partition, err);
		}
		else if (1) {
			printk(KERN_NOTICE "Trying static partition map...\n");
			
			if (FIONA_BOARD_EVT1 == fiona_board_revision)
				err = add_mtd_partitions(fiona_onenand_mtd,
						static_partition_evt1,
						ARRAY_SIZE(static_partition_evt1));
			else
				err = add_mtd_partitions(fiona_onenand_mtd,
						static_partition,
						ARRAY_SIZE(static_partition));
		}
		else
#endif
			err = add_mtd_device(fiona_onenand_mtd);
		
		if (err)
			goto out_buf;
		break;

	default:
		printk(KERN_WARNING "Unsupported OneNAND device\n");
		err = -ENXIO;
		goto out_buf;
	}

	return 0;




out_buf:
#if FIONA_ONENAND_SUPPORT_EVT1
	onenand_release(fiona_onenand_mtd);
	mtd_concat_destroy(fiona_onenand_mtd);
out_con:
#endif
	onenand_release(fiona_onenand_mtd1);
#if FIONA_ONENAND_SUPPORT_EVT1
	onenand_release(fiona_onenand_mtd2);
#endif
	iounmap(this1->base);
#if FIONA_ONENAND_SUPPORT_EVT1
	iounmap(this2->base);
#endif
out_mtd:
	kfree(fiona_onenand_mtd1);
#if FIONA_ONENAND_SUPPORT_EVT1
	kfree(fiona_onenand_mtd2);
#endif

out:
	return err;
}

/*
 * Clean up routine
 */
static void __exit fiona_onenand_cleanup (void)
{
#if FIONA_ONENAND_SUPPORT_EVT1
	struct onenand_chip *this1, *this2;

	this1 = fiona_onenand_mtd1->priv;
	this2 = fiona_onenand_mtd2->priv;

	/* onenand_release frees MTD partitions, MTD structure
	   and onenand internal buffers */
	onenand_release(fiona_onenand_mtd);
	onenand_release(fiona_onenand_mtd1);
	onenand_release(fiona_onenand_mtd2);

	iounmap(this1->base);
	iounmap(this2->base);

	mtd_concat_destroy(fiona_onenand_mtd);
	kfree(fiona_onenand_mtd1);
	kfree(fiona_onenand_mtd2);
#else	
	struct onenand_chip *this1;

	this1 = fiona_onenand_mtd1->priv;

	/* onenand_release frees MTD partitions, MTD structure
	   and onenand internal buffers */
	onenand_release(fiona_onenand_mtd1);

	iounmap(this1->base);

	kfree(fiona_onenand_mtd1);
#endif
}

module_init(fiona_onenand_init);
module_exit(fiona_onenand_cleanup);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Kyungmin Park <kyungmin.park@samsung.com>");
MODULE_DESCRIPTION("Glue layer for OneNAND flash on FIONA boards");
