/*
 * JFFS2 -- Journalling Flash File System, Version 2.
 *
 * Copyright (C) 2005  Zoltan Sogor <weth@inf.u-szeged.hu>,
 *                     Ferenc Havasi <havasi@inf.u-szeged.hu>,
 *                     Patrik Kluba <pajko@halom.u-szeged.hu>,
 *                     University of Szeged, Hungary
 *
 *          Thanks for Peter Grayson <pgrayson@realmsys.com>
 *
 * For licensing information, see the file 'LICENCE' in this directory.
 *
 * $Id$
 *
 */

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/mtd/mtd.h>
#include <linux/pagemap.h>
#include <linux/crc32.h>
#include <linux/compiler.h>
#include <linux/vmalloc.h>
#include <linux/parser.h>
#include "nodelist.h"

struct erase_priv_struct {
	struct jffs2_eraseblock *jeb;
	struct jffs2_sb_info *c;
};

static int jffs2_cs_erase_block(struct jffs2_sb_info *c, struct jffs2_eraseblock *jeb);
static int jffs2_cs_mark_erased_block(struct jffs2_sb_info *c, struct jffs2_eraseblock *jeb);
static int jffs2_cs_write_inocache(struct jffs2_sb_info *c, struct jffs2_cs_inocache *buff,
										uint32_t cachelen, uint32_t count, uint32_t offset);
static int jffs2_cs_write_jebinfo(struct jffs2_sb_info *c, struct jffs2_cs_jeb *buff,
									uint32_t jebinfolen, uint32_t num, uint32_t offset);
static int jffs2_cs_write_rawref(struct jffs2_sb_info *c, struct jffs2_cs_rawref *buff,
									uint32_t rawreflen, uint32_t num, uint32_t offset);
static int jffs2_cs_write_lists(struct jffs2_sb_info *c, jint16_t *buff,
									uint32_t listslen, uint32_t offset);
static int jffs2_cs_add_offset(struct jffs2_sb_info *c, uint32_t offset, uint32_t length);
static void jffs2_cs_clean_offset_list(struct jffs2_sb_info *c);
static int jffs2_cs_read_log_jeb(struct jffs2_sb_info *c);
static int jffs2_cs_process_log(struct jffs2_sb_info *c, struct jffs2_raw_cs_log *log);
static int jffs2_cs_process_node(struct jffs2_sb_info *c, union jffs2_node_union *node, uint32_t offset);
static int jffs2_cs_process_inocache_node(struct jffs2_sb_info *c, struct jffs2_raw_cs_inocache *node);
static int jffs2_cs_process_jebinfo_node(struct jffs2_sb_info *c, struct jffs2_raw_cs_jebinfo *node);
static int jffs2_cs_process_rawref_node(struct jffs2_sb_info *c, struct jffs2_raw_cs_rawref *node);
static int jffs2_cs_process_lists_node(struct jffs2_sb_info *c, struct jffs2_raw_cs_lists *node);
static int jffs2_cs_correct_jebinfo(struct jffs2_sb_info *c);
static void *jffs2_cs_alloc_flashbuf(struct jffs2_sb_info *c);
static void jffs2_cs_free_flashbuf(struct jffs2_sb_info *c, void *flashbuf);

#ifdef CONFIG_JFFS2_CS_FEB
static int jffs2_cs_write_umount_log_jeb(struct jffs2_sb_info *c);
#endif

#ifdef CONFIG_JFFS2_CS_MP_SYSFS
static int jffs2_cs_parse_mount_params(struct jffs2_sb_info *c, char *options);
static int jffs2_cs_write_umount_log_mntsys(struct jffs2_sb_info *c);
#endif

void jffs2_free_all_node_refs(struct jffs2_sb_info *c, struct jffs2_eraseblock *jeb);

// move all data from jeb to nextblock, and erase jeb
int jffs2_cs_move_block(struct jffs2_sb_info *c, struct jffs2_eraseblock *jeb)
{
	int ret;

	if (c->gcblock) {
		dbg_cs("error GC is in use!\n");
		return -1;
	}

	c->gcblock = jeb;
	list_del(&c->gcblock->list);

	c->gcblock->gc_node = c->gcblock->first_node;

	/* Have we accidentally picked a clean block with wasted space ? */
	if (c->gcblock->wasted_size) {
		dbg_cs("Converting wasted_size %08x to dirty_size\n",c->gcblock->wasted_size);
		c->gcblock->dirty_size += c->gcblock->wasted_size;
		c->wasted_size -= c->gcblock->wasted_size;
		c->dirty_size += c->gcblock->wasted_size;
		c->gcblock->wasted_size = 0;
	}

	if (c->gcblock->gc_node) {

		while(c->gcblock) {
			ret = jffs2_garbage_collect_pass(c);
			if (ret) {
				printk("%s: Error occured during gc!\n", __FUNCTION__);
				/* TODO ERROR HANDLING*/
				return -1;
			}
		}
	} else {
		JFFS2_WARNING("Eep. c->gcblock->gc_node for block at 0x%08x is NULL\n", c->gcblock->offset);
	}

	c->erasing_size += c->sector_size;
	c->wasted_size -= jeb->wasted_size;
	c->free_size -= jeb->free_size;
	c->used_size -= jeb->used_size;
	c->dirty_size -= jeb->dirty_size;
	jeb->wasted_size = jeb->used_size = jeb->dirty_size = jeb->free_size = 0;
	jffs2_free_all_node_refs(c, jeb);

	ret = jffs2_cs_erase_block(c, jeb);

	if (ret) {
		c->erasing_size -= c->sector_size;
		c->bad_size += c->sector_size;
		list_add(&jeb->list, &c->bad_list);

		return -1;
	}

	c->erasing_size -= c->sector_size;

	if (jffs2_cs_mark_erased_block(c, jeb)) {
		JFFS2_ERROR("Error during cs_mark_erase_block\n");
		// TODO : we should correct sb size info !!!
		// maybe :
		c->bad_size += c->sector_size;
		list_add(&jeb->list, &c->bad_list);

		return -1;
	}

	// adding cenztralized summary log block size as used_space
	c->used_size += c->sector_size;

	// for debugging
	//D1(jffs2_dump_block_lists(c));

	return 0;
}

// init centralized summary structures
int jffs2_cs_init(struct jffs2_sb_info *c, void *data)
{

	c->cs_struct = kmalloc(sizeof(struct cent_sum), GFP_KERNEL);
	if(!c->cs_struct)
		return -ENOMEM;

	c->cs_struct->csblock = NULL;
	c->cs_struct->csblock_free = 0;
	c->cs_struct->cs_log_offset = 0;
	c->cs_struct->cs_flags &= ~JFFS2_CS_CSBLOCK_VALID; // flag valid = 0
	c->cs_struct->cs_flags &= ~JFFS2_CS_DURING_MOUNT; // flag during_mount= 0
	c->cs_struct->cs_flags &= ~JFFS2_CS_DURING_UNMOUNT; // flag during_unmount= 0
	c->cs_struct->cs_flags &= ~JFFS2_CS_MOUNT_OFFSET_PARAM; // flag mount_offset_param = 0

	c->cs_struct->offset_list_tail = NULL;
	c->cs_struct->offset_list_head = NULL;
	c->cs_struct->offset_list_num = 0;

	if (jffs2_cs_mode() == JFFS2_CS_MODE_MP_SYSFS) {
		if(!jffs2_cs_parse_mount_params(c, data)) {
			JFFS2_ERROR("Error at mount options parsing!\n");
			return -EINVAL;
		}
	}

	return 0;
}

// init centralized summary log block
int jffs2_cs_init_cs_block_jeb(struct jffs2_sb_info *c)
{
	int ret;

	c->cs_struct->cs_flags |= JFFS2_CS_DURING_MOUNT; // flag during_mount = 1

	ret = jffs2_cs_move_block(c, c->cs_struct->csblock);

	if (!ret) {
		c->cs_struct->csblock_free = c->sector_size;
		c->cs_struct->cs_flags |= JFFS2_CS_CSBLOCK_VALID; // flag valid = 1;
		c->cs_struct->cs_flags &= ~JFFS2_CS_DURING_MOUNT; // flag during_mount = 0

		if (jffs2_cs_write_mount_log(c)) {
			dbg_cs("write_mount_log returned with error\n");

			// disable cent_sum
			c->cs_struct->csblock = NULL;
			c->cs_struct->cs_flags &= ~JFFS2_CS_CSBLOCK_VALID; // flag valid = 0
			c->cs_struct->csblock_free = 0;

			return -1;
		}

		return 0;
	}

	dbg_cs("Can't empty erase block for centralized summary log\n");

	c->cs_struct->csblock = NULL;
	c->cs_struct->cs_flags &= ~JFFS2_CS_CSBLOCK_VALID; // flag valid = 0
	c->cs_struct->cs_flags &= ~JFFS2_CS_DURING_MOUNT; // flag during_mount = 0

	return -1;
}


// erase the jeb
static int jffs2_cs_erase_block(struct jffs2_sb_info *c, struct jffs2_eraseblock *jeb)
{
	int ret;
	uint32_t bad_offset;
	struct erase_info *instr;

	dbg_cs("erase block %#x (range %#x-%#x)\n", jeb->offset, jeb->offset,
				jeb->offset + c->sector_size);
	instr = kmalloc(sizeof(struct erase_info) + sizeof(struct erase_priv_struct), GFP_KERNEL);
	if (!instr) {
		JFFS2_ERROR("Can't allocate memory for erase_info!\n");
		return -ENOMEM;
	}

	memset(instr, 0, sizeof(*instr));

	instr->mtd = c->mtd;
	instr->addr = jeb->offset;
	instr->len = c->sector_size;
	instr->callback = NULL;
	instr->priv = (unsigned long)(&instr[1]);
	instr->fail_addr = 0xffffffff;

	((struct erase_priv_struct *)instr->priv)->jeb = jeb;
	((struct erase_priv_struct *)instr->priv)->c = c;

	ret = c->mtd->erase(c->mtd, instr);

	if (!ret) {
		c->cs_struct->csblock_free = c->sector_size;
		return 0;
	}

	bad_offset = instr->fail_addr;
	kfree(instr);

	if (!c->mtd->block_markbad)
		return 1; // What else can we do?

	dbg_cs("Marking bad block at %08x\n", bad_offset);
	ret = c->mtd->block_markbad(c->mtd, bad_offset);

	if (ret) {
		dbg_cs("Write failed for block at %08x: error %d\n", jeb->offset, ret);
		return ret;
	}

	return 1;
}


// mark block
static int jffs2_cs_mark_erased_block(struct jffs2_sb_info *c, struct jffs2_eraseblock *jeb)
{
	unsigned char *ebuf;
	size_t retlen;
	int ret;
	uint32_t bad_offset;

	ebuf = kmalloc(PAGE_SIZE, GFP_KERNEL);
	if (!ebuf) {
		JFFS2_WARNING ("Failed to allocate page buffer for verifying "
				"erase at 0x%08x. Assuming it worked\n", jeb->offset);
	} else {
		uint32_t ofs = jeb->offset;

		dbg_cs("Verifying erase at 0x%08x\n", jeb->offset);
		while(ofs < jeb->offset + c->sector_size) {
			uint32_t readlen = min((uint32_t)PAGE_SIZE, jeb->offset + c->sector_size - ofs);
			int i;

			bad_offset = ofs;

			ret = jffs2_flash_read(c, ofs, readlen, &retlen, ebuf);
			if (ret) {
				JFFS2_WARNING("Read of newly-erased block at 0x%08x "
						"failed: %d. Putting on bad_list\n", ofs, ret);
				goto bad;
			}
			if (retlen != readlen) {
				JFFS2_WARNING("Short read from newly-erased block at"
						"0x%08x. Wanted %d, got %zd\n", ofs, readlen, retlen);
				goto bad;
			}
			for (i=0; i<readlen; i += sizeof(unsigned long)) {
				/* It's OK. We know it's properly aligned */
				unsigned long datum = *(unsigned long *)(&ebuf[i]);
				if (datum + 1) {
					bad_offset += i;
					JFFS2_WARNING("Newly-erased block contained word "
							"0x%lx at offset 0x%08x\n", datum, bad_offset);
				bad:
					kfree(ebuf);
				bad2:
					// mark block as bad
					if (!c->mtd->block_markbad)
						return 1; // What else can we do?

					dbg_cs("Marking bad block at %08x\n", bad_offset);
					ret = c->mtd->block_markbad(c->mtd, bad_offset);

					if (ret) {
						dbg_cs("Write failed for block at "
								"%08x: error %d\n", jeb->offset, ret);
						return ret;
					}

					return -1;
				}
			}
			ofs += readlen;
		}
		kfree(ebuf);
	}

	bad_offset = jeb->offset;

	/* Write the erase complete marker */
	dbg_cs("Writing erased marker to block at 0x%08x\n", jeb->offset);

	if (!EBFLAGS_HAS_EBH(jeb)) {
		printk("EBH header not found!!!");
		if (c->nr_blocks_with_ebh != 0) {
			jeb->erase_count = c->total_erase_count/c->nr_blocks_with_ebh;
		} else {
			jeb->erase_count = 0;
		}
		EBFLAGS_SET_EBH(jeb);
	}

	jeb->erase_count++;
	dbg_cs("CS block erase count : %d\n", jeb->erase_count);

	if (jffs2_write_nand_ebh(c, jeb))
		goto bad2;

	jeb->first_node = jeb->last_node = NULL;

	jeb->free_size = c->sector_size;
	jeb->used_size = 0;
	jeb->dirty_size = 0;
	jeb->wasted_size = 0;

	return 0;
}


// free cs_struct->offset_list
void jffs2_cs_free_offset_list(struct jffs2_sb_info *c)
{
	struct cent_sum_offset *temp;

	while (c->cs_struct->offset_list_head != NULL) {
		temp = c->cs_struct->offset_list_head;
		c->cs_struct->offset_list_head = temp->next;
		kfree(temp);
		c->cs_struct->offset_list_num--;
	}

	c->cs_struct->offset_list_tail = NULL;

	if (c->cs_struct->offset_list_num) {
		dbg_cs("List is empty, but list_num != 0\n");
		BUG();
	}
}

// free cs_struct
void jffs2_cs_free(struct jffs2_sb_info *c)
{
	if (c->cs_struct->offset_list_head != NULL) {
		jffs2_cs_free_offset_list(c);
	}

	if (c->cs_struct)
		kfree(c->cs_struct);

}

// free c->lists (clean, free, dirty, ...)
void jffs2_cs_free_lists(struct jffs2_sb_info *c)
{
	int i;

	for(i=0; i<c->nr_blocks; i++) {
		list_del_init(&c->blocks[i]->list);
	}

}

static inline struct jffs2_inode_cache *
first_inode_chain(int *i, struct jffs2_sb_info *c)
{
	for (; *i < INOCACHE_HASHSIZE; (*i)++) {
		if (c->inocache_list[*i])
			return c->inocache_list[*i];
	}
	return NULL;
}


static inline struct jffs2_inode_cache *
next_inode(int *i, struct jffs2_inode_cache *ic, struct jffs2_sb_info *c)
{
	/* More in this chain? */
	if (ic->next)
		return ic->next;
	(*i)++;
	return first_inode_chain(i, c);
}


#define for_each_inode(i, c, ic)	\
	for (i = 0, ic = first_inode_chain(&i, (c));	\
		ic;		\
		ic = next_inode(&i, ic, (c)))

// count list
static int count_list(struct list_head *l)
{
	uint32_t count = 0;
	struct list_head *tmp;

	list_for_each(tmp, l) {
		count++;
	}
	return count;
}

// count the number of rawref entries
static unsigned int jffs2_cs_count_rawrefs(struct jffs2_sb_info *c)
{
	int i;
	unsigned int count = 0;
	struct jffs2_inode_cache *ic;
	struct jffs2_raw_node_ref *raw;

	for_each_inode(i, c, ic) {
		raw = ic->nodes;
		while (raw) {
			raw = raw->next_in_ino;
			count++;
		}
	}

	return count;
}

struct jffs2_cent_sum_tmp_raws {
	struct jffs2_raw_node_ref *raw;
	struct jffs2_raw_node_ref *nii;
};

// save rawrefs next_in_ino field
int jffs2_cs_save_rawrefs_nii(struct jffs2_sb_info *c)
{
	unsigned int size, count;
	int i;
	void *ptr;
	struct jffs2_inode_cache *ic;
	struct jffs2_raw_node_ref *raw;
	struct jffs2_cent_sum_tmp_raws *rn;

	count = jffs2_cs_count_rawrefs(c);
	size = sizeof(count) + count * 2 * sizeof(void *);

	c->cs_struct->raw_nii = vmalloc(size);
	if (!c->cs_struct->raw_nii) {
		JFFS2_WARNING("Can't allocate memory!\n");
		return -ENOMEM;
	}

	ptr = c->cs_struct->raw_nii;

	*((int *) ptr) = count;
	ptr += sizeof(count);
	rn = ptr;

	for_each_inode(i, c, ic) {
		raw = ic->nodes;
		while (raw) {
			rn->raw = raw; 
			rn->nii = raw->next_in_ino;
			rn++;
			raw = raw->next_in_ino;
		}
	}

	return 0;
}

// restore rawrefs next_in_ino field
int jffs2_cs_restore_rawrefs_nii(struct jffs2_sb_info *c)
{
	unsigned int i, count;
	struct jffs2_cent_sum_tmp_raws *rn;

	count = *((int *)c->cs_struct->raw_nii);
	rn = c->cs_struct->raw_nii + sizeof(count);

	for (i=0; i<count; i++) {
		rn->raw->next_in_ino = rn->nii;
		rn++;
	}

	vfree(c->cs_struct->raw_nii);

	return 0;
}

// dump ino_cache data
static int jffs2_cs_dump_ic(struct jffs2_sb_info *c)
{
	int i, ret;
	unsigned int minsize, count = 0;
	uint32_t offset, cachesize = 0, avail = 0;
	struct jffs2_inode_cache *ic;
	struct jffs2_raw_node_ref *raw, *temp;
	struct jffs2_cs_inocache *buff = NULL, *ptr;

	minsize = sizeof(struct jffs2_raw_cs_inocache) + JFFS2_MIN_INOCACHE_NUM * JFFS2_INOCACHE_ENTRY_SIZE;

	dbg_cs2("Start cs_dump_inocache\n");

	ret = jffs2_reserve_space(c, minsize, &offset, &avail, ALLOC_CENT_SUM, 0);
	if (ret) {
		dbg_cs("reserve_space returned: %d\n", ret);
		return ret;
	}

	buff = kmalloc(CENT_SUM_BUFFER_SIZE, GFP_KERNEL);
	if (!buff) {
		JFFS2_ERROR("Can't allocate memory for cent_cache!!!\n");
		return -ENOMEM;
	}

	memset(buff, 0, CENT_SUM_BUFFER_SIZE);
	ptr = buff;

	for_each_inode(i, c, ic) {

		if ((cachesize + JFFS2_INOCACHE_ENTRY_SIZE + sizeof(struct jffs2_raw_cs_inocache) > avail) ||
			(cachesize + JFFS2_INOCACHE_ENTRY_SIZE + sizeof(struct jffs2_raw_cs_inocache) > CENT_SUM_BUFFER_SIZE)) {

			ret = jffs2_cs_write_inocache(c, buff, cachesize, count, offset);
			up(&c->alloc_sem); // locked by jffs2_reserve_space

			if (ret) { // error during write
				dbg_cs("Error during write!\n");
				kfree(buff);
				return ret;
			}

			// alloc for next inocache node
			ret = jffs2_reserve_space(c, minsize, &offset, &avail, ALLOC_CENT_SUM, 0);
			if (ret) {
				dbg_cs("reserve_space returned: %d\n", ret);
				kfree(buff);
				return ret;
			}

			ptr = buff;
			cachesize = 0;
			count = 0; // count = cachesize / JFFS2_INOCACHE_ENTRY_SIZE
		}

		dbg_cs2("Dumping inocache %d \n", ic->ino);

		ptr->ino = cpu_to_je32(ic->ino);
		ptr->nlink = cpu_to_je32(ic->nlink);

		if (ic->state == INO_STATE_UNCHECKED) {
			ptr->state = cpu_to_je32(ic->state);
		} else {
			ptr->state = cpu_to_je32(INO_STATE_CHECKEDABSENT);
		}

		raw = ic->nodes;
		while (raw) {
			temp = raw->next_in_ino;
			raw->next_in_ino = (void *)ic->ino;
			raw = temp;
		}

		cachesize += JFFS2_INOCACHE_ENTRY_SIZE;
		count++;
		ptr++;
	}

	if (cachesize != 0) {

		ret = jffs2_cs_write_inocache(c, buff, cachesize, count, offset);

		up(&c->alloc_sem); // locked by jffs2_reserve_space

		if (ret) { // error during write
			dbg_cs("Error during write!\n");
			kfree(buff);
			return ret;
		}
	} else {
		dbg_cs2("empty buffer -> up alloc_sem\n");
		up(&c->alloc_sem); // locked by jffs2_reserve_space
	}

	kfree(buff);
	dbg_cs2("End cs_dump_inocache\n");

	return 0;
}

// dump jebinfo
static int jffs2_cs_dump_jebs(struct jffs2_sb_info *c)
{
	int i, ret;
	unsigned int minsize, count = 0;
	uint32_t offset, jebinfosize = 0, avail = 0;
	uint32_t erase_cnt;
	struct jffs2_cs_jeb *buff = NULL, *ptr;

	minsize = sizeof(struct jffs2_raw_cs_jebinfo) + JFFS2_MIN_JEBINFO_NUM * JFFS2_JEBINFO_ENTRY_SIZE;

	dbg_cs2("Start cs_dump_jebinfo\n");

	ret = jffs2_reserve_space(c, minsize, &offset, &avail, ALLOC_CENT_SUM, 0);
	if (ret) {
		dbg_cs("reserve_space returned: %d\n", ret);
		return ret;
	}

	buff = kmalloc(CENT_SUM_BUFFER_SIZE, GFP_KERNEL);
	if (!buff) {
		JFFS2_ERROR("Can't allocate memory for cent_jebinfo!!!\n");
		return -ENOMEM;
	}

	memset(buff, 0, CENT_SUM_BUFFER_SIZE);
	ptr = buff;

	for(i=0; i<c->nr_blocks; i++) {

		if (c->blocks[i] == c->cs_struct->csblock)
			continue;

		if ((jebinfosize + JFFS2_JEBINFO_ENTRY_SIZE + sizeof(struct jffs2_raw_cs_jebinfo) > avail) ||
			(jebinfosize + JFFS2_JEBINFO_ENTRY_SIZE + sizeof(struct jffs2_raw_cs_jebinfo) > CENT_SUM_BUFFER_SIZE)) {

			ret = jffs2_cs_write_jebinfo(c, buff, jebinfosize, count, offset);
			up(&c->alloc_sem); // locked by jffs2_reserve_space

			if (ret) { // error during write
				dbg_cs("Error during write!\n");
				kfree(buff);
				return ret;
			}

			// alloc for next inocache node
			ret = jffs2_reserve_space(c, minsize, &offset, &avail, ALLOC_CENT_SUM, 0);
			if (ret) {
				dbg_cs("reserve_space returned: %d\n", ret);
				kfree(buff);
				return ret;
			}

			ptr = buff;
			jebinfosize = 0;
			count = 0; // count = jebinfosize / JFFS2_JEBINFO_ENTRY_SIZE
		}

		ptr->offset = cpu_to_je32(c->blocks[i]->offset);
		ptr->unchecked_size = cpu_to_je32(c->blocks[i]->unchecked_size);
		ptr->used_size = cpu_to_je32(c->blocks[i]->used_size);
		ptr->dirty_size = cpu_to_je32(c->blocks[i]->dirty_size);
		ptr->wasted_size = cpu_to_je32(c->blocks[i]->wasted_size);

		/* EBH extension */
		erase_cnt = c->blocks[i]->erase_count;
		erase_cnt |= ((c->blocks[i]->flags & 1) << 31);
		ptr->erase_count = cpu_to_je32(erase_cnt);

		jebinfosize += JFFS2_JEBINFO_ENTRY_SIZE;
		count++;
		ptr++;

	}

	if (jebinfosize != 0) {

		ret = jffs2_cs_write_jebinfo(c, buff, jebinfosize, count, offset);
		up(&c->alloc_sem); // locked by jffs2_reserve_space

		if (ret) { // error during write
			dbg_cs("Error during write!\n");
			kfree(buff);
			return ret;
		}
	} else {
		dbg_cs2("empty buffer -> up alloc_sem\n");
		up(&c->alloc_sem); // locked by jffs2_reserve_space
	}

	kfree(buff);
	dbg_cs2("End cs_dump_jebinfo\n");

	return 0;
}

// dump raw_node_ref information
static int jffs2_cs_dump_rawrefs(struct jffs2_sb_info *c)
{
	int i, ret;
	unsigned int minsize, count = 0;
	uint32_t offset, rawrefssize = 0, avail = 0;
	struct jffs2_cs_rawref *buff = NULL, *ptr;
	struct jffs2_raw_node_ref *walk = NULL, *temp_raw = NULL;

	minsize = sizeof(struct jffs2_raw_cs_rawref) + JFFS2_MIN_RAWREF_NUM * JFFS2_RAWREF_ENTRY_SIZE;

	dbg_cs2("Start cs_dump_rawrefs\n");

	ret = jffs2_reserve_space(c, minsize, &offset, &avail, ALLOC_CENT_SUM, 0);
	if (ret) {
		dbg_cs("jffs2_reserve_space returned: %d\n", ret);
		return ret;
	}

	buff = kmalloc(CENT_SUM_BUFFER_SIZE, GFP_KERNEL);
	if (!buff) {
		JFFS2_ERROR("Can't allocate memory for cent_rawref!\n");
		return -ENOMEM;
	}

	memset(buff, 0, CENT_SUM_BUFFER_SIZE);
	ptr = buff;

	for(i=0; i<c->nr_blocks; i++) {

		walk = c->blocks[i]->first_node;

		while (walk) {
			if ((rawrefssize + JFFS2_RAWREF_ENTRY_SIZE + sizeof(struct jffs2_raw_cs_rawref) > avail) ||
				(rawrefssize + JFFS2_RAWREF_ENTRY_SIZE + sizeof(struct jffs2_raw_cs_rawref) > CENT_SUM_BUFFER_SIZE)) {

				ret = jffs2_cs_write_rawref(c, buff, rawrefssize, count, offset);

				up(&c->alloc_sem); // locked by jffs2_reserve_space

				if (ret) { // error during write
					dbg_cs("Error during write\n");
					kfree(buff);
					return ret;
				}

				// alloc for next inocache node
				ret = jffs2_reserve_space(c, minsize, &offset, &avail, ALLOC_CENT_SUM, 0);
				if (ret) {
					dbg_cs("reserve_space returned: %d\n", ret);
					kfree(buff);
					return ret;
				}

				ptr = buff;
				rawrefssize = 0;
				count = 0; // count = rawrefssize / 20
			}

			temp_raw = walk;
			walk = walk->next_phys;

			ptr->flash_offset = cpu_to_je32(temp_raw->flash_offset);
			ptr->totlen = cpu_to_je32(temp_raw->__totlen);
			ptr->ino = cpu_to_je32(((unsigned int) temp_raw->next_in_ino));

			rawrefssize += JFFS2_RAWREF_ENTRY_SIZE;
			count++;
			ptr++;
		}
	}

	if (rawrefssize != 0) {

		ret = jffs2_cs_write_rawref(c, buff, rawrefssize, count, offset);

		up(&c->alloc_sem); // locked by jffs2_reserve_space

		if (ret) { // error during write
			dbg_cs("Error during write!\n");
			kfree(buff);
			return ret;
		}
	} else {
		dbg_cs2("empty buffer -> up alloc_sem\n");
		up(&c->alloc_sem); // locked by jffs2_reserve_space
	}

	kfree(buff);
	dbg_cs2("End cs_dump_rawrefs\n");

	return 0;
}

// dump list
static void jffs2_cs_dump_list(struct jffs2_sb_info *c, jint16_t **ptr, struct list_head *l, int count)
{
	struct list_head *tmp;
	**ptr = cpu_to_je16(count);
	(*ptr)++;

	if (count) {
		list_for_each(tmp, l) {
			struct jffs2_eraseblock *jeb = list_entry(tmp, struct jffs2_eraseblock, list);
			**ptr = cpu_to_je16(jeb->offset / c->sector_size);
			(*ptr)++;
		}
	}
}

// dump lists (free, clean , dirty, ...)
static int jffs2_cs_dump_lists(struct jffs2_sb_info *c)
{
	int ret, buff_size, nodesize;
	uint32_t offset, avail;
	int count[7];
	jint16_t *ptr = NULL, *buff = NULL;

	dbg_cs2("Start cs_dump_lists\n");

	buff_size = PAD((7 + c->nr_blocks) * 2);
	nodesize = sizeof(struct jffs2_raw_cs_lists) + buff_size;

	ret = jffs2_reserve_space(c, nodesize, &offset, &avail, ALLOC_CENT_SUM, 0);
	if (ret) {
		dbg_cs("reserve_space returned: %d\n", ret);
		return ret;
	}

	count[0] = count_list(&c->clean_list);
	count[1] = count_list(&c->very_dirty_list);
	count[2] = count_list(&c->dirty_list);
	count[3] = count_list(&c->erasable_list);
	count[4] = count_list(&c->free_list);
	count[5] = count_list(&c->bad_list);
	count[6] = count_list(&c->bad_used_list);

	buff = kmalloc(buff_size, GFP_KERNEL);
	if (!buff) {
		JFFS2_ERROR("Can't allocate memory for cent_lists!!!\n");
		return -ENOMEM;
	}

	memset(buff, 0, buff_size);
	ptr = buff;

	jffs2_cs_dump_list(c, &ptr, &c->clean_list, count[0]);
	jffs2_cs_dump_list(c, &ptr, &c->very_dirty_list, count[1]);
	jffs2_cs_dump_list(c, &ptr, &c->dirty_list, count[2]);
	jffs2_cs_dump_list(c, &ptr, &c->erasable_list, count[3]);
	jffs2_cs_dump_list(c, &ptr, &c->free_list, count[4]);
	jffs2_cs_dump_list(c, &ptr, &c->bad_list, count[5]);
	jffs2_cs_dump_list(c, &ptr, &c->bad_used_list, count[6]);

	ret = jffs2_cs_write_lists(c, buff, nodesize, offset);

	up(&c->alloc_sem); // locked by jffs2_reserve_space

	if (ret) { // error during write
		dbg_cs("Error during write\n");
		kfree(buff);
		return ret;
	}

	kfree(buff);
	dbg_cs2("End cs_dump_lists\n");

	return 0;
}


// dump centralized summary data
void jffs2_cs_dump_centsum_data(struct jffs2_sb_info *c)
{
	dbg_cs2("Dumping centsum data\n");

	c->cs_struct->cs_flags |= JFFS2_CS_DURING_UNMOUNT; // flag during_umount = 1

	if (jffs2_cs_dump_ic(c)) {
		goto out;
	}

	if (jffs2_cs_dump_jebs(c)) {
		goto out;
	}

	if (jffs2_cs_dump_rawrefs(c)) {
		goto out;
	}

	if (jffs2_cs_dump_lists(c)) {
		goto out;
	}

	if (jffs2_cs_write_umount_log(c)) {
		/* TODO */
		/* maybe : find new cs_block
		   now skip, do nothing
		*/
	}

out:
	c->cs_struct->cs_flags &= ~JFFS2_CS_DURING_UNMOUNT; // flag during_umount = 0
}

// write inocache node to flash
static int jffs2_cs_write_inocache(struct jffs2_sb_info *c, struct jffs2_cs_inocache *buff,
										uint32_t cachelen, uint32_t count, uint32_t offset)
{
	int ret, infosize;
	size_t retlen;
	struct kvec vecs[2];
	struct jffs2_raw_cs_inocache cent_cache;
	struct jffs2_eraseblock *jeb;

	dbg_cs("Write inocache node to 0x%x\n", offset);

	infosize = PAD(sizeof(cent_cache) + cachelen);

	cent_cache.magic = cpu_to_je16(JFFS2_MAGIC_BITMASK);
	cent_cache.nodetype = cpu_to_je16(JFFS2_NODETYPE_CENT_SUM_INOCACHE);
	cent_cache.totlen = cpu_to_je32(infosize);
	cent_cache.hdr_crc = cpu_to_je32(crc32(0, &cent_cache, sizeof(struct jffs2_unknown_node) - 4));
	cent_cache.cache_num = cpu_to_je32(count);
	cent_cache.cache_crc = cpu_to_je32(crc32(0, buff, cachelen));
	cent_cache.node_crc = cpu_to_je32(crc32(0, &cent_cache, sizeof(cent_cache) - 8));

	vecs[0].iov_base = &cent_cache;
	vecs[0].iov_len = sizeof(cent_cache);
	vecs[1].iov_base = buff;
	vecs[1].iov_len = cachelen;

	jeb = c->blocks[offset / c->sector_size];

	ret = jffs2_flash_writev(c, vecs, 2, offset, &retlen, 0);
	if (ret || (retlen != infosize)) {
		JFFS2_WARNING("write of %zd bytes at 0x%08x failed. returned %d, retlen %zd\n",
						infosize, offset, ret, retlen);
		WASTED_SPACE(infosize);
		return -EIO;
	}

	DECREASE_FREE_SPACE(infosize);

	ret = jffs2_cs_add_offset(c, offset, infosize);
	if (ret)
		return ret;

	D1(printk("%s: End write\n", __FUNCTION__));

	return 0;
}

// write jebinfo node to flash
static int jffs2_cs_write_jebinfo(struct jffs2_sb_info *c, struct jffs2_cs_jeb *buff,
									uint32_t jebinfolen, uint32_t num, uint32_t offset)
{
	int ret, infosize;
	size_t retlen;
	struct kvec vecs[2];
	struct jffs2_raw_cs_jebinfo cent_jebinfo;
	struct jffs2_eraseblock *jeb;

	dbg_cs("Write jebinfo node to 0x%x\n", offset);

	infosize = PAD(sizeof(cent_jebinfo) + jebinfolen);

	cent_jebinfo.magic = cpu_to_je16(JFFS2_MAGIC_BITMASK);
	cent_jebinfo.nodetype = cpu_to_je16(JFFS2_NODETYPE_CENT_SUM_JEBINFO);
	cent_jebinfo.totlen = cpu_to_je32(infosize);
	cent_jebinfo.hdr_crc = cpu_to_je32(crc32(0, &cent_jebinfo, sizeof(struct jffs2_unknown_node) - 4));
	cent_jebinfo.jebinfo_num = cpu_to_je32(num);
	cent_jebinfo.jebinfo_crc = cpu_to_je32(crc32(0, buff, jebinfolen));
	cent_jebinfo.node_crc = cpu_to_je32(crc32(0, &cent_jebinfo, sizeof(cent_jebinfo) - 8));

	vecs[0].iov_base = &cent_jebinfo;
	vecs[0].iov_len = sizeof(cent_jebinfo);
	vecs[1].iov_base = buff;
	vecs[1].iov_len = jebinfolen;

	jeb = c->blocks[offset / c->sector_size];

	ret = jffs2_flash_writev(c, vecs, 2, offset, &retlen, 0);
	if (ret || (retlen != infosize)) {
		JFFS2_WARNING("write of %zd bytes at 0x%08x failed. returned %d, retlen %zd\n",
						infosize, offset, ret, retlen);
		WASTED_SPACE(infosize);
		return -EIO;
	}

	DECREASE_FREE_SPACE(infosize);

	ret = jffs2_cs_add_offset(c, offset, infosize);
	if (ret)
		return ret;

	dbg_cs("End write\n");

	return 0;
}

// write rawref node to flash
static int jffs2_cs_write_rawref(struct jffs2_sb_info *c, struct jffs2_cs_rawref *buff,
									uint32_t rawreflen, uint32_t num, uint32_t offset)
{
	int ret, infosize;
	size_t retlen;
	struct kvec vecs[2];
	struct jffs2_raw_cs_rawref cent_rawref;
	struct jffs2_eraseblock *jeb;

	dbg_cs("Write rawref node to 0x%x\n", offset);

	infosize = PAD(sizeof(cent_rawref) + rawreflen);

	cent_rawref.magic = cpu_to_je16(JFFS2_MAGIC_BITMASK);
	cent_rawref.nodetype = cpu_to_je16(JFFS2_NODETYPE_CENT_SUM_RAWREF);
	cent_rawref.totlen = cpu_to_je32(infosize);
	cent_rawref.hdr_crc = cpu_to_je32(crc32(0, &cent_rawref, sizeof(struct jffs2_unknown_node) - 4));
	cent_rawref.rawref_num = cpu_to_je32(num);
	cent_rawref.rawref_crc = cpu_to_je32(crc32(0, buff, rawreflen));
	cent_rawref.node_crc = cpu_to_je32(crc32(0, &cent_rawref, sizeof(cent_rawref) - 8));

	vecs[0].iov_base = &cent_rawref;
	vecs[0].iov_len = sizeof(cent_rawref);
	vecs[1].iov_base = buff;
	vecs[1].iov_len = rawreflen;

	jeb = c->blocks[offset / c->sector_size];

	ret = jffs2_flash_writev(c, vecs, 2, offset, &retlen, 0);
	if (ret || (retlen != infosize)) {
		JFFS2_WARNING("write of %zd bytes at 0x%08x failed. returned %d, retlen %zd\n",
						infosize, offset, ret, retlen);
		WASTED_SPACE(infosize);
		return -EIO;
	}

	DECREASE_FREE_SPACE(infosize);

	ret = jffs2_cs_add_offset(c, offset, infosize);
	if (ret)
		return ret;

	dbg_cs("End write\n");

	return 0;
}

// write lists node to flash
static int jffs2_cs_write_lists(struct jffs2_sb_info *c, jint16_t *buff,
									uint32_t listslen, uint32_t offset)
{
	int ret, infosize;
	size_t retlen;
	struct kvec vecs[2];
	struct jffs2_raw_cs_lists cent_lists;
	struct jffs2_eraseblock *jeb;

	dbg_cs("Write lists node to 0x%x\n", offset);

	infosize = PAD(sizeof(cent_lists) + listslen);

	cent_lists.magic = cpu_to_je16(JFFS2_MAGIC_BITMASK);
	cent_lists.nodetype = cpu_to_je16(JFFS2_NODETYPE_CENT_SUM_LISTS);
	cent_lists.totlen = cpu_to_je32(infosize);
	cent_lists.hdr_crc = cpu_to_je32(crc32(0, &cent_lists, sizeof(struct jffs2_unknown_node) - 4));
	cent_lists.lists_crc = cpu_to_je32(crc32(0, buff, listslen));

	vecs[0].iov_base = &cent_lists;
	vecs[0].iov_len = sizeof(cent_lists);
	vecs[1].iov_base = buff;
	vecs[1].iov_len = listslen;

	jeb = c->blocks[offset / c->sector_size];

	ret = jffs2_flash_writev(c, vecs, 2, offset, &retlen, 0);
	if (ret || (retlen != infosize)) {
		JFFS2_WARNING("write of %zd bytes at 0x%08x failed. returned %d, retlen %zd\n",
								infosize, offset, ret, retlen);
		WASTED_SPACE(infosize);
		return -EIO;
	}

	DECREASE_FREE_SPACE(infosize);

	ret = jffs2_cs_add_offset(c, offset, infosize);
	if (ret)
		return ret;

	dbg_cs("End write\n");

	return 0;
}

// write mount log entry to flash
int jffs2_cs_write_mount_log_jeb(struct jffs2_sb_info *c)
{
	struct jffs2_raw_cs_log cent_log;
	struct kvec vecs[1];
	int ret;
	size_t retlen;
	uint32_t offset;

retry:
	if (c->cs_struct->csblock_free < PAD(sizeof(cent_log))) {
		dbg_cs("There is not enough free space for log -> erase it\n");
		// erase_cs_block
		if (jffs2_cs_erase_block(c, c->cs_struct->csblock)) {
			JFFS2_WARNING("Can't erase cent_sum log block\n");
			return -1;
		}

		// write ebh
		if (jffs2_cs_mark_erased_block(c, c->cs_struct->csblock)) {
			JFFS2_WARNING("Can't mark CS log block\n");
			return -1;
		}
	}

	offset = c->sector_size - c->cs_struct->csblock_free; // padded

	cent_log.magic = cpu_to_je16(JFFS2_MAGIC_BITMASK);
	cent_log.nodetype = cpu_to_je16(JFFS2_NODETYPE_CENT_SUM_LOG);
	cent_log.totlen = cpu_to_je32(PAD(sizeof(cent_log)));
	cent_log.hdr_crc = cpu_to_je32(crc32(0, &cent_log, sizeof(struct jffs2_unknown_node) - 4));

	cent_log.valid = 0;
	cent_log.reserved = 0;
	cent_log.erase_size = cpu_to_je32(c->sector_size);
	cent_log.ofs_num = cpu_to_je16(0);
	cent_log.version = cpu_to_je32(JFFS2_CS_VERSION);

	cent_log.ofs_crc = cpu_to_je32(0);
	cent_log.node_crc = cpu_to_je32(crc32(0, &cent_log, sizeof(cent_log) - 8));

	vecs[0].iov_base = &cent_log;
	vecs[0].iov_len = PAD(sizeof(cent_log));

	dbg_cs("Writing mount_log node to offset : %x\n", offset);

	down(&c->alloc_sem);
	ret = jffs2_flash_writev(c, vecs, 1, offset, &retlen, 0);
	up(&c->alloc_sem);

	if (ret || (retlen != PAD(sizeof(cent_log)))) {
		JFFS2_WARNING("write of %zd bytes at 0x%08x failed. returned %d, retlen %zd\n",
						PAD(sizeof(cent_log)), offset, ret, retlen);
		c->cs_struct->csblock_free -= PAD(sizeof(cent_log));
		goto retry;
	}

	c->cs_struct->csblock_free -= PAD(sizeof(cent_log));

	return 0;
}


#ifdef CONFIG_JFFS2_CS_FEB
// write umount log entry to flash
static int jffs2_cs_write_umount_log_jeb(struct jffs2_sb_info *c)
{
	struct jffs2_raw_cs_log cent_log;
	struct cent_sum_offset *walk = NULL, *temp = NULL;
	struct kvec vecs[2];
	int ret, i, infosize, datasize, failure = 0;
	size_t retlen;
	uint32_t offset;
	jint32_t *buff = NULL, *ptr = NULL;

	infosize = PAD(sizeof(cent_log) + c->cs_struct->offset_list_num * 4);
	datasize = infosize - sizeof(cent_log);

retry:
	if (c->cs_struct->csblock_free < infosize) {
		dbg_cs("There is not enough free space for log -> erase it\n");
		// erase_cs_block
		if (jffs2_cs_erase_block(c, c->cs_struct->csblock)){
			JFFS2_WARNING("Can't erase CS log block\n");
			return -1;
		}

		// write ebh
		if (jffs2_cs_mark_erased_block(c, c->cs_struct->csblock)) {
			JFFS2_WARNING("Can't mark CS log block\n");
			return -1;
		}
	}

	offset = c->sector_size - c->cs_struct->csblock_free; // padded

	if (!failure) {

		buff = kmalloc(datasize, GFP_KERNEL);

		if (!buff) {
			JFFS2_WARNING("Can't allocate memory for cent_sum offset buffer!\n");
			return -ENOMEM;
		}

		memset(buff, 0, datasize);

		cent_log.magic = cpu_to_je16(JFFS2_MAGIC_BITMASK);
		cent_log.nodetype = cpu_to_je16(JFFS2_NODETYPE_CENT_SUM_LOG);
		cent_log.totlen = cpu_to_je32(infosize);
		cent_log.hdr_crc = cpu_to_je32(crc32(0, &cent_log, sizeof(struct jffs2_unknown_node) - 4));

		cent_log.valid = 1;
		cent_log.reserved = 0;
		cent_log.erase_size = cpu_to_je32(c->sector_size);
		cent_log.ofs_num = cpu_to_je16(c->cs_struct->offset_list_num);
		cent_log.version = cpu_to_je32(JFFS2_CS_VERSION);

		ptr = buff;
		temp = c->cs_struct->offset_list_head;

		for (i=0; i<c->cs_struct->offset_list_num; i++) {
			walk = temp->next;
			*(ptr++) = cpu_to_je32(temp->offset);
			temp = walk;
		}

		if (walk) {
			JFFS2_ERROR("i == c->cs_struct->offset_list_num, but walk != 0 !(???)\n");
			//BUG();
		}

		cent_log.ofs_crc = cpu_to_je32(crc32(0, buff, datasize));
		cent_log.node_crc = cpu_to_je32(crc32(0, &cent_log, sizeof(cent_log) - 8));

		vecs[0].iov_base = &cent_log;
		vecs[0].iov_len = sizeof(cent_log);
		vecs[1].iov_base = buff;
		vecs[1].iov_len = datasize;

	}

	dbg_cs("Writing umount node to offset : %x\n", offset);

	down(&c->alloc_sem);
	ret = jffs2_flash_writev(c, vecs, 2, offset, &retlen, 0);
	up(&c->alloc_sem);

	if (ret || (retlen != infosize)) {
		JFFS2_WARNING("write of %zd bytes at 0x%08x failed. returned %d, retlen %zd\n",
						infosize, offset, ret, retlen);
		c->cs_struct->csblock_free -= infosize;
		failure = 1;
		goto retry;
	}

	kfree(buff);
	c->cs_struct->csblock_free -= infosize;

	return 0;

}
#endif // CONFIG_JFFS2_CS_FEB

// add a new offset to offset list
static int jffs2_cs_add_offset(struct jffs2_sb_info *c, uint32_t offset, uint32_t length)
{
	struct cent_sum_offset *new;

	new = kmalloc(sizeof(new), GFP_KERNEL);
	if (!new) {
		JFFS2_WARNING("Can't allocate memory for cent_sum_offset!\n");
		return -ENOMEM;
	}

	new->next = NULL;
	new->offset = offset;
	new->length = length;

	if (c->cs_struct->offset_list_head == NULL) {
		c->cs_struct->offset_list_head = new;
		c->cs_struct->offset_list_tail = new;
	} else {
		c->cs_struct->offset_list_tail->next = new;
		c->cs_struct->offset_list_tail = new;
	}

	c->cs_struct->offset_list_num++;
	dbg_cs("Offset list num = %d\n", c->cs_struct->offset_list_num);

	return 0;
}

// clean up offset_list
static void jffs2_cs_clean_offset_list(struct jffs2_sb_info *c)
{
	struct cent_sum_offset *walk;

	while (c->cs_struct->offset_list_head) {
		walk = c->cs_struct->offset_list_head->next;
		kfree(c->cs_struct->offset_list_head);
		c->cs_struct->offset_list_head = walk;
	}

	c->cs_struct->offset_list_tail = NULL;
	c->cs_struct->offset_list_num = 0;

}

// find csblock, call read_log, handle errors
int jffs2_cs_build_filesystem_jeb(struct jffs2_sb_info *c)
{
	int i, ret;
	uint32_t offset;

	for (i=0; i < CENT_SUM_RETRY_EMPTY; i++) {
		offset = c->blocks[i]->offset;
		if (c->mtd->block_isbad(c->mtd, offset) == 2) {
			dbg_cs2("Erase block #%d (offset 0x08%x) is bad! Skip", i, offset);
			continue;
		}

		dbg_cs("csblock is the %d. block\n", i);
		c->cs_struct->csblock = c->blocks[i];
		break;
	}

	if (c->cs_struct->csblock) {

		ret = jffs2_cs_read_log_jeb(c);

		switch (ret) {
			case 0:		// LOG OKK;
				c->cs_struct->cs_flags |= JFFS2_CS_CSBLOCK_VALID; // flag valid = 1
				break;

			case 1:		// MOUNT LOG found (umount failed) -> normal scan
				c->cs_struct->cs_flags |= JFFS2_CS_CSBLOCK_VALID; // flag valid = 1
				return 1;

			case 2:		// LOG not found -> empty and erase block, write mount log, normal scan
				c->cs_struct->cs_flags &= ~JFFS2_CS_CSBLOCK_VALID; // flag valid = 0
				return 1;

			case 3:		// LOG crc check failed => erase block, write mount log -> normal scan
				c->cs_struct->cs_flags |= JFFS2_CS_CSBLOCK_VALID; // flag vlaid = 1

				if (!jffs2_is_readonly(c)) {
					int ret;
					uint32_t len;

					// read the ebh of the csblock from the flash, orherwise we loose it
					ret = jffs2_check_nand_cleanmarker_ebh(c, c->cs_struct->csblock, &len);
					if (ret) {
						dbg_cs("check_nand_cleanmarker_ebh returned with error code (%d)!\n",ret);
					}
					dbg_cs("the erase_count of the CS block is %d\n", c->cs_struct->csblock->erase_count);

					// erase block
					if (jffs2_cs_erase_block(c, c->cs_struct->csblock)) {
						c->cs_struct->csblock = NULL;	// centralized summary disabled
						dbg_cs("csblock erase error -> DISABLING cent_sum!\n");
						return 1;
					}

					// write ebh
					if (jffs2_cs_mark_erased_block(c, c->cs_struct->csblock)) {
						JFFS2_WARNING("Can't mark CS log block\n");
						c->cs_struct->csblock = NULL;	// centralized summary disabled
						return 1;
					}
				}
				return 1;

			default:
				JFFS2_WARNING("cs_read_log returned error\n");
				c->cs_struct->csblock = NULL;	// centralized summary disabled
				return 1;
		}
	} else {
		// TODO disable cent_sum
		c->cs_struct->csblock = NULL;
		return 1;
	}

	dbg_cs("End of cs_build_fs\n");
	return 0;
}

// read csblock, find last log, call process_log, handle errors
static int jffs2_cs_read_log_jeb(struct jffs2_sb_info *c)
{
	int ret;
	uint32_t ofs, prev_ofs, crc;
	unsigned char *flashbuf = NULL;
	struct jffs2_unknown_node *node = NULL;
	struct jffs2_raw_cs_log *log = NULL, *temp_log = NULL;
	struct jffs2_unknown_node crcnode;

	flashbuf = jffs2_cs_alloc_flashbuf(c);
	if (!flashbuf) {
		JFFS2_WARNING("Can't allocate memory for flashbuf !\n");
		return -ENOMEM;
	}

	ret = jffs2_flash_read_safe(c, c->cs_struct->csblock->offset,
										c->sector_size, flashbuf);
	if (ret) {
		jffs2_cs_free_flashbuf(c, flashbuf);
		return -EIO;
	}

	ofs = prev_ofs = 0;
	c->cs_struct->csblock_free = c->sector_size;

	while ( *((uint32_t *) (flashbuf + ofs)) != 0xffffffff) {

		node = (struct jffs2_unknown_node *) (flashbuf + ofs);

		if (je16_to_cpu(node->magic) != JFFS2_MAGIC_BITMASK) {
			JFFS2_WARNING("Magic bitmask not found!\n");

			if(je16_to_cpu(node->magic) == KSAMTIB_CIGAM_2SFFJ) {
				JFFS2_WARNING("Wrong endian filesystem!\n");
				jffs2_cs_free_flashbuf(c, flashbuf);
				return -EIO;
			}
			// TODO handle problem: may be -EIO
			return -EIO;
		}

		/* We seem to have a node of sorts. Check the CRC */
		crcnode.magic = node->magic;
		crcnode.nodetype = cpu_to_je16( je16_to_cpu(node->nodetype) | JFFS2_NODE_ACCURATE);
		crcnode.totlen = node->totlen;
		crc = crc32(0, &crcnode, sizeof(crcnode)-4);

		if (crc != je32_to_cpu(node->hdr_crc)) {
			JFFS2_WARNING("Header crc error!\n");

			if (je16_to_cpu(node->nodetype) == JFFS2_NODETYPE_CENT_SUM_LOG) {
				JFFS2_WARNING("LOG node has crc failure => erase block, write mount log\n");
				jffs2_cs_free_flashbuf(c, flashbuf);
				return 3; // TODO: DANGEROUS SOLUTION
			}

			jffs2_cs_free_flashbuf(c, flashbuf);
			return -EIO;
		}

		if (je16_to_cpu(node->nodetype) != JFFS2_NODETYPE_CENT_SUM_LOG &&
			je16_to_cpu(node->nodetype) != JFFS2_NODETYPE_PADDING) {
				dbg_cs("Not LOG node found!\n");
				jffs2_cs_free_flashbuf(c, flashbuf);
				return 2;
		}

		if (je16_to_cpu(node->nodetype) == JFFS2_NODETYPE_CENT_SUM_LOG)
			prev_ofs = ofs;

		c->cs_struct->csblock_free -= PAD(je32_to_cpu(node->totlen));
		ofs += PAD(je32_to_cpu(node->totlen));

		dbg_cs("c->cs_struct->csblock_free = %d\n", c->cs_struct->csblock_free);

		if (ofs >= c->sector_size) {
			dbg_cs("Ofs points to the next block => node=last_log\n");
			break;
		}
	}

	if (ofs == 0) {
		dbg_cs("No node found in the begining of the block \n");
		// TODO delete ???
		jffs2_cs_free_flashbuf(c, flashbuf);
		return 1;
	}

	dbg_cs("LOG found!!!\n");

	log = (struct jffs2_raw_cs_log *) (flashbuf + prev_ofs);
	crc = crc32(0, log, sizeof(struct jffs2_raw_cs_log) - 8);

	/* check node crc */
	if (crc != je32_to_cpu(log->node_crc)) {
		JFFS2_WARNING("log->node_crc != calculated node_crc\n");
		jffs2_cs_free_flashbuf(c, flashbuf);
		return 3;
	}

	/* check erase block size */
	if (je32_to_cpu(log->erase_size) != c->sector_size) {
		JFFS2_WARNING("Image erase_size (%d) != jffs2 erase_size (%d) -> normal scan\n",
						je32_to_cpu(log->erase_size), c->sector_size);
		jffs2_cs_free_flashbuf(c, flashbuf);
		return 1;
	}

	/* check CS version */
	if (je32_to_cpu(log->version) != JFFS2_CS_VERSION) {
		JFFS2_WARNING("Image has CS with version number %d, which is not supported (%d)\n",
						je32_to_cpu(log->version), JFFS2_CS_VERSION);
		JFFS2_WARNING("Skipping CS information => erase CS block, normal scan\n");
		jffs2_cs_free_flashbuf(c, flashbuf);
		return 3;
	}

	/* check CS log type */
	if (!log->valid) {
		dbg_cs("Mount log found! -> normal scan\n");
		jffs2_cs_free_flashbuf(c, flashbuf);
		return 1;
	}

	/* check offsets crc */
	crc = crc32(0, log->ofs, je32_to_cpu(log->totlen) - sizeof(struct jffs2_raw_cs_log));
	if (crc != je32_to_cpu(log->ofs_crc)) {
		JFFS2_WARNING("log->ofs_crc != calculated ofs_crc\n");
		jffs2_cs_free_flashbuf(c, flashbuf);
		return 3;
	}

	temp_log = kmalloc(PAD(je32_to_cpu(log->totlen)), GFP_KERNEL);
	if (!temp_log) {
		JFFS2_WARNING("Can't allocate memory for temp_log!\n");
		jffs2_cs_free_flashbuf(c, flashbuf);
		return 1;
	}

	memcpy(temp_log, log, PAD(je32_to_cpu(log->totlen)));
	jffs2_cs_free_flashbuf(c, flashbuf);

	ret = jffs2_cs_process_log(c, temp_log);
	if (ret) {
		kfree(temp_log);
		return ret;
	}

	kfree(temp_log);
	return 0;
}

// process log info
static int jffs2_cs_process_log(struct jffs2_sb_info *c, struct jffs2_raw_cs_log *log)
{
	int i, ret;
	int loaded = 0;
	uint32_t loaded_offset = 0;
	uint32_t offset, orig_offset;
	unsigned char *flashbuf = NULL;
	union jffs2_node_union *node;

	flashbuf = jffs2_cs_alloc_flashbuf(c);

	if (!flashbuf) {
		JFFS2_WARNING("Can't allocate memory for flashbuf !\n");
		return 1; //-ENOMEM;
	}

	for(i=0; i<je16_to_cpu(log->ofs_num); i++) {

		offset = je32_to_cpu(log->ofs[i]);
		orig_offset = offset;

		dbg_cs("Offset(%d) = 0x%x\n", i, offset);

		if (!loaded) {

			ret = jffs2_flash_read_safe(c, offset & (~(c->sector_size - 1)), c->sector_size, flashbuf);

			if (ret) {
				JFFS2_WARNING("fill_scan_buf returned error\n");
				jffs2_cs_free_flashbuf(c, flashbuf);
				return 1; //-EIO;
			}

			loaded = 1;
			loaded_offset = offset & (~(c->sector_size - 1));
		} else {
			if (loaded_offset != (offset & (~(c->sector_size - 1)))) {

				ret = jffs2_flash_read_safe(c, offset & (~(c->sector_size - 1)), c->sector_size, flashbuf);

				if (ret) {
					JFFS2_WARNING("fill_scan_buf returned error\n");
					jffs2_cs_free_flashbuf(c, flashbuf);
					return 1; //-EIO;
				}

				loaded_offset = offset & (~(c->sector_size - 1));
			}
		}

		offset = offset % c->sector_size;
		node = (union jffs2_node_union *) (flashbuf + offset);

		ret = jffs2_cs_process_node(c, node, orig_offset);

		if (ret) {
			jffs2_cs_free_flashbuf(c, flashbuf);
			jffs2_cs_clean_offset_list(c);
			return ret;
		}
	}

	jffs2_cs_free_flashbuf(c, flashbuf);

	if (jffs2_cs_mode() == JFFS2_CS_MODE_MP_SYSFS) {
		ret = jffs2_cs_add_offset(c, c->cs_struct->cs_log_offset, je32_to_cpu(log->totlen));
		if (ret) {
			return ret;
		}
	}

	// all data was loaded, correct jeb & sb space counters
	jffs2_cs_correct_jebinfo(c);

	return 0;
}

// process cent_sum node
static int jffs2_cs_process_node(struct jffs2_sb_info *c, union jffs2_node_union *node, uint32_t offset)
{
	uint32_t crc;

	crc = crc32(0, &(node->u), sizeof(struct jffs2_unknown_node) - 4);
	if (crc != je32_to_cpu(node->u.hdr_crc)) {
		JFFS2_WARNING("Hdr_crc failed on cent_sum (offset) node!\n");
		return 1;
	}

	if (jffs2_cs_add_offset(c, offset, je32_to_cpu(node->u.totlen))) {
		return 1;
	}

	switch (je16_to_cpu(node->u.nodetype)) {

		case JFFS2_NODETYPE_CENT_SUM_INOCACHE:
			return jffs2_cs_process_inocache_node(c, &node->ic);

		case JFFS2_NODETYPE_CENT_SUM_JEBINFO:
			return jffs2_cs_process_jebinfo_node(c, &node->j);

		case JFFS2_NODETYPE_CENT_SUM_RAWREF:
			return jffs2_cs_process_rawref_node(c, &node->r);

		case JFFS2_NODETYPE_CENT_SUM_LISTS:
			return jffs2_cs_process_lists_node(c, &node->li);

		default:
			JFFS2_WARNING("Unknown nodetype!\n");
			return 1;
	}

	return 0;
}

// process cent_sum inocache node
static int jffs2_cs_process_inocache_node(struct jffs2_sb_info *c, struct jffs2_raw_cs_inocache *node)
{
	int i;
	uint32_t crc;
	struct jffs2_cs_inocache *ptr = NULL;
	struct jffs2_inode_cache *ic;

	crc = crc32(0, node, sizeof(struct jffs2_raw_cs_inocache) - 8);
	if (crc != je32_to_cpu(node->node_crc)) {
		JFFS2_WARNING("Crc failure on cent_sum inocache: node_crc!\n");
		return 1;
	}

	crc = crc32(0, node->cache, je32_to_cpu(node->cache_num) * JFFS2_INOCACHE_ENTRY_SIZE);
	if (crc != je32_to_cpu(node->cache_crc)) {
		JFFS2_WARNING("Crc failure on cent_sum inocache: cache_crc!\n");
		return 1;
	}

	ptr = (struct jffs2_cs_inocache *) node->cache;

	for(i=0; i<je32_to_cpu(node->cache_num); i++) {

		ic = jffs2_alloc_inode_cache();
		if (!ic) {
			JFFS2_WARNING("allocation of inode cache failed\n");
			return 1;
		}
		memset(ic, 0, sizeof(*ic));

		ic->ino = je32_to_cpu(ptr->ino);
		ic->nlink = je32_to_cpu(ptr->nlink);
		ic->state = je32_to_cpu(ptr->state);
		ptr++;

		dbg_cs2("Adding inocache %d\n", ic->ino);

		ic->nodes = (void *)ic;
		jffs2_add_ino_cache(c, ic);

		if (ic->ino > c->highest_ino)
			c->highest_ino = ic->ino;

	}

	return 0;
}

// process cent_sum jebinfo node
static int jffs2_cs_process_jebinfo_node(struct jffs2_sb_info *c, struct jffs2_raw_cs_jebinfo *node)
{
	int i, tmp;
	uint32_t crc;
	uint32_t erase_cnt;
	struct jffs2_cs_jeb *ptr = NULL;
	struct jffs2_eraseblock *jeb;

	crc = crc32(0, node, sizeof(struct jffs2_raw_cs_jebinfo) - 8);
	if (crc != je32_to_cpu(node->node_crc)) {
		JFFS2_WARNING("Crc failure on cent_sum jebinfo: node_crc!\n");
		return 1;
	}

	crc = crc32(0, node->jebinfo, je32_to_cpu(node->jebinfo_num) * JFFS2_JEBINFO_ENTRY_SIZE);
	if (crc != je32_to_cpu(node->jebinfo_crc)) {
		JFFS2_WARNING("Crc failure on cent_sum jebinfo: jebinfo_crc!\n");
		return 1;
	}

	ptr = (struct jffs2_cs_jeb *) node->jebinfo;

	for(i=0; i<je32_to_cpu(node->jebinfo_num); i++) {

		jeb = c->blocks[je32_to_cpu(ptr->offset) / c->sector_size];
		jeb->unchecked_size = je32_to_cpu(ptr->unchecked_size);
		jeb->used_size = je32_to_cpu(ptr->used_size);
		jeb->dirty_size = je32_to_cpu(ptr->dirty_size);
		jeb->wasted_size = je32_to_cpu(ptr->wasted_size);

		/* EBH extension*/
		erase_cnt = je32_to_cpu(ptr->erase_count);
		jeb->erase_count = erase_cnt & ~(1 << 31);
		jeb->flags |= erase_cnt >> 31;

		ptr++;

		c->unchecked_size += jeb->unchecked_size;
		c->used_size += jeb->used_size;
		c->dirty_size += jeb->dirty_size;
		c->wasted_size += jeb->wasted_size;

		// calculate free size
		tmp = jeb->unchecked_size + jeb->used_size + jeb->dirty_size + jeb->wasted_size;
		jeb->free_size = c->sector_size - tmp;

		c->free_size -= tmp;

	}

	return 0;
}

// process cent_sum rawref node
static int jffs2_cs_process_rawref_node(struct jffs2_sb_info *c, struct jffs2_raw_cs_rawref *node)
{
	int i;
	uint32_t crc;
	uint32_t ino;
	struct jffs2_cs_rawref *ptr = NULL;
	struct jffs2_eraseblock *jeb;
	struct jffs2_raw_node_ref *raw;
	struct jffs2_inode_cache *ic;

	crc = crc32(0, node, sizeof(struct jffs2_raw_cs_rawref) - 8);
	if (crc != je32_to_cpu(node->node_crc)) {
		JFFS2_WARNING("Crc failure on cent_sum rawref: node_crc!\n");
		return 1;
	}

	crc = crc32(0, node->rawrefs, je32_to_cpu(node->rawref_num) * JFFS2_RAWREF_ENTRY_SIZE);
	if (crc != je32_to_cpu(node->rawref_crc)) {
		JFFS2_WARNING("Crc failure on cent_sum rawref: rawref_crc!\n");
		return 1;
	}

	ptr = (struct jffs2_cs_rawref *) node->rawrefs;

	for(i=0; i<je32_to_cpu(node->rawref_num); i++) {

		raw = jffs2_alloc_raw_node_ref();

		if (!raw) {
			JFFS2_WARNING("allocation of node reference failed\n");
			return 1;
		}

		raw->flash_offset = je32_to_cpu(ptr->flash_offset);
		raw->__totlen = je32_to_cpu(ptr->totlen);
		ino = je32_to_cpu(ptr->ino);
		ptr++;

		jeb = c->blocks[raw->flash_offset / c->sector_size];

		raw->next_phys = NULL;

		if (!jeb->first_node)
			jeb->first_node = raw;
		if (jeb->last_node)
			jeb->last_node->next_phys = raw;
		jeb->last_node = raw;

		if (ino) {
			ic = jffs2_get_ino_cache(c, ino);
			if (!ic) {
				JFFS2_WARNING("Inocache not found for ino %d at offset 0x%x!\n",
								ino, raw->flash_offset);
				BUG();
			}

			raw->next_in_ino = ic->nodes;
			ic->nodes = raw;
		} else {
			raw->next_in_ino = NULL;
		}
	}

	return 0;
}


// process list entry
static void jffs2_cs_process_list(struct jffs2_sb_info *c, jint16_t **ptr, struct list_head *l)
{
	int count, i;

	count = je16_to_cpu(**ptr);
	(*ptr)++;

	if (!count) {
		return;
	}

	for (i=0; i<count; i++) {
		list_add_tail(&c->blocks[je16_to_cpu(**ptr)]->list, l);
		(*ptr)++;
	}

}

// process cent_sum lists node
static int jffs2_cs_process_lists_node(struct jffs2_sb_info *c, struct jffs2_raw_cs_lists *node)
{
	uint32_t crc;
	jint16_t *ptr;

	crc = crc32(0, node->lists, je32_to_cpu(node->totlen) - sizeof (struct jffs2_raw_cs_lists));
	if (crc != je32_to_cpu(node->lists_crc)) {
		JFFS2_WARNING("Crc failure on cent_sum lists: lists_crc!\n");
		return 1;
	}

	ptr = node->lists;

	jffs2_cs_process_list(c, &ptr, &c->clean_list);
	jffs2_cs_process_list(c, &ptr, &c->very_dirty_list);
	jffs2_cs_process_list(c, &ptr, &c->dirty_list);
	jffs2_cs_process_list(c, &ptr, &c->erasable_list);

	c->nr_free_blocks = je16_to_cpu(*ptr);
	jffs2_cs_process_list(c, &ptr, &c->free_list);

	c->bad_size = je16_to_cpu(*ptr) * c->sector_size;
	c->free_size -= c->bad_size;
	jffs2_cs_process_list(c, &ptr, &c->bad_list);

	// TODO !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	jffs2_cs_process_list(c, &ptr, &c->bad_used_list);

	return 0;
}


// add a list to a specified list of wl
static void jffs2_cs_add_list_wl(struct jffs2_sb_info *c, struct list_head *li, uint8_t flag)
{
	/* ugly debug */
	struct jffs2_eraseblock *jeb;
	int i = 0;
	list_for_each_entry(jeb, li, list) {
		if (EBFLAGS_HAS_EBH(jeb)) {
			i++;
			D2(printk("%03d (%d) ", jeb->offset / c->sector_size, jeb->erase_count));
			record_erase_count(c, jeb);
			jffs2_add_to_hash_table(c, jeb, flag);
			if (i == 10) {
				i = 0;
				D2(printk(" \n"));
			}
		}
	}
	D2(printk(" \n"));
}

// build wear leveling
static void jffs2_cs_build_wear_leveling(struct jffs2_sb_info *c)
{
	dbg_cs("Processing free_list ...\n");
	// add to free wl list
	jffs2_cs_add_list_wl(c, &c->free_list, 2);

	// add to used wl list
	dbg_cs("Processing clean_list ...\n");
	jffs2_cs_add_list_wl(c, &c->clean_list, 1);
	dbg_cs("Processing dirty_list ...\n");
	jffs2_cs_add_list_wl(c, &c->dirty_list, 1);
	dbg_cs("Processing very_dirty_list ...\n");
	jffs2_cs_add_list_wl(c, &c->very_dirty_list, 1);

	if (jffs2_cs_mode() == JFFS2_CS_MODE_FEB) {
		uint32_t len;
		int ret;

		ret = jffs2_check_nand_cleanmarker_ebh(c, c->cs_struct->csblock, &len);
		if (ret) {
			dbg_cs("check_nand_cleanmarker_ebh returned with error code (%d)!\n",ret);
		}
		dbg_cs("the erase_count of the CS block is %d\n", c->cs_struct->csblock->erase_count);
	}
}

// correct jeb space information
static int jffs2_cs_correct_jebinfo(struct jffs2_sb_info *c)
{
	struct cent_sum_offset *walk;
	struct jffs2_eraseblock *jeb;

	while(c->cs_struct->offset_list_head) {
		walk = c->cs_struct->offset_list_head->next;

		jeb = c->blocks[c->cs_struct->offset_list_head->offset / c->sector_size];

		WASTED_SPACE(c->cs_struct->offset_list_head->length);

		// correct wasted & dirty size by jeb change, move jeb to an other list if it is nessesery
		if (walk) {
			if (jeb != c->blocks[walk->offset / c->sector_size]) {
				list_del(&jeb->list);

				c->wasted_size += jeb->free_size;
				c->free_size -= jeb->free_size;
				jeb->wasted_size += jeb->free_size;
				jeb->free_size = 0;

				if (ISDIRTY (jeb->wasted_size + jeb->dirty_size)) {
					c->dirty_size += jeb->wasted_size;
					c->wasted_size -= jeb->wasted_size;
					jeb->dirty_size += jeb->wasted_size;
					jeb->wasted_size = 0;
					if (VERYDIRTY(c, jeb->dirty_size)) {
						dbg_cs("Adding full erase block at 0x%08x to very_dirty_list "
								"(free 0x%08x, dirty 0x%08x, used 0x%08x\n",
								jeb->offset, jeb->free_size, jeb->dirty_size,
								jeb->used_size);
						list_add_tail(&jeb->list, &c->very_dirty_list);
					} else {
						dbg_cs("Adding full erase block at 0x%08x to dirty_list "
								"(free 0x%08x, dirty 0x%08x, used 0x%08x\n",
								jeb->offset, jeb->free_size, jeb->dirty_size,
								jeb->used_size);
						list_add_tail(&jeb->list, &c->dirty_list);
					}
				} else {
					dbg_cs("Adding full erase block at 0x%08x to clean_list "
							"(free 0x%08x, dirty 0x%08x, used 0x%08x\n",
							jeb->offset, jeb->free_size, jeb->dirty_size,
							jeb->used_size);
					list_add_tail(&jeb->list, &c->clean_list);
				}

			}
		} else { // this is the last offset
			uint32_t padding;

			padding = jeb->free_size % c->wbuf_pagesize;

			if (padding) {
				DIRTY_SPACE(padding);
			}

			if (!jffs2_sum_active()) {
				list_del(&jeb->list);
				c->nextblock = jeb; //last used block
			} else {
				list_del(&jeb->list);

				c->wasted_size += jeb->free_size;
				c->free_size -= jeb->free_size;
				jeb->wasted_size += jeb->free_size;
				jeb->free_size = 0;

				if (ISDIRTY (jeb->wasted_size + jeb->dirty_size)) {
					c->dirty_size += jeb->wasted_size;
					c->wasted_size -= jeb->wasted_size;
					jeb->dirty_size += jeb->wasted_size;
					jeb->wasted_size = 0;
					if (VERYDIRTY(c, jeb->dirty_size)) {
						dbg_cs("Adding full erase block at 0x%08x to very_dirty_list "
								"(free 0x%08x, dirty 0x%08x, used 0x%08x\n",
								jeb->offset, jeb->free_size, jeb->dirty_size,
								jeb->used_size);
						list_add_tail(&jeb->list, &c->very_dirty_list);
					} else {
						dbg_cs("Adding full erase block at 0x%08x to dirty_list "
								"(free 0x%08x, dirty 0x%08x, used 0x%08x\n",
							jeb->offset, jeb->free_size, jeb->dirty_size,
							jeb->used_size);
						list_add_tail(&jeb->list, &c->dirty_list);
					}
				} else {
					dbg_cs("Adding full erase block at 0x%08x to clean_list "
							"(free 0x%08x, dirty 0x%08x, used 0x%08x\n",
						jeb->offset, jeb->free_size, jeb->dirty_size,
						jeb->used_size);
					list_add_tail(&jeb->list, &c->clean_list);
				}
			}
		}

		kfree(c->cs_struct->offset_list_head);
		c->cs_struct->offset_list_head = walk;
		c->cs_struct->offset_list_num--;
	}

	if (jffs2_cs_mode() == JFFS2_CS_MODE_FEB) {
		// correct c->free_size
		dbg_cs2("unc : %d, was : %d, dir : %d, use : %d, fre : %d\n",
			c->unchecked_size, c->wasted_size, c->dirty_size, c->used_size, c->free_size);

		c->used_size += c->sector_size;
		c->free_size -= c->sector_size;

		dbg_cs2("unc : %d, was : %d, dir : %d, use : %d, fre : %d\n",
			c->unchecked_size, c->wasted_size, c->dirty_size, c->used_size, c->free_size);
	}

	jffs2_cs_build_wear_leveling(c);

	return 0;
}

// correct jeb space information at remount
int jffs2_cs_correct_jebinfo_remount(struct jffs2_sb_info *c)
{
	struct cent_sum_offset *walk;
	struct jffs2_eraseblock *jeb;

	while(c->cs_struct->offset_list_head) {
		walk = c->cs_struct->offset_list_head->next;

		jeb = c->blocks[c->cs_struct->offset_list_head->offset / c->sector_size];

		INCREASE_WASTED_SPACE(c->cs_struct->offset_list_head->length);

		// correct wasted & dirty size by jeb change, move jeb to an other list if it is nessesery
		if (walk) {
			if (jeb != c->blocks[walk->offset / c->sector_size]) {
				list_del(&jeb->list);

				c->wasted_size += jeb->free_size;
				c->free_size -= jeb->free_size;
				jeb->wasted_size += jeb->free_size;
				jeb->free_size = 0;

				if (ISDIRTY (jeb->wasted_size + jeb->dirty_size)) {
					c->dirty_size += jeb->wasted_size;
					c->wasted_size -= jeb->wasted_size;
					jeb->dirty_size += jeb->wasted_size;
					jeb->wasted_size = 0;
					if (VERYDIRTY(c, jeb->dirty_size)) {
						dbg_cs("Adding full erase block at 0x%08x to very_dirty_list "
								"(free 0x%08x, dirty 0x%08x, used 0x%08x\n",
								jeb->offset, jeb->free_size, jeb->dirty_size,
								jeb->used_size);
						list_add_tail(&jeb->list, &c->very_dirty_list);
					} else {
						dbg_cs("Adding full erase block at 0x%08x to dirty_list "
								"(free 0x%08x, dirty 0x%08x, used 0x%08x\n",
								jeb->offset, jeb->free_size, jeb->dirty_size,
								jeb->used_size);
						list_add_tail(&jeb->list, &c->dirty_list);
					}
				} else {
					dbg_cs("Adding full erase block at 0x%08x to clean_list "
							"(free 0x%08x, dirty 0x%08x, used 0x%08x\n",
							jeb->offset, jeb->free_size, jeb->dirty_size,
							jeb->used_size);
					list_add_tail(&jeb->list, &c->clean_list);
				}

			}
		} else {
			// this is the last offset
		}

		kfree(c->cs_struct->offset_list_head);
		c->cs_struct->offset_list_head = walk;
		c->cs_struct->offset_list_num--;
	}

	if (c->cs_struct->offset_list_num) {
		dbg_cs("List is empty, but list_num != 0\n");
		BUG();
	}

	return 0;
}

// if we can't load the log => reset sb_info
void jffs2_cs_reset_sb_info(struct jffs2_sb_info *c)
{
	int i;

	c->used_size = 0;
	c->dirty_size = 0;
	c->wasted_size = 0;
	c->free_size = c->flash_size;
	c->erasing_size = 0;
	c->bad_size = 0;
	c->unchecked_size = 0;

	c->nr_free_blocks = 0;
	c->nr_erasing_blocks = 0;

	c->nextblock = NULL;

	for (i=0; i<c->nr_blocks; i++) {
		INIT_LIST_HEAD(&c->blocks[i]->list);
		INIT_LIST_HEAD(&c->blocks[i]->hash_list);
		c->blocks[i]->offset = i * c->sector_size;
		c->blocks[i]->free_size = c->sector_size;
		c->blocks[i]->dirty_size = 0;
		c->blocks[i]->wasted_size = 0;
		c->blocks[i]->unchecked_size = 0;
		c->blocks[i]->used_size = 0;
		c->blocks[i]->first_node = NULL;
		c->blocks[i]->last_node = NULL;
		c->blocks[i]->bad_count = 0;
		c->blocks[i]->erase_count = 0;
		c->blocks[i]->flags = 0;
	}

	INIT_LIST_HEAD(&c->clean_list);
	INIT_LIST_HEAD(&c->very_dirty_list);
	INIT_LIST_HEAD(&c->dirty_list);
	INIT_LIST_HEAD(&c->erasable_list);
	INIT_LIST_HEAD(&c->erasing_list);
	INIT_LIST_HEAD(&c->erase_pending_list);
	INIT_LIST_HEAD(&c->erasable_pending_wbuf_list);
	INIT_LIST_HEAD(&c->erase_complete_list);
	INIT_LIST_HEAD(&c->free_list);
	INIT_LIST_HEAD(&c->bad_list);
	INIT_LIST_HEAD(&c->bad_used_list);
	c->highest_ino = 1;
}

static void* jffs2_cs_alloc_flashbuf(struct jffs2_sb_info *c)
{
	if (c->sector_size > 128*1024)
		return vmalloc(c->sector_size);
	else
		return kmalloc(c->sector_size, GFP_KERNEL);
}

static void jffs2_cs_free_flashbuf(struct jffs2_sb_info *c, void *flashbuf)
{
	if (c->sector_size > 128*1024)
		vfree(flashbuf);
	else
		kfree(flashbuf);
}

#ifdef CONFIG_JFFS2_CS_MP_SYSFS
/*******************************************************************************
*                MOUNT PARAMETER / SYSFS MODE SPECIFIC FUNTIONS                *
*******************************************************************************/

enum {
	Opt_cs_offset, Opt_err,
};

static match_table_t tokens = {
	{Opt_cs_offset, "cs_offset=%u"},
	{Opt_err, NULL},
};


// parse mount parameters for "cs_offset = value"
static int jffs2_cs_parse_mount_params(struct jffs2_sb_info *c, char *options)
{
	char *p;
	substring_t args[MAX_OPT_ARGS];

	if (!options) {
		JFFS2_WARNING("no mount parameter found!\n");
		return 1;
	}

	while ((p = strsep(&options, ",")) != NULL) {

		int token, n;
		if (!*p)
			continue;

		token = match_token(p, tokens, args);

		switch (token) {

			case Opt_cs_offset:
				if (match_int(&args[0], &n)) {
					JFFS2_WARNING("Match_int failure!\n");
					return 0;
				}

				if (n < 0) {
					JFFS2_WARNING("negative parameter is not allowed!\n");
					return 0;
				}

				c->cs_struct->cs_log_offset = n;
				c->cs_struct->cs_flags |= JFFS2_CS_MOUNT_OFFSET_PARAM;
				dbg_cs("cs_log_offset = %d\n", c->cs_struct->cs_log_offset);
				break;

			default:
				JFFS2_WARNING("Unrecognized mount option \"%s\" "
					"or missing value!\n", p);
				return 0;
		}
	}

	return 1;
}

// write mount log
int jffs2_cs_write_mount_log_mntsys(struct jffs2_sb_info *c)
{
	return 0;
}

// write umount log (set centsum_offset /sysfs/)
static int jffs2_cs_write_umount_log_mntsys(struct jffs2_sb_info *c)
{
	struct jffs2_raw_cs_log cent_log;
	struct cent_sum_offset *walk = NULL, *temp = NULL;
	struct kvec vecs[2];
	struct jffs2_eraseblock *jeb = NULL;
	int ret, i, infosize, datasize;
	size_t retlen;
	uint32_t offset, avail;
	jint32_t *buff = NULL, *ptr = NULL;

	infosize = PAD(sizeof(cent_log) + c->cs_struct->offset_list_num * 4);
	datasize = infosize - sizeof(cent_log);

	ret = jffs2_reserve_space(c, infosize, &offset, &avail, ALLOC_CENT_SUM, 0);
	if (ret) {
		dbg_cs("reserve_space returned: %d\n", ret);
		return ret;
	}

	buff = kmalloc (datasize, GFP_KERNEL);

	if (!buff) {
		JFFS2_WARNING("Can't allocate memory for cent_sum offset buffer!\n");
		return -ENOMEM;
	}

	memset(buff, 0, datasize);

	cent_log.magic = cpu_to_je16(JFFS2_MAGIC_BITMASK);
	cent_log.nodetype = cpu_to_je16(JFFS2_NODETYPE_CENT_SUM_LOG);
	cent_log.totlen = cpu_to_je32(infosize);
	cent_log.hdr_crc = cpu_to_je32(crc32(0, &cent_log, sizeof(struct jffs2_unknown_node) - 4));

	cent_log.valid = 1;
	cent_log.reserved = 0;
	cent_log.erase_size = cpu_to_je32(c->sector_size);
	cent_log.ofs_num = cpu_to_je16(c->cs_struct->offset_list_num);
	cent_log.version = cpu_to_je32(JFFS2_CS_VERSION);

	ptr = buff;
	temp = c->cs_struct->offset_list_head;

	for (i=0; i<c->cs_struct->offset_list_num; i++) {
		walk = temp->next;
		*(ptr++) = cpu_to_je32(temp->offset);
		temp = walk;
	}

	if (walk) {
		JFFS2_WARNING("i == c->cs_struct->offset_list_num, but walk != 0 !(???)\n");
		//BUG();
	}

	cent_log.ofs_crc = cpu_to_je32(crc32(0, buff, datasize));
	cent_log.node_crc = cpu_to_je32(crc32(0, &cent_log, sizeof(cent_log) - 8));

	vecs[0].iov_base = &cent_log;
	vecs[0].iov_len = sizeof(cent_log);
	vecs[1].iov_base = buff;
	vecs[1].iov_len = datasize;

	jeb = c->blocks[offset / c->sector_size];

	dbg_cs("Writing umount node to offset : %x\n", offset);

	ret = jffs2_flash_writev(c, vecs, 2, offset, &retlen, 0);
	up(&c->alloc_sem);

	if (ret || (retlen != infosize)) {
		JFFS2_WARNING("write of %zd bytes at 0x%08x failed. returned %d, retlen %zd\n",
						infosize, offset, ret, retlen);

		kfree(buff);
		WASTED_SPACE(infosize);
		return -EIO;
	}

	kfree(buff);
	DECREASE_FREE_SPACE(infosize);
	centsum_offset = offset;

	ret = jffs2_cs_add_offset(c, offset, infosize);
	if (ret)
		return ret;

	return 0;
}


// read cs log from flash, check it, load it
static int jffs2_cs_read_log_mntsys(struct jffs2_sb_info *c)
{
	int ret, load_size;
	uint32_t crc;
	unsigned char *flashbuf = NULL;
	struct jffs2_unknown_node *node = NULL, crcnode;
	struct jffs2_raw_cs_log *log = NULL, *temp_log = NULL;

	flashbuf = kmalloc(c->sector_size, GFP_KERNEL);

	if (!flashbuf) {
		JFFS2_WARNING("Can't allocate memory for flashbuf !\n");
		return -ENOMEM;
	}

	load_size = c->sector_size - (c->cs_struct->cs_log_offset % c->sector_size);

	ret = jffs2_flash_read_safe(c, c->cs_struct->cs_log_offset, load_size, flashbuf);
	if (ret) {
		jffs2_cs_free_flashbuf(c, flashbuf);
		return -EIO;
	}

	node = (struct jffs2_unknown_node *) flashbuf;

	if (je16_to_cpu(node->magic) != JFFS2_MAGIC_BITMASK) {
		JFFS2_WARNING("Magic bitmask not found!\n");

		if(je16_to_cpu(node->magic) == KSAMTIB_CIGAM_2SFFJ) {
			JFFS2_WARNING("Wrong endian filesystem!\n");
			jffs2_cs_free_flashbuf(c, flashbuf);
			return -EIO;
		}
		// TODO handle problem: may be -EIO
		return -EIO;
	}

	crcnode.magic = node->magic;
	crcnode.nodetype = cpu_to_je16(je16_to_cpu(node->nodetype) | JFFS2_NODE_ACCURATE);
	crcnode.totlen = node->totlen;

	crc = crc32(0, &crcnode, sizeof(crcnode) - 4);
	if (crc != je32_to_cpu(node->hdr_crc)) {
		JFFS2_WARNING("Header crc error!\n");
		jffs2_cs_free_flashbuf(c, flashbuf);
		return -EIO;
	}

	if (je16_to_cpu(node->nodetype) == JFFS2_NODETYPE_CENT_SUM_LOG) {
		dbg_cs("LOG found !!!\n");
		log = (struct jffs2_raw_cs_log *) node;
	} else {
		dbg_cs("not LOG found !!!\n");
		jffs2_cs_free_flashbuf(c, flashbuf);
		return 1;
	}

	/* check node crc */
	crc = crc32(0, log, sizeof(struct jffs2_raw_cs_log) - 8);
	if (crc != je32_to_cpu(log->node_crc)) {
		JFFS2_WARNING("log->node_crc != calculated node_crc\n");
		jffs2_cs_free_flashbuf(c, flashbuf);
		return 1;
	}

	/* check erase block size */
	if (je32_to_cpu(log->erase_size) != c->sector_size) {
		JFFS2_WARNING("Image erase_size (%d) != jffs2 erase_size (%d) -> normal scan\n",
					je32_to_cpu(log->erase_size), c->sector_size);
		jffs2_cs_free_flashbuf(c, flashbuf);
		return 1;
	}

	/* check CS version */
	if (je32_to_cpu(log->version) != JFFS2_CS_VERSION) {
		JFFS2_WARNING("Image has CS with version number %d, which is not supported (%d)\n",
						je32_to_cpu(log->version), JFFS2_CS_VERSION);
		JFFS2_WARNING("Skipping CS information => erase CS block, normal scan\n");
		jffs2_cs_free_flashbuf(c, flashbuf);
		return 1;
	}

	/* check CS log type */
	if (!log->valid) {
		dbg_cs("Mount log found! (??? impossible in mount param / sysfs mode)-> normal scan\n");
		jffs2_cs_free_flashbuf(c, flashbuf);
		return 1;
	}

	/* check offsets crc */
	crc = crc32(0, log->ofs, je32_to_cpu(log->totlen) - sizeof(struct jffs2_raw_cs_log));
	if (crc != je32_to_cpu(log->ofs_crc)) {
		JFFS2_WARNING("log->ofs_crc != calculated ofs_crc\n");
		jffs2_cs_free_flashbuf(c, flashbuf);
		return 1;
	}

	temp_log = kmalloc(PAD(je32_to_cpu(log->totlen)), GFP_KERNEL);
	if (!temp_log) {
		JFFS2_WARNING("Can't allocate memory for temp_log!\n");
		jffs2_cs_free_flashbuf(c, flashbuf);
		return 1; //-ENOMEM;
	}

	memcpy(temp_log, log, PAD(je32_to_cpu(log->totlen)));
	jffs2_cs_free_flashbuf(c, flashbuf);

	ret = jffs2_cs_process_log(c, temp_log);
	if (ret) {
		kfree(temp_log);
		return ret;
	}

	kfree(temp_log);
	return 0;
}

// build filesystem
int jffs2_cs_build_filesystem_mntsys(struct jffs2_sb_info *c)
{
	int ret;

	if (!HAS_MOUNT_OFFSET_PARAM) {
		dbg_cs("No mount parameter -> normal scan\n");
		return 1;
	}

	ret = jffs2_cs_read_log_mntsys(c);

	if (ret) {
		dbg_cs("returned with %d\n", ret);
		return 1;
	}

	return 0;
}
#endif /* CONFIG_JFFS2_CS_MP_SYSFS */
