/*
 * JFFS2 -- Journalling Flash File System, Version 2.
 *
 * Copyright (C) 2001-2003 Red Hat, Inc.
 *
 * Created by David Woodhouse <dwmw2@infradead.org>
 *
 * For licensing information, see the file 'LICENCE' in the
 * jffs2 directory.
 *
 * $Id: jffs2.h,v 1.40 2005/11/07 11:14:51 gleixner Exp $
 *
 */

/*
 * Fiona device changes Copyright (C) 2006, Lab126, Inc.  All rights reserved.
 */

#ifndef __LINUX_JFFS2_H__
#define __LINUX_JFFS2_H__

/* You must include something which defines the C99 uintXX_t types.
   We don't do it from here because this file is used in too many
   different environments. */

#define JFFS2_SUPER_MAGIC 0x72b6

/* Values we may expect to find in the 'magic' field */
#define JFFS2_OLD_MAGIC_BITMASK 0x1984
#define JFFS2_MAGIC_BITMASK 0x1985
#define KSAMTIB_CIGAM_2SFFJ 0x8519 /* For detecting wrong-endian fs */
#define JFFS2_EMPTY_BITMASK 0xffff
#define JFFS2_DIRTY_BITMASK 0x0000

/* NOTE:  the Fiona OneNAND part only has 20 bytes of space, which is four
   bytes too small for struct jffs2_raw_ebh.  as this is a closed device,
   the EBH compatibility flags will not be used */
#ifndef CONFIG_ARCH_FIONA
/* JFFS2 eraseblock header compat/incompat/rocompat features set */
#define JFFS2_EBH_COMPAT_FSET 0x00
#define JFFS2_EBH_INCOMPAT_FSET 0x00
#define JFFS2_EBH_ROCOMPAT_FSET 0x00
#endif

/* Summary node MAGIC marker */
#define JFFS2_SUM_MAGIC	0x02851885

/* We only allow a single char for length, and 0xFF is empty flash so
   we don't want it confused with a real length. Hence max 254.
*/
#define JFFS2_MAX_NAME_LEN 254

/* How small can we sensibly write nodes? */
#define JFFS2_MIN_DATA_LEN 128

#define JFFS2_COMPR_NONE	0x00
#define JFFS2_COMPR_ZERO	0x01
#define JFFS2_COMPR_RTIME	0x02
#define JFFS2_COMPR_RUBINMIPS	0x03
#define JFFS2_COMPR_COPY	0x04
#define JFFS2_COMPR_DYNRUBIN	0x05
#define JFFS2_COMPR_ZLIB	0x06
/* Compatibility flags. */
#define JFFS2_COMPAT_MASK 0xc000      /* What do to if an unknown nodetype is found */
#define JFFS2_NODE_ACCURATE 0x2000
/* INCOMPAT: Fail to mount the filesystem */
#define JFFS2_FEATURE_INCOMPAT 0xc000
/* ROCOMPAT: Mount read-only */
#define JFFS2_FEATURE_ROCOMPAT 0x8000
/* RWCOMPAT_COPY: Mount read/write, and copy the node when it's GC'd */
#define JFFS2_FEATURE_RWCOMPAT_COPY 0x4000
/* RWCOMPAT_DELETE: Mount read/write, and delete the node when it's GC'd */
#define JFFS2_FEATURE_RWCOMPAT_DELETE 0x0000

#define JFFS2_NODETYPE_DIRENT (JFFS2_FEATURE_INCOMPAT | JFFS2_NODE_ACCURATE | 1)
#define JFFS2_NODETYPE_INODE (JFFS2_FEATURE_INCOMPAT | JFFS2_NODE_ACCURATE | 2)
#define JFFS2_NODETYPE_CLEANMARKER (JFFS2_FEATURE_RWCOMPAT_DELETE | JFFS2_NODE_ACCURATE | 3)
#define JFFS2_NODETYPE_PADDING (JFFS2_FEATURE_RWCOMPAT_DELETE | JFFS2_NODE_ACCURATE | 4)

#define JFFS2_NODETYPE_ERASEBLOCK_HEADER (JFFS2_FEATURE_RWCOMPAT_DELETE | JFFS2_NODE_ACCURATE | 5)
#define JFFS2_NODETYPE_SUMMARY (JFFS2_FEATURE_RWCOMPAT_DELETE | JFFS2_NODE_ACCURATE | 6)

#define JFFS2_NODETYPE_CENT_SUM_LOG (JFFS2_FEATURE_RWCOMPAT_DELETE | JFFS2_NODE_ACCURATE | 7)
#define JFFS2_NODETYPE_CENT_SUM_INOCACHE (JFFS2_FEATURE_RWCOMPAT_DELETE | JFFS2_NODE_ACCURATE | 8)
#define JFFS2_NODETYPE_CENT_SUM_JEBINFO (JFFS2_FEATURE_RWCOMPAT_DELETE | JFFS2_NODE_ACCURATE | 9)
#define JFFS2_NODETYPE_CENT_SUM_RAWREF (JFFS2_FEATURE_RWCOMPAT_DELETE | JFFS2_NODE_ACCURATE | 10)
#define JFFS2_NODETYPE_CENT_SUM_LISTS (JFFS2_FEATURE_RWCOMPAT_DELETE | JFFS2_NODE_ACCURATE | 11)

// Maybe later...
//#define JFFS2_NODETYPE_CHECKPOINT (JFFS2_FEATURE_RWCOMPAT_DELETE | JFFS2_NODE_ACCURATE | 3)
//#define JFFS2_NODETYPE_OPTIONS (JFFS2_FEATURE_RWCOMPAT_COPY | JFFS2_NODE_ACCURATE | 4)


#define JFFS2_INO_FLAG_PREREAD	  1	/* Do read_inode() for this one at
					   mount time, don't wait for it to
					   happen later */
#define JFFS2_INO_FLAG_USERCOMPR  2	/* User has requested a specific
					   compression type */


/* These can go once we've made sure we've caught all uses without
   byteswapping */

typedef struct {
	uint32_t v32;
} __attribute__((packed))  jint32_t;

typedef struct {
	uint32_t m;
} __attribute__((packed))  jmode_t;

typedef struct {
	uint16_t v16;
} __attribute__((packed)) jint16_t;

struct jffs2_unknown_node
{
	/* All start like this */
	jint16_t magic;
	jint16_t nodetype;
	jint32_t totlen; /* So we can skip over nodes we don't grok */
	jint32_t hdr_crc;
} __attribute__((packed));

struct jffs2_raw_dirent
{
	jint16_t magic;
	jint16_t nodetype;	/* == JFFS2_NODETYPE_DIRENT */
	jint32_t totlen;
	jint32_t hdr_crc;
	jint32_t pino;
	jint32_t version;
	jint32_t ino; /* == zero for unlink */
	jint32_t mctime;
	uint8_t nsize;
	uint8_t type;
	uint8_t unused[2];
	jint32_t node_crc;
	jint32_t name_crc;
	uint8_t name[0];
} __attribute__((packed));

/* The JFFS2 raw inode structure: Used for storage on physical media.  */
/* The uid, gid, atime, mtime and ctime members could be longer, but
   are left like this for space efficiency. If and when people decide
   they really need them extended, it's simple enough to add support for
   a new type of raw node.
*/
struct jffs2_raw_inode
{
	jint16_t magic;      /* A constant magic number.  */
	jint16_t nodetype;   /* == JFFS2_NODETYPE_INODE */
	jint32_t totlen;     /* Total length of this node (inc data, etc.) */
	jint32_t hdr_crc;
	jint32_t ino;        /* Inode number.  */
	jint32_t version;    /* Version number.  */
	jmode_t mode;       /* The file's type or mode.  */
	jint16_t uid;        /* The file's owner.  */
	jint16_t gid;        /* The file's group.  */
	jint32_t isize;      /* Total resultant size of this inode (used for truncations)  */
	jint32_t atime;      /* Last access time.  */
	jint32_t mtime;      /* Last modification time.  */
	jint32_t ctime;      /* Change time.  */
	jint32_t offset;     /* Where to begin to write.  */
	jint32_t csize;      /* (Compressed) data size */
	jint32_t dsize;	     /* Size of the node's data. (after decompression) */
	uint8_t compr;       /* Compression algorithm used */
	uint8_t usercompr;   /* Compression algorithm requested by the user */
	jint16_t flags;	     /* See JFFS2_INO_FLAG_* */
	jint32_t data_crc;   /* CRC for the (compressed) data.  */
	jint32_t node_crc;   /* CRC for the raw inode (excluding data)  */
	uint8_t data[0];
} __attribute__((packed));

struct jffs2_raw_summary
{
	jint16_t magic;
	jint16_t nodetype; 	/* = JFFS2_NODETYPE_SUMMARY */
	jint32_t totlen;
	jint32_t hdr_crc;
	jint32_t sum_num;	/* number of sum entries*/
	jint32_t cln_mkr;	/* clean marker size, 0 = no cleanmarker */
	jint32_t padded;	/* sum of the size of padding nodes */
	jint32_t sum_crc;	/* summary information crc */
	jint32_t node_crc; 	/* node crc */
	jint32_t sum[0]; 	/* inode summary info */
} __attribute__((packed));

struct jffs2_raw_ebh
{
	jint16_t magic;
	jint16_t nodetype; /* == JFFS2_NODETYPE_ERASEBLOCK_HEADER */
	jint32_t totlen;
	jint32_t hdr_crc;
	jint32_t node_crc;
#ifndef CONFIG_ARCH_FIONA
	uint8_t  reserved; /* reserved for future use and alignment */
	uint8_t  compat_fset;
	uint8_t  incompat_fset;
	uint8_t  rocompat_fset;
#endif
	jint32_t erase_count; /* the erase count of this erase block */
	jint32_t data[0];
} __attribute__((packed));

struct jffs2_raw_cs_log
{
	jint16_t magic;
	jint16_t nodetype; 	/* = JFFS2_NODETYPE_CENT_SUM_LOG */
	jint32_t totlen;
	jint32_t hdr_crc;
	uint8_t valid;		/* 0 = mount log, 1 = umount log*/
	uint8_t reserved;
	jint16_t ofs_num;	/* number of jeb offsets */
	jint32_t version;	/* CS log version */
	jint32_t erase_size; /* erase block size */
	jint32_t ofs_crc;	/* jebs crc */
	jint32_t node_crc; 	/* node crc */
	jint32_t ofs[0];	/* jeb offsets, which contains centralized summary information */
} __attribute__((packed));

struct jffs2_raw_cs_inocache
{
	jint16_t magic;
	jint16_t nodetype; 	/* = JFFS2_NODETYPE_CENT_SUM_INOCACHE */
	jint32_t totlen;
	jint32_t hdr_crc;
	jint32_t cache_num;	/* number of cache entries */
	jint32_t cache_crc;	/* cache information crc */
	jint32_t node_crc; 	/* node crc */
	jint32_t cache[0]; 	/* cache info */
} __attribute__((packed));

struct jffs2_raw_cs_jebinfo
{
	jint16_t magic;
	jint16_t nodetype; 	/* = JFFS2_NODETYPE_CENT_SUM_JEBINFO */
	jint32_t totlen;
	jint32_t hdr_crc;
	jint32_t jebinfo_num;	/* number of jeb entries */
	jint32_t jebinfo_crc;	/* jeb crc */
	jint32_t node_crc; 		/* node crc */
	jint32_t jebinfo[0]; 	/* jeb info */
} __attribute__((packed));

struct jffs2_raw_cs_rawref
{
	jint16_t magic;
	jint16_t nodetype; 	/* = JFFS2_NODETYPE_CENT_SUM_RAWREF */
	jint32_t totlen;
	jint32_t hdr_crc;
	jint32_t rawref_num;	/* number of jeb entries */
	jint32_t rawref_crc;	/* rawrefs crc */
	jint32_t node_crc; 		/* node crc */
	jint32_t rawrefs[0]; 	/* rawref info */
} __attribute__((packed));

struct jffs2_raw_cs_lists
{
	jint16_t magic;
	jint16_t nodetype; 	/* = JFFS2_NODETYPE_CENT_SUM_LISTS */
	jint32_t totlen;
	jint32_t hdr_crc;
	jint32_t lists_crc;	/* lists crc */
	jint16_t lists[0]; 	/* list info */
} __attribute__((packed));

union jffs2_node_union
{
	struct jffs2_raw_inode i;
	struct jffs2_raw_dirent d;
	struct jffs2_raw_summary s;
	struct jffs2_raw_ebh eh;
	struct jffs2_unknown_node u;
	struct jffs2_raw_cs_log l;
	struct jffs2_raw_cs_inocache ic;
	struct jffs2_raw_cs_jebinfo j;
	struct jffs2_raw_cs_rawref r;
	struct jffs2_raw_cs_lists li;
};

#endif /* __LINUX_JFFS2_H__ */
