/*
 * wm8731.c  --  WM8731 Soc Audio driver
 *
 * Copyright 2005 Openedhand Ltd.
 *
 * Author: Richard Purdie <richard@openedhand.com>
 *
 * Based on wm8753.c by Liam Girdwood
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/i2c.h>
#include <linux/pm.h>
#include <sound/driver.h>
#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/pcm_params.h>
#include <sound/soc.h>
#include <sound/soc-dpm.h>
#include <sound/initval.h>

#include "wm8731.h"

#define AUDIO_NAME "wm8731"
#define WM8731_VERSION "0.4"


/*
 * Debug
 */
 
#define PFX AUDIO_NAME
#define WM8731_DEBUG 0

#ifdef WM8731_DEBUG
#define dbg(format, arg...) printk(KERN_DEBUG PFX ": " format "\n" , ## arg)
#else
#define dbg(format, arg...) do {} while (0)
#endif
#define err(format, arg...) printk(KERN_ERR PFX ": " format "\n" , ## arg)
#define info(format, arg...) printk(KERN_INFO PFX ": " format "\n" , ## arg)
#define warn(format, arg...) printk(KERN_WARNING PFX ": " format "\n" , ## arg)

#define I2C_DRIVERID_WM8731 0xf2fe

static unsigned short normal_i2c[] = { 0x1a, 0x1b,
                                       I2C_CLIENT_END };

/* Magic definition of all other variables and things */
I2C_CLIENT_INSMOD;

static struct i2c_driver wm8731_i2c_driver;
static struct i2c_client client_template;

/*
 * wm8731 register cache
 * We can't read the WM8731 register space when we are
 * using 2 wire for device control, so we cache them instead. 
 * There is no point in caching the reset register
 */
static const u16 wm8731_reg[WM8731_CACHEREGNUM] = {
    0x0097, 0x0097, 0x0079, 0x0079, 
    0x000a, 0x0008, 0x009f, 0x000a, 
    0x0000, 0x0000
};

#define WM8731_HWFMT \
	(SND_SOC_HWFMT_I2S | SND_SOC_HWFMT_LEFT_J | SND_SOC_HWFMT_RIGHT_J | \
	SND_SOC_HWFMT_CBS_CFS | SND_SOC_HWFMT_CBM_CFM | \
	SND_SOC_HWFMT_NB_NF | SND_SOC_HWFMT_NB_IF | SND_SOC_HWFMT_IB_NF | \
	SND_SOC_HWFMT_IB_IF)

/*
 * priv1 is srate register setting
 */
static struct snd_soc_hw_mode wm8731_hwfmt[] = {

	{WM8731_HWFMT, SND_SOC_HWBITS(16), 8000 ,  64, 0x000c},
	{WM8731_HWFMT, SND_SOC_HWBITS(16), 8000,   96, 0x000e},
	{WM8731_HWFMT, SND_SOC_HWBITS(16), 8000 , 128, 0x004c},
	{WM8731_HWFMT, SND_SOC_HWBITS(16), 8000,  192, 0x004e},

	{WM8731_HWFMT, SND_SOC_HWBITS(16), 32000,  64, 0x0018},
	{WM8731_HWFMT, SND_SOC_HWBITS(16), 32000,  96, 0x001a},
	{WM8731_HWFMT, SND_SOC_HWBITS(16), 32000, 128, 0x0058},
	{WM8731_HWFMT, SND_SOC_HWBITS(16), 32000, 192, 0x005a},

	{WM8731_HWFMT, SND_SOC_HWBITS(16), 44100,  64, 0x0020},
	{WM8731_HWFMT, SND_SOC_HWBITS(16), 44100,  96, 0x0022},
	{WM8731_HWFMT, SND_SOC_HWBITS(16), 44100, 128, 0x0060},
	{WM8731_HWFMT, SND_SOC_HWBITS(16), 44100, 192, 0x0062},

	{WM8731_HWFMT, SND_SOC_HWBITS(16), 48000,  64, 0x0000},
	{WM8731_HWFMT, SND_SOC_HWBITS(16), 48000,  96, 0x0002},
	{WM8731_HWFMT, SND_SOC_HWBITS(16), 48000, 128, 0x0040},
	{WM8731_HWFMT, SND_SOC_HWBITS(16), 48000, 192, 0x0042},

	{WM8731_HWFMT, SND_SOC_HWBITS(16), 88200,  32, 0x003c},
	{WM8731_HWFMT, SND_SOC_HWBITS(16), 88200,  48, 0x003e},
	{WM8731_HWFMT, SND_SOC_HWBITS(16), 88200,  64, 0x007c},
	{WM8731_HWFMT, SND_SOC_HWBITS(16), 88200,  96, 0x007e},

	{WM8731_HWFMT, SND_SOC_HWBITS(16), 96000,  32, 0x001c},
	{WM8731_HWFMT, SND_SOC_HWBITS(16), 96000,  48, 0x001e},
	{WM8731_HWFMT, SND_SOC_HWBITS(16), 96000,  64, 0x005c},
	{WM8731_HWFMT, SND_SOC_HWBITS(16), 96000,  96, 0x005e},
};


/*
 * read wm8731 register cache
 */
static inline unsigned int wm8731_read_reg_cache(struct snd_soc_codec * codec,
	unsigned int reg)
{
	u16 *cache = codec->reg_cache;
	if (reg == WM8731_RESET) 
		return 0;
	if (reg > WM8731_CACHEREGNUM)
		return -1;
	return cache[reg];
}

/*
 * write wm8731 register cache
 */
static inline void wm8731_write_reg_cache(struct snd_soc_codec * codec,
	u16 reg, unsigned int value)
{
	u16 *cache = codec->reg_cache;
	if (reg > WM8731_CACHEREGNUM)
		return;
	cache[reg] = value;
}

/*
 * write to the WM8731 register space
 */
static int wm8731_write(struct snd_soc_codec * codec, unsigned int reg, 
	unsigned int value)
{
	u8 data[2];
	struct i2c_client* i2c = (struct i2c_client*)codec->control_data;
	/* data is 
	 *   D15..D9 WM8731 register offset
	 *   D8...D0 register data
	 */
	data[0] = (reg << 1) | ((value >> 8) & 0x0001);
	data[1] = value & 0x00ff;

	wm8731_write_reg_cache (codec, reg, value);

	if (i2c_master_send(i2c, data, 2) == 2)
		return 0;
	else
		return -1;
}

#define wm8731_reset(c)	wm8731_write(c, WM8731_RESET, 0)

static const char* wm8731_input_select[] = {"Line In", "Mic"};
static const char* wm8731_deemph[] = {"None", "32Khz", "44.1Khz", "48Khz"};

static const struct soc_enum wm8731_enum[] = {
	SOC_ENUM_SINGLE(WM8731_APANA, 2, 2, wm8731_input_select),
	SOC_ENUM_SINGLE(WM8731_APDIGI, 1, 4, wm8731_deemph),
};

static const snd_kcontrol_new_t wm8731_snd_controls[] = {

SOC_SINGLE("Left Playback Volume", WM8731_LOUT1V, 0, 127, 0),
SOC_SINGLE("Left Playback ZC Switch", WM8731_LOUT1V, 7, 1, 0),
SOC_SINGLE("Right Playback Volume", WM8731_ROUT1V, 0, 127, 0),
SOC_SINGLE("Right Playback ZC Switch", WM8731_ROUT1V, 7, 1, 0),

SOC_SINGLE("Line Left Capture Volume", WM8731_LINVOL, 0, 31, 0),
SOC_SINGLE("Line Left Capture Switch", WM8731_LINVOL, 7, 1, 1),
SOC_SINGLE("Line Right Capture Volume", WM8731_RINVOL, 0, 31, 0),
SOC_SINGLE("Line Right Capture Switch", WM8731_RINVOL, 7, 1, 1),

SOC_SINGLE("Mic Boost (+20dB)", WM8731_APANA, 0, 1, 0),
SOC_SINGLE("Mic Capture Switch", WM8731_APANA, 1, 1, 0),

SOC_SINGLE("Sidetone Playback Volume", WM8731_APANA, 6, 3, 1),

SOC_SINGLE("ADC High Pass Filter Switch", WM8731_APDIGI, 0, 1, 1),
SOC_SINGLE("Store DC Offset Switch", WM8731_APDIGI, 4, 1, 0),
	
SOC_ENUM("Playback De-emphasis", wm8731_enum[1]),
};

/* add non dpm controls */
static int wm8731_add_controls(struct snd_soc_codec *codec)
{
	int err, i;

	for (i = 0; i < ARRAY_SIZE(wm8731_snd_controls); i++) {
		if ((err = snd_ctl_add(codec->card, snd_soc_cnew(&wm8731_snd_controls[i],codec, NULL))) < 0)
			return err;
	}
	
	return 0;
}


/* Output Mixer */
static const snd_kcontrol_new_t wm8731_output_mixer_controls[] = {
SOC_DPM_SINGLE("Line Bypass Switch", WM8731_APANA, 3, 1, 0),
SOC_DPM_SINGLE("Mic Sidetone Switch", WM8731_APANA, 5, 1, 0),
SOC_DPM_SINGLE("HiFi Playback Switch", WM8731_APANA, 4, 1, 0),
};

/* Input mux */
static const snd_kcontrol_new_t wm8753_input_mux_controls = 
SOC_DPM_ENUM("Input Select", wm8731_enum[0]);


static const struct snd_soc_dpm_widget wm8731_dpm_widgets[] = {
SND_SOC_DPM_CLOCK("WM8731 Clock", WM8731_PWR, 6, 1),
SND_SOC_DPM_CLOCK("WM8731 Osc", WM8731_PWR, 5, 1),
SND_SOC_DPM_MUTE("DAC Mute", WM8731_APDIGI, 3, 0),
SND_SOC_DPM_MIXER("Output Mixer", WM8731_PWR, 4, 1, &wm8731_output_mixer_controls[0], 
	ARRAY_SIZE(wm8731_output_mixer_controls)),
SND_SOC_DPM_DAC("DAC", "HiFi Playback", WM8731_PWR, 3, 1, NULL, 0),
SND_SOC_DPM_OUTPUT("LOUT"),
SND_SOC_DPM_OUTPUT("LHPOUT"),
SND_SOC_DPM_OUTPUT("ROUT"),
SND_SOC_DPM_OUTPUT("LRHPOUT"),
SND_SOC_DPM_ADC("ADC", "HiFi Capture", WM8731_PWR, 2, 1, NULL, 0),
SND_SOC_DPM_MUX("Input Mux", -1, 0, 0, &wm8753_input_mux_controls),
SND_SOC_DPM_VOLUME("Line Input", WM8731_PWR, 0, 1, NULL, 0),
SND_SOC_DPM_MICBIAS("Mic", WM8731_PWR, 1, 1),
SND_SOC_DPM_INPUT("MICIN"),
SND_SOC_DPM_INPUT("RLINEIN"),
SND_SOC_DPM_INPUT("LLINEIN"),
};

static const char* intercon[][3] = {
	/* output mixer */
	{"Output Mixer", "Line Bypass Switch", "Line Input"},
	{"Output Mixer", "HiFi Playback Switch", "DAC"},
	{"Output Mixer", "Mic Sidetone Switch", "Mic"},
	
	/* outputs */
	{"RHPOUT", NULL, "Output Mixer"},
	{"ROUT", NULL, "Output Mixer"},
	{"LHPOUT", NULL, "Output Mixer"},
	{"LOUT", NULL, "Output Mixer"},
	
	/* input mux */
	{"Input Mux", "Line In", "Line Input"},
	{"Input Mux", "Mic", "Mic"},
	
	/* outputs */
	{"Line Input", NULL, "LLINEIN"},
	{"Line Input", NULL, "RLINEIN"},
	{"Mic", NULL, "MICIN"},
	
	/* terminator */
	{NULL, NULL, NULL},
};

static int wm8731_add_widgets(struct snd_soc_codec *codec)
{
	int i;
	
	for(i = 0; i < ARRAY_SIZE(wm8731_dpm_widgets); i++) {
		snd_soc_dpm_new_control(codec, &wm8731_dpm_widgets[i]);
	}
	
	/* set up audio path interconnects */
	for(i = 0; intercon[i][0] != NULL; i++) {
		snd_soc_dpm_connect_input(codec, intercon[i][0], intercon[i][1], intercon[i][2]);
	}
	
	snd_soc_dpm_sync(codec);
	return 0;
}

static int wm8731_pcm_prepare(snd_pcm_substream_t *substream)
{
	struct snd_soc_pcm_codec *pcm_c = substream->private_data;
	struct snd_soc_codec *codec = pcm_c->codec;
	u16 iface = 0;
	
	/* set master/slave audio interface */
	switch(pcm_c->hw_runtime.hformat & SND_SOC_CLOCK_MASK) {
		case SND_SOC_HWFMT_CBM_CFM:
			iface |= 0x0040;
			break;
		case SND_SOC_HWFMT_CBS_CFS:
			break;
	
	}
	
	/* interface format */
	switch(pcm_c->hw_runtime.hformat & SND_SOC_FORMAT_MASK) {
		case SND_SOC_HWFMT_I2S:
			iface |= 0x0002;
			break;
		case SND_SOC_HWFMT_RIGHT_J:
			break;
		case SND_SOC_HWFMT_LEFT_J:
			iface |= 0x0001;
			break;
		case SND_SOC_HWFMT_DSP_A:
			iface |= 0x0003;
			break;
		case SND_SOC_HWFMT_DSP_B:
			iface |= 0x0013;
			break;
	}
	
	/* bit size */
	switch(pcm_c->hw_runtime.hbits) {
		case SND_SOC_HWBITS(16):
			break;
		case SND_SOC_HWBITS(20):
			iface |= 0x0004;
			break;
		case SND_SOC_HWBITS(24):
			iface |= 0x0008;
			break;
		case SND_SOC_HWBITS(32):
			iface |= 0x000c;
			break;
	}

	/* clock inversion */
	switch(pcm_c->hw_runtime.hformat & SND_SOC_INV_MASK) {
		case SND_SOC_HWFMT_NB_NF:
			break;
		case SND_SOC_HWFMT_IB_IF:
			iface |= 0x0090;
			break;
		case SND_SOC_HWFMT_IB_NF:
			iface |= 0x0080;
			break;
		case SND_SOC_HWFMT_NB_IF:
			iface |= 0x0010;
			break;
	}
	
	if (wm8731_write(codec, WM8731_ACTIVE, 0x00) != 0)
		return -EIO;

	/* set rate */
	wm8731_write(codec, WM8731_SRATE, pcm_c->hw_runtime.priv1);
	wm8731_write(codec, WM8731_IFACE, iface);

	return 0;
}

static void wm8731_device_release(struct device * dev)
{
}

static int wm8731_dpm_event(struct snd_soc_codec *codec, int event)
{
	switch (event) {
		case SNDRV_CTL_POWER_D0: /* full On */
			/* vref/mid, clk and osc on, dac unmute, active */
			wm8731_write(codec, WM8731_ACTIVE, 0x0001);
			break;
		case SNDRV_CTL_POWER_D1: /* partial On */
		case SNDRV_CTL_POWER_D2: /* partial On */
			break;
		case SNDRV_CTL_POWER_D3hot: /* Off, with power */
			/* everything off except vref/vmid, dac mute, inactive */
			wm8731_write(codec, WM8731_ACTIVE, 0x0);
			wm8731_write(codec, WM8731_PWR, 0xff7f);
			break;
		case SNDRV_CTL_POWER_D3cold: /* Off, without power */
			/* everything off, dac mute, inactive */
			wm8731_write(codec, WM8731_ACTIVE, 0x0);
			wm8731_write(codec, WM8731_PWR, 0xffff);
			break;
	}

	return 0;
}

static struct snd_soc_pcm_codec wm8731_pcm_client = {
	.name = "WM8731",
	.playback = {
		.sname = "Playback",
		.rates = SNDRV_PCM_RATE_8000 | SNDRV_PCM_RATE_16000 |
		SNDRV_PCM_RATE_32000 | SNDRV_PCM_RATE_48000 | SNDRV_PCM_RATE_44100 |
		SNDRV_PCM_RATE_96000,
		.formats = SNDRV_PCM_FMTBIT_S16_LE,
		.channels_min = 1,
		.channels_max = 2,
	},
	.nplayback = 1,
	.capture = {
		.sname = "Capture",
		.rates = SNDRV_PCM_RATE_8000 | SNDRV_PCM_RATE_16000 | 
		SNDRV_PCM_RATE_32000 | SNDRV_PCM_RATE_48000 | SNDRV_PCM_RATE_44100 |
		SNDRV_PCM_RATE_96000,
		.formats = SNDRV_PCM_FMTBIT_S16_LE,
		.channels_min = 1,
		.channels_max = 2,
	},
	.ncapture = 1,
	.ops = {
		.prepare = wm8731_pcm_prepare,
	},
	.hw = {
		.num_hmodes = ARRAY_SIZE(wm8731_hwfmt),
		.hmodes = &wm8731_hwfmt[0],
	},
};

/*
 * initialise the WM8731 driver
 * register the mixer and dsp interfaces with the kernel 
 */
static int wm8731_probe(struct snd_soc_codec *codec)
{
	int reg;
	
	wm8731_reset(codec);

	wm8731_dpm_event(codec, SNDRV_CTL_POWER_D3hot);
	
	/* set the update bits */
	reg = wm8731_read_reg_cache(codec, WM8731_LOUT1V);
	wm8731_write(codec, WM8731_LOUT1V, reg | 0x0100);
	reg = wm8731_read_reg_cache(codec, WM8731_ROUT1V);
	wm8731_write(codec, WM8731_ROUT1V, reg | 0x0100);
	reg = wm8731_read_reg_cache(codec, WM8731_LINVOL);
	wm8731_write(codec, WM8731_LINVOL, reg | 0x0100);
	reg = wm8731_read_reg_cache(codec, WM8731_RINVOL);
	wm8731_write(codec, WM8731_RINVOL, reg | 0x0100);

	wm8731_add_controls(codec);
	wm8731_add_widgets(codec);

	return 0;
}

/* 
 * unregister interfaces and clean up
 */
static void wm8731_remove(struct snd_soc_codec *codec)
{
	wm8731_dpm_event(codec, SNDRV_CTL_POWER_D3cold);
}

/*
 * Attach WM8731 2 wire client 
 */
static int wm8731_i2c_probe(struct i2c_adapter *adap, int addr, int kind)
{		
	int ret = 0;
	struct snd_soc_codec *codec;
	struct i2c_client *i2c;

	client_template.adapter = adap;
	client_template.addr = addr;
	
	if ((codec = kzalloc(sizeof(struct snd_soc_codec), GFP_KERNEL)) == NULL)
		return -ENOMEM;
 
	if ((i2c = kzalloc(sizeof(struct i2c_client), GFP_KERNEL)) == NULL){
		kfree(codec);
		return -ENOMEM;
	}
	memcpy(i2c, &client_template, sizeof(struct i2c_client));
	codec->control_data = i2c;
	i2c->dev.platform_data = codec;

	if((ret = i2c_attach_client(i2c)) < 0) {
		err("failed to attach codec at addr %x\n", addr);
		kfree(i2c);
		kfree(codec);
		return ret;
	}

	codec->name = "WM8731";
	codec->longname = "WM8731 Portable Codec";
	codec->owner = THIS_MODULE;
	codec->pdev.name = "alsa-wm8731";
	codec->pdev.id = -1;
	codec->pdev.dev.release = wm8731_device_release;
	codec->probe = wm8731_probe;
	codec->remove = wm8731_remove;
	codec->read = wm8731_read_reg_cache;
	codec->write = wm8731_write;
	codec->dpm_event = wm8731_dpm_event;
	codec->pcms = &wm8731_pcm_client;
	codec->npcms = 1;
	if ((codec->reg_cache = 
			kzalloc(sizeof(u16) * ARRAY_SIZE(wm8731_reg), GFP_KERNEL)) == NULL)
		return -ENOMEM;
	memcpy(codec->reg_cache, wm8731_reg, sizeof(u16) * ARRAY_SIZE(wm8731_reg));

	if((ret = snd_soc_register_codec(codec)) < 0) {
		err("can't register codec");
		i2c_del_driver(&wm8731_i2c_driver);
		kfree(i2c);
		kfree(codec->reg_cache);
		kfree(codec);
	}

	return ret;
}

static int wm8731_i2c_detach(struct i2c_client *client)
{
	struct snd_soc_codec *codec = 
		(struct snd_soc_codec*)client->dev.platform_data;

	i2c_detach_client(client);
	snd_soc_unregister_codec(codec);
	kfree(codec->reg_cache);
	kfree(codec->control_data);
	kfree(codec);
	return 0;
}

static int wm8731_i2c_attach(struct i2c_adapter *adap)
{
	return i2c_probe(adap, &addr_data, wm8731_i2c_probe);
}

/* WM8731 2 Wire layer */ 
static struct i2c_driver wm8731_i2c_driver = {
	.name =           "i2c wm8731 driver",
	.id =             I2C_DRIVERID_WM8731,
	.flags =          I2C_DF_NOTIFY,
	.attach_adapter = wm8731_i2c_attach,
	.detach_client =  wm8731_i2c_detach,
	.command =        NULL,
};

static struct i2c_client client_template = {
	.name =   "WM8731",
	.flags =  I2C_CLIENT_ALLOW_USE,
	.driver = &wm8731_i2c_driver,
};

static int __init wm8731_init(void)
{	
	int ret = 0;
	
	info("WM8731 Audio Codec %s", WM8731_VERSION);

	if ((ret = i2c_add_driver(&wm8731_i2c_driver)) != 0)
		err("can't add i2c driver");
	
	return ret;
}

static void __exit wm8731_exit(void)
{
	i2c_del_driver(&wm8731_i2c_driver);
}

module_init(wm8731_init);
module_exit(wm8731_exit);

MODULE_DESCRIPTION("Soc WM8731 driver");
MODULE_AUTHOR("Richard Purdie");
MODULE_LICENSE("GPL");
