/*
 * wm8753.c  --  WM8753 Soc Audio driver
 *
 * Copyright 2003 Wolfson Microelectronics PLC.
 * Author: Liam Girdwood
 *         liam.girdwood@wolfsonmicro.com or linux@wolfsonmicro.com
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  either version 2 of the  License, or (at your
 *  option) any later version.
 *
 * Notes:
 *  The WM8753 is a low power, high quality stereo codec with integrated PCM
 *  codec designed for portable digital telephony applications.
 */

#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/i2c.h>
#include <linux/pm.h>
#include <sound/driver.h>
#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/pcm_params.h>
#include <sound/soc.h>
#include <sound/soc-dpm.h>
#include <sound/initval.h>

#include "wm8753.h"

#define AUDIO_NAME "wm8753"
#define WM8753_VERSION "0.5"

/*
 * Debug
 */
 
#define PFX AUDIO_NAME
#define WM8753_DEBUG 0

#ifdef WM8753_DEBUG
#define dbg(format, arg...) printk(KERN_DEBUG PFX ": " format "\n" , ## arg)
#else
#define dbg(format, arg...) do {} while (0)
#endif
#define err(format, arg...) printk(KERN_ERR PFX ": " format "\n" , ## arg)
#define info(format, arg...) printk(KERN_INFO PFX ": " format "\n" , ## arg)
#define warn(format, arg...) printk(KERN_WARNING PFX ": " format "\n" , ## arg)



/* 
 * WM8753 2 wire address is determined by GPIO5
 * state during powerup.
 *    low  = WM8753_2W_ADDR1
 *    high = WM8753_2W_ADDR2
 */
#define WM8753_2W_ADDR1	0x1a
#define WM8753_2W_ADDR2	0x1b
#define I2C_DRIVERID_WM8753 0xfefe /* liam -  need a proper id */

static unsigned short normal_i2c[] = { WM8753_2W_ADDR1, WM8753_2W_ADDR2,
                                       I2C_CLIENT_END };

/* Magic definition of all other variables and things */
I2C_CLIENT_INSMOD;

static struct i2c_driver wm8753_i2c_driver;
static struct i2c_client client_template;

/*
 * wm8753 register cache
 * We can't read the WM8753 register space when we 
 * are using 2 wire for device control, so we cache them instead. 
 */
static const u16 wm8753_reg[] = {
    0x0008, 0x0000, 0x000a, 0x000a,  
    0x0033, 0x0000, 0x0007, 0x00ff,
    0x00ff, 0x000f, 0x000f, 0x007b, 
    0x0000, 0x0032, 0x0000, 0x00c3,
    0x00c3, 0x00c0, 0x0000, 0x0000,
    0x0000, 0x0000, 0x0000, 0x0000,
    0x0000, 0x0000, 0x0000, 0x0000,
    0x0000, 0x0000, 0x0000, 0x0055, 
	0x0005, 0x0050, 0x0055, 0x0050,
	0x0055, 0x0050, 0x0055, 0x0079, 
	0x0079, 0x0079, 0x0079, 0x0079,
	0x0000, 0x0000, 0x0000, 0x0000, 
	0x0097, 0x0097, 0x0000, 0x0004,
	0x0000, 0x0083, 0x0024, 0x01ba,
	0x0000, 0x0083, 0x0024, 0x01ba, 
	0x0000, 0x0000
};

/* PLL divisors */
struct _pll_div {
	u32 hz;
	u32 n;
	u32 k;
};

/* 
 * MCLCK crystal frequency
 */
#define MCLK_13MHZ

static struct _pll_div pll_div[] = {
#ifdef MCLK_13MHZ	
   {12288000,    0x7,    0x23F54A},
   {11289600,    0x6,    0x3CA2F5},
#endif
#ifdef MCLK_12MHZ	
   {12288000,    0x8,    0x0C49BA},
   {11289600,    0x7,    0x21B08A},
#endif
#ifdef MCLK_12_288MHZ
   {12288000,	0x0,    0x000000},
   {11289600,	0x7,	0x166667},
#endif
   {0,           0,      0}  /* must be last */
};

#define WM8753_HIFI_HWFMT \
	(SND_SOC_HWFMT_I2S | SND_SOC_HWFMT_LEFT_J | SND_SOC_HWFMT_RIGHT_J | \
	SND_SOC_HWFMT_CBM_CFM | SND_SOC_HWFMT_CBM_CFS | SND_SOC_HWFMT_CBS_CFS | \
	SND_SOC_HWFMT_NB_NF | SND_SOC_HWFMT_NB_IF | SND_SOC_HWFMT_IB_NF | \
	SND_SOC_HWFMT_IB_IF)

/*
 * priv1 is pll/mclk freq
 * priv2 is bits[15..8] = MCLK divisor, bits[7..0] = sr
 */
static struct snd_soc_hw_mode wm8753_hifi[] = {
	{WM8753_HIFI_HWFMT, 	SND_SOC_HWBITS(16),		8000,	768,	12288000, 	0x020c},
	{WM8753_HIFI_HWFMT, 	SND_SOC_HWBITS(16),		8000,	384,	12288000, 	0x040c},
	{WM8753_HIFI_HWFMT, 	SND_SOC_HWBITS(16),		8000,	192,	12288000, 	0x080c},
	{WM8753_HIFI_HWFMT, 	SND_SOC_HWBITS(16),		8000,	96,		12288000, 	0x100c},
	
	{WM8753_HIFI_HWFMT, 	SND_SOC_HWBITS(16),		11025,	512,	11289600,	0x0230},
	{WM8753_HIFI_HWFMT, 	SND_SOC_HWBITS(16),		11025,	256,	11289600,	0x0430},
	{WM8753_HIFI_HWFMT, 	SND_SOC_HWBITS(16),		11025, 	128,	11289600,	0x0830},
	{WM8753_HIFI_HWFMT, 	SND_SOC_HWBITS(16),		11025,	64,		11289600,	0x1030},
	
	{WM8753_HIFI_HWFMT, 	SND_SOC_HWBITS(16),		12000, 	512,	12288000,   0x0210},
	{WM8753_HIFI_HWFMT, 	SND_SOC_HWBITS(16),		12000, 	256,	12288000,   0x0410},
	{WM8753_HIFI_HWFMT, 	SND_SOC_HWBITS(16),		12000, 	128,	12288000,   0x0810},
	{WM8753_HIFI_HWFMT, 	SND_SOC_HWBITS(16),		12000, 	64,		12288000,   0x1010},
	
	{WM8753_HIFI_HWFMT, 	SND_SOC_HWBITS(16),		16000, 	384,	12288000,   0x0214},
	{WM8753_HIFI_HWFMT, 	SND_SOC_HWBITS(16),		16000, 	192,	12288000,   0x0414},
	{WM8753_HIFI_HWFMT, 	SND_SOC_HWBITS(16),		16000, 	96,		12288000,   0x0814},
	{WM8753_HIFI_HWFMT, 	SND_SOC_HWBITS(16),		16000, 	48, 	12288000,   0x1014},

	{WM8753_HIFI_HWFMT, 	SND_SOC_HWBITS(16),		22050, 	512,	11289600,   0x0034},
	{WM8753_HIFI_HWFMT, 	SND_SOC_HWBITS(16),		22050, 	256,	11289600,   0x0234},
	{WM8753_HIFI_HWFMT, 	SND_SOC_HWBITS(16),		22050, 	128,	11289600,   0x0434},
	{WM8753_HIFI_HWFMT, 	SND_SOC_HWBITS(16),		22050, 	64,		11289600,   0x0834},
	{WM8753_HIFI_HWFMT, 	SND_SOC_HWBITS(16),		22050, 	32,		11289600,   0x1034},
	
	{WM8753_HIFI_HWFMT, 	SND_SOC_HWBITS(16),		24000, 	512,	12288000,   0x0038},
	{WM8753_HIFI_HWFMT, 	SND_SOC_HWBITS(16),		24000, 	256,	12288000,   0x0238},
	{WM8753_HIFI_HWFMT, 	SND_SOC_HWBITS(16),		24000, 	128,	12288000,   0x0438},
	{WM8753_HIFI_HWFMT, 	SND_SOC_HWBITS(16),		24000, 	64,		12288000,   0x0838},
	{WM8753_HIFI_HWFMT, 	SND_SOC_HWBITS(16),		24000, 	32,		12288000,   0x1038},
	
	{WM8753_HIFI_HWFMT, 	SND_SOC_HWBITS(16),		32000, 	384,	12288000,	0x0018},
	{WM8753_HIFI_HWFMT, 	SND_SOC_HWBITS(16),		32000, 	192,	12288000,	0x0218},
	{WM8753_HIFI_HWFMT, 	SND_SOC_HWBITS(16),		32000, 	96,		12288000,	0x0418},
	{WM8753_HIFI_HWFMT, 	SND_SOC_HWBITS(16),		32000, 	48,		12288000,	0x0818},
	
	{WM8753_HIFI_HWFMT, 	SND_SOC_HWBITS(16),		44100, 	256,	11289600,  	0x0020},
	{WM8753_HIFI_HWFMT, 	SND_SOC_HWBITS(16),		44100, 	128,	11289600,  	0x0220},
	{WM8753_HIFI_HWFMT, 	SND_SOC_HWBITS(16),		44100, 	64,		11289600,  	0x0420},
	{WM8753_HIFI_HWFMT, 	SND_SOC_HWBITS(16),		44100, 	32,		11289600,  	0x0820},
	
	{WM8753_HIFI_HWFMT, 	SND_SOC_HWBITS(16),		48000, 	256,	12288000,  	0x0000},
	{WM8753_HIFI_HWFMT, 	SND_SOC_HWBITS(16),		48000, 	128,	12288000,  	0x0200},
	{WM8753_HIFI_HWFMT, 	SND_SOC_HWBITS(16),		48000, 	64,		12288000,  	0x0400},
	{WM8753_HIFI_HWFMT, 	SND_SOC_HWBITS(16),		48000, 	32,		12288000,  	0x0800},
	
	{WM8753_HIFI_HWFMT, 	SND_SOC_HWBITS(16),		88200, 	128,	11289600,   0x003c},
	{WM8753_HIFI_HWFMT, 	SND_SOC_HWBITS(16),		88200, 	64,		11289600,   0x023c},
	{WM8753_HIFI_HWFMT, 	SND_SOC_HWBITS(16),		88200, 	32,		11289600,   0x043c},
	
	{WM8753_HIFI_HWFMT, 	SND_SOC_HWBITS(16),		96000, 	128,	12288000,   0x001c},
	{WM8753_HIFI_HWFMT, 	SND_SOC_HWBITS(16),		96000, 	64,		12288000,   0x021c},
	{WM8753_HIFI_HWFMT, 	SND_SOC_HWBITS(16),		96000, 	32,		12288000,   0x041c},
};

#define WM8753_VOICE_HWFMT \
	(SND_SOC_HWFMT_I2S | SND_SOC_HWFMT_LEFT_J | SND_SOC_HWFMT_RIGHT_J | \
	SND_SOC_HWFMT_DSP_A | SND_SOC_HWFMT_DSP_B | SND_SOC_HWFMT_CBM_CFM | \
	SND_SOC_HWFMT_CBM_CFS | SND_SOC_HWFMT_CBS_CFS | SND_SOC_HWFMT_NB_NF | \
	SND_SOC_HWFMT_NB_IF | SND_SOC_HWFMT_IB_NF | SND_SOC_HWFMT_IB_IF)

/*
 * priv1 is pll/mclk freq
 * priv2 is bits[15..8] = PCM divisor, bits[7..0] = psr
 */
static struct snd_soc_hw_mode wm8753_voice[] = {
	{WM8753_VOICE_HWFMT, 	SND_SOC_HWBITS(16),		48000,	256,	12288000, 	0x0000},
	{WM8753_VOICE_HWFMT, 	SND_SOC_HWBITS(16),		44100,	256,	11289600, 	0x0000},
	{WM8753_VOICE_HWFMT, 	SND_SOC_HWBITS(16),		32000,	384,	12288000, 	0x0001},
	{WM8753_VOICE_HWFMT, 	SND_SOC_HWBITS(16),		24000,	256,	12288000, 	0x0400},
	{WM8753_VOICE_HWFMT, 	SND_SOC_HWBITS(16),		22050,	256,	11289600, 	0x0400},
	{WM8753_VOICE_HWFMT, 	SND_SOC_HWBITS(16),		16000,	256,	12288000, 	0x0200},
	{WM8753_VOICE_HWFMT, 	SND_SOC_HWBITS(16),		16000,	384,	12288000, 	0x0401},
	{WM8753_VOICE_HWFMT, 	SND_SOC_HWBITS(16),		8000,	256,	12288000, 	0x0600},
	{WM8753_VOICE_HWFMT, 	SND_SOC_HWBITS(16),		8000,	384,	12288000, 	0x0501},
};

/*
 * read wm8753 register cache
 */
static inline unsigned int wm8753_read_reg_cache(struct snd_soc_codec * codec, 
	unsigned int reg)
{
	u16 *cache = codec->reg_cache;
	if(reg < 1 || reg > (ARRAY_SIZE(wm8753_reg) + 1))
		return -1;
	return cache[reg - 1];
}

/*
 * write wm8753 register cache
 */
static inline void wm8753_write_reg_cache(struct snd_soc_codec * codec, 
	unsigned int reg, unsigned int value)
{
	u16 *cache = codec->reg_cache;
	if(reg < 1 || reg > (ARRAY_SIZE(wm8753_reg) + 1))
		return;
	cache[reg - 1] = value;
}

static struct i2c_client client_template;

/*
 * write to the WM8753 register space
 */
static int wm8753_write(struct snd_soc_codec * codec, unsigned int reg,
	unsigned int value)
{
	u8 data[2];
	struct i2c_client *i2c = (struct i2c_client*)codec->control_data;
	/* data is 
	 *   D15..D9 WM8753 register offset
	 *   D8...D0 register data
	 */
	data[0] = (reg << 1) | ((value >> 8) & 0x0001);
	data[1] = value & 0x00ff;

	wm8753_write_reg_cache (codec, reg, value);
	if (i2c_master_send(i2c, data, 2) == 2)
		return 0;
	else
		return -1;
}

#define wm8753_reset(c) wm8753_write(c, WM8753_RESET, 0)


/*
 * WM8753 Controls
 */
static const char* wm8753_base[] = {"Linear Control", "Adaptive Boost"};
static const char* wm8753_base_filter[] = 
	{"130Hz @ 48kHz", "200Hz @ 48kHz", "100Hz @ 16kHz", "400Hz @ 48kHz", 
	"100Hz @ 8kHz", "200Hz @ 8kHz"};
static const char* wm8753_treble[] = {"8kHz", "4kHz"};
static const char* wm8753_alc_func[] = {"Off", "Right", "Left", "Stereo"};
static const char* wm8753_ng_type[] = {"Constant PGA Gain", "Mute ADC Output"};
static const char* wm8753_3d_func[] = {"Capture", "Playback"};
static const char* wm8753_3d_uc[] = {"2.2kHz", "1.5kHz"};
static const char* wm8753_3d_lc[] = {"200Hz", "500Hz"};
static const char* wm8753_deemp[] = {"None", "32kHz", "44.1kHz", "48kHz"};
static const char* wm8753_mono_mix[] = {"Stereo", "Left", "Right", "Mono"};
static const char* wm8753_dac_phase[] = {"Non Inverted", "Inverted"};
static const char* wm8753_line_mix[] = {"Line 1 + 2", "Line 1 - 2", 
	"Line 1", "Line 2"};
static const char* wm8753_mono_mux[] = {"Line Mix", "Rx Mix"};
static const char* wm8753_right_mux[] = {"Line 2", "Rx Mix"};
static const char* wm8753_left_mux[] = {"Line 1", "Rx Mix"};
static const char* wm8753_rxmsel[] = {"RXP - RXN", "RXP + RXN", "RXP", "RXN"};
static const char* wm8753_sidetone_mux[] = {"Left PGA", "Mic 1", "Mic 2", "Right PGA"};
static const char* wm8753_mono2_src[] = {"Inverted Mono 1", "Left", "Right", "Left + Right"};
static const char* wm8753_out3[] = {"VREF", "ROUT2", "Left + Right"};
static const char* wm8753_out4[] = {"VREF", "Capture ST", "LOUT2"};
static const char* wm8753_radcsel[] = {"PGA", "Line or RXP-RXN", "Sidetone"};
static const char* wm8753_ladcsel[] = {"PGA", "Line or RXP-RXN", "Line"};
static const char* wm8753_mono_adc[] = {"Stereo", "Analogue Mix Left", "Analogue Mix Right",
	"Digital Mono Mix"};
static const char* wm8753_adc_hp[] = {"3.4Hz @ 48kHz", "82Hz @ 16k", "82Hz @ 8kHz", "170Hz @ 8kHz"};
static const char* wm8753_adc_filter[] = {"HiFi", "Voice"};
static const char* wm8753_mic_sel[] = {"Mic 1", "Mic 2", "Mic 3"};

static const struct soc_enum wm8753_enum[] = {
SOC_ENUM_SINGLE(WM8753_BASS, 7, 2, wm8753_base),		// 0
SOC_ENUM_SINGLE(WM8753_BASS, 4, 6, wm8753_base_filter),	// 1
SOC_ENUM_SINGLE(WM8753_TREBLE, 6, 2, wm8753_treble), 	// 2
SOC_ENUM_SINGLE(WM8753_ALC1, 7, 4, wm8753_alc_func),	// 3
SOC_ENUM_SINGLE(WM8753_NGATE, 1, 2, wm8753_ng_type),	// 4
SOC_ENUM_SINGLE(WM8753_3D, 7, 2, wm8753_3d_func),		// 5
SOC_ENUM_SINGLE(WM8753_3D, 6, 2, wm8753_3d_uc),			// 6
SOC_ENUM_SINGLE(WM8753_3D, 5, 2, wm8753_3d_lc),			// 7
SOC_ENUM_SINGLE(WM8753_DAC, 1, 4, wm8753_deemp),		// 8
SOC_ENUM_SINGLE(WM8753_DAC, 4, 4, wm8753_mono_mix),		// 9
SOC_ENUM_SINGLE(WM8753_DAC, 6, 2, wm8753_dac_phase),	// 10
SOC_ENUM_SINGLE(WM8753_INCTL1, 3, 4, wm8753_line_mix),	// 11
SOC_ENUM_SINGLE(WM8753_INCTL1, 2, 2, wm8753_mono_mux),	// 12
SOC_ENUM_SINGLE(WM8753_INCTL1, 1, 2, wm8753_right_mux),	// 13
SOC_ENUM_SINGLE(WM8753_INCTL1, 0, 2, wm8753_left_mux),	// 14
SOC_ENUM_SINGLE(WM8753_INCTL2, 6, 4, wm8753_rxmsel),	// 15
SOC_ENUM_SINGLE(WM8753_INCTL2, 4, 4, wm8753_sidetone_mux),// 16
SOC_ENUM_SINGLE(WM8753_OUTCTL, 7, 4, wm8753_mono2_src),	// 17
SOC_ENUM_SINGLE(WM8753_OUTCTL, 0, 3, wm8753_out3),		// 18
SOC_ENUM_SINGLE(WM8753_ADCTL2, 7, 3, wm8753_out4),		// 19
SOC_ENUM_SINGLE(WM8753_ADCIN, 2, 3, wm8753_radcsel),	// 20
SOC_ENUM_SINGLE(WM8753_ADCIN, 0, 3, wm8753_ladcsel),	// 21
SOC_ENUM_SINGLE(WM8753_ADCIN, 4, 4, wm8753_mono_adc),	// 22
SOC_ENUM_SINGLE(WM8753_ADC, 2, 4, wm8753_adc_hp),		// 23
SOC_ENUM_SINGLE(WM8753_ADC, 4, 2, wm8753_adc_filter),	// 24
SOC_ENUM_SINGLE(WM8753_MICBIAS, 6, 3, wm8753_mic_sel),	// 25
};

static const snd_kcontrol_new_t wm8753_snd_controls[] = {
SOC_SINGLE("Left PCM Volume", WM8753_LDAC, 0, 255, 0),
SOC_SINGLE("Right PCM Volume", WM8753_RDAC, 0, 255, 0),
	
SOC_SINGLE("Left ADC Capture Volume", WM8753_LADC, 0, 63, 0),
SOC_SINGLE("Left ADC Capture Switch", WM8753_LADC, 7, 1, 0),
SOC_SINGLE("Left ADC Capture ZC Switch", WM8753_LADC, 6, 1, 0),
SOC_SINGLE("Right ADC Capture Volume", WM8753_RADC, 0, 63, 0),
SOC_SINGLE("Right ADC Capture Switch", WM8753_RADC, 7, 1, 0),
SOC_SINGLE("Right ADC Capture ZC Switch", WM8753_RADC, 6, 1, 0),

SOC_SINGLE("Left Out1 Playback ZC Switch", WM8753_LOUT1V, 7, 1, 0),
SOC_SINGLE("Left Out2 Playback ZC Switch", WM8753_LOUT2V, 7, 1, 0),
	
SOC_SINGLE("Left Bypass Playback Volume", WM8753_LOUTM1, 4, 7, 1),
SOC_SINGLE("Left Sidetone Playback Volume", WM8753_LOUTM2, 4, 7, 1),
SOC_SINGLE("Left Voice Playback Volume", WM8753_LOUTM2, 0, 7, 1),

SOC_SINGLE("Right Out1 Playback ZC Switch", WM8753_ROUT1V, 7, 1, 0),
SOC_SINGLE("Right Out2 Playback ZC Switch", WM8753_ROUT2V, 7, 1, 0),

SOC_SINGLE("Right Bypass Playback Volume", WM8753_ROUTM1, 4, 7, 1),
SOC_SINGLE("Right Sidetone Playback Volume", WM8753_ROUTM2, 4, 7, 1),
SOC_SINGLE("Right Voice Playback Volume", WM8753_ROUTM2, 0, 7, 1),

SOC_SINGLE("Mono Bypass Playback Volume", WM8753_MOUTM1, 4, 7, 1),
SOC_SINGLE("Mono Sidetone Playback Volume", WM8753_MOUTM2, 4, 7, 1),
SOC_SINGLE("Mono Voice Playback Volume", WM8753_MOUTM2, 4, 7, 1),

SOC_SINGLE("Mono Playback ZC Switch", WM8753_MOUTV, 7, 1, 0),

SOC_ENUM("Bass Boost", wm8753_enum[0]),
SOC_ENUM("Bass Filter", wm8753_enum[1]),
SOC_SINGLE("Bass Volume", WM8753_BASS, 0, 7, 1),

SOC_SINGLE("Treble Volume", WM8753_TREBLE, 0, 7, 0),
SOC_ENUM("Treble Cut-off", wm8753_enum[2]),

SOC_SINGLE("Right Sidetone Capture Volume", WM8753_RECMIX1, 4, 7, 1),
SOC_SINGLE("Left Sidetone Capture Volume", WM8753_RECMIX1, 0, 7, 1),
SOC_SINGLE("Voice Sidetone Capture Volume", WM8753_RECMIX2, 0, 7, 1),

SOC_SINGLE("Left Capture Volume", WM8753_LINVOL, 0, 63, 0),
SOC_SINGLE("Left Capture ZC Switch", WM8753_LINVOL, 6, 1, 0),
SOC_SINGLE("Left Capture Switch", WM8753_LINVOL, 7, 1, 1),
SOC_SINGLE("Right Capture Volume", WM8753_RINVOL, 0, 63, 0),
SOC_SINGLE("Right Capture ZC Switch", WM8753_RINVOL, 6, 1, 0),
SOC_SINGLE("Right Capture Switch", WM8753_RINVOL, 7, 1, 1),

SOC_ENUM("Capture Filter Select", wm8753_enum[23]),
SOC_ENUM("Capture Filter Cut-off", wm8753_enum[24]),
SOC_SINGLE("Capture Filter Switch", WM8753_ADC, 0, 1, 1),

SOC_SINGLE("ALC Capture Target Volume", WM8753_ALC1, 0, 7, 0),
SOC_SINGLE("ALC Capture Max Volume", WM8753_ALC1, 4, 7, 0),
SOC_ENUM("ALC Capture Function", wm8753_enum[3]),
SOC_SINGLE("ALC Capture ZC Switch", WM8753_ALC2, 8, 1, 0),
SOC_SINGLE("ALC Capture Hold Time", WM8753_ALC2, 0, 15, 1),
SOC_SINGLE("ALC Capture Decay Time", WM8753_ALC3, 4, 15, 1),
SOC_SINGLE("ALC Capture Attack Time", WM8753_ALC3, 0, 15, 0),
SOC_SINGLE("ALC Capture NG Threshold", WM8753_NGATE, 3, 31, 0),
SOC_ENUM("ALC Capture NG Type", wm8753_enum[4]),
SOC_SINGLE("ALC Capture NG Switch", WM8753_NGATE, 0, 1, 0),

SOC_ENUM("3D Function", wm8753_enum[5]),
SOC_ENUM("3D Upper Cut-off", wm8753_enum[6]),
SOC_ENUM("3D Lower Cut-off", wm8753_enum[7]),
SOC_SINGLE("3D Volume", WM8753_3D, 1, 15, 0),
SOC_SINGLE("3D Switch", WM8753_3D, 0, 1, 0),

SOC_SINGLE("Capture 6dB Attenuate", WM8753_ADCTL1, 2, 1, 0),
SOC_SINGLE("Playback 6dB Attenuate", WM8753_ADCTL1, 1, 1, 0),

SOC_ENUM("De-emphasis", wm8753_enum[8]),
SOC_ENUM("Playback Mono Mix", wm8753_enum[9]),
SOC_ENUM("Playback Phase", wm8753_enum[10]),

SOC_SINGLE("Mic2 Capture Volume", WM8753_INCTL1, 7, 3, 0),
SOC_SINGLE("Mic1 Capture Volume", WM8753_INCTL1, 5, 3, 0),
};

/* add non dpm controls */
static int wm8753_add_controls(struct snd_soc_codec *codec)
{
	int err, i;

	for (i = 0; i < ARRAY_SIZE(wm8753_snd_controls); i++) {
		if ((err = snd_ctl_add(codec->card, snd_soc_cnew(&wm8753_snd_controls[i],codec, NULL))) < 0)
			return err;
	}
	return 0;
}

/*
 * DPM Controls
 */

/* Left Mixer */
static const snd_kcontrol_new_t wm8753_left_mixer_controls[] = {
SOC_DPM_SINGLE("Voice Playback Switch", WM8753_LOUTM2, 8, 1, 0),
SOC_DPM_SINGLE("Sidetone Playback Switch", WM8753_LOUTM2, 7, 1, 0),
SOC_DPM_SINGLE("Left Playback Switch", WM8753_LOUTM1, 8, 1, 0),
SOC_DPM_SINGLE("Bypass Playback Switch", WM8753_LOUTM1, 7, 1, 0),
};

/* Right mixer */
static const snd_kcontrol_new_t wm8753_right_mixer_controls[] = {
SOC_DPM_SINGLE("Voice Playback Switch", WM8753_ROUTM2, 8, 1, 0),
SOC_DPM_SINGLE("Sidetone Playback Switch", WM8753_ROUTM2, 7, 1, 0),
SOC_DPM_SINGLE("Right Playback Switch", WM8753_ROUTM1, 8, 1, 0),
SOC_DPM_SINGLE("Bypass Playback Switch", WM8753_ROUTM1, 7, 1, 0),
};

/* Mono mixer */
static const snd_kcontrol_new_t wm8753_mono_mixer_controls[] = {
SOC_DPM_SINGLE("Left Playback Switch", WM8753_MOUTM1, 8, 1, 0),
SOC_DPM_SINGLE("Right Playback Switch", WM8753_MOUTM2, 8, 1, 0),
SOC_DPM_SINGLE("Voice Playback Switch", WM8753_MOUTM2, 4, 1, 0),
SOC_DPM_SINGLE("Sidetone Playback Switch", WM8753_MOUTM2, 7, 1, 0),
SOC_DPM_SINGLE("Bypass Playback Switch", WM8753_ROUTM1, 7, 1, 0),
};

/* Mono 2 Mux */
static const snd_kcontrol_new_t wm8753_mono2_controls =
SOC_DPM_ENUM("Route", wm8753_enum[17]);

/* Out 3 Mux */
static const snd_kcontrol_new_t wm8753_out3_controls =
SOC_DPM_ENUM("Route", wm8753_enum[18]);

/* Out 4 Mux */
static const snd_kcontrol_new_t wm8753_out4_controls =
SOC_DPM_ENUM("Route", wm8753_enum[19]);

/* ADC Mono Mix */
static const snd_kcontrol_new_t wm8753_adc_mono_controls =
SOC_DPM_ENUM("Route", wm8753_enum[22]);

/* Record mixer */
static const snd_kcontrol_new_t wm8753_record_mixer_controls[] = {
SOC_DPM_SINGLE("Voice Capture Switch", WM8753_RECMIX2, 3, 1, 0),
SOC_DPM_SINGLE("Left Capture Switch", WM8753_RECMIX1, 3, 1, 0),
SOC_DPM_SINGLE("Right Capture Switch", WM8753_RECMIX1, 7, 1, 0),
};

/* Left ADC mux */
static const snd_kcontrol_new_t wm8753_adc_left_controls =
SOC_DPM_ENUM("Route", wm8753_enum[21]);

/* Right ADC mux */
static const snd_kcontrol_new_t wm8753_adc_right_controls =
SOC_DPM_ENUM("Route", wm8753_enum[20]);

/* MIC mux */
static const snd_kcontrol_new_t wm8753_mic_mux_controls =
SOC_DPM_ENUM("Route", wm8753_enum[16]);

/* ALC mixer */
static const snd_kcontrol_new_t wm8753_alc_mixer_controls[] = {
SOC_DPM_SINGLE("Line Capture Switch", WM8753_INCTL2, 3, 1, 0),
SOC_DPM_SINGLE("Mic2 Capture Switch", WM8753_INCTL2, 2, 1, 0),
SOC_DPM_SINGLE("Mic1 Capture Switch", WM8753_INCTL2, 1, 1, 0),
SOC_DPM_SINGLE("Rx Capture Switch", WM8753_INCTL2, 0, 1, 0),
};

/* Left Line mux */
static const snd_kcontrol_new_t wm8753_line_left_controls =
SOC_DPM_ENUM("Route", wm8753_enum[14]);

/* Right Line mux */
static const snd_kcontrol_new_t wm8753_line_right_controls =
SOC_DPM_ENUM("Route", wm8753_enum[13]);

/* Mono Line mux */
static const snd_kcontrol_new_t wm8753_line_mono_controls =
SOC_DPM_ENUM("Route", wm8753_enum[12]);

/* Line mux and mixer */
static const snd_kcontrol_new_t wm8753_line_mux_mix_controls =
SOC_DPM_ENUM("Route", wm8753_enum[11]);

/* Rx mux and mixer */
static const snd_kcontrol_new_t wm8753_rx_mux_mix_controls =
SOC_DPM_ENUM("Route", wm8753_enum[15]);

/* Mic Selector Mux */
static const snd_kcontrol_new_t wm8753_mic_sel_mux_controls =
SOC_DPM_ENUM("Route", wm8753_enum[25]);

/* LOUT 1 Volume */
static const snd_kcontrol_new_t wm8753_lout1_vol_controls =
SOC_DPM_SINGLE("Left Out1 Playback Volume", WM8753_LOUT1V, 0, 127, 0);

/* LOUT 2 Volume */
static const snd_kcontrol_new_t wm8753_lout2_vol_controls =
SOC_DPM_SINGLE("Left Out2 Playback Volume", WM8753_LOUT2V, 0, 127, 0);

/* ROUT 1 Volume */
static const snd_kcontrol_new_t wm8753_rout1_vol_controls =
SOC_DPM_SINGLE("Right Out1 Playback Volume", WM8753_ROUT1V, 0, 127, 0);

/* ROUT 2 Volume */
static const snd_kcontrol_new_t wm8753_rout2_vol_controls =
SOC_DPM_SINGLE("Right Out2 Playback Volume", WM8753_ROUT2V, 0, 127, 0);

/* Mono 1 Volume */
static const snd_kcontrol_new_t wm8753_mout1_vol_controls =
SOC_DPM_SINGLE("Mono Playback Volume", WM8753_MOUTV, 0, 127, 0);

static const struct snd_soc_dpm_widget wm8753_dpm_widgets[] = {
SND_SOC_DPM_CLOCK("WM8753 Clock", WM8753_PWR1, 0, 1),
SND_SOC_DPM_MUTE("DAC Mute", WM8753_DAC, 3, 0),
SND_SOC_DPM_MICBIAS("Mic Bias", WM8753_PWR1, 5, 0),
SND_SOC_DPM_MIXER("Left Mixer", WM8753_PWR4, 0, 0, &wm8753_left_mixer_controls[0], 
	ARRAY_SIZE(wm8753_left_mixer_controls)),
SND_SOC_DPM_VOLUME("Left Out 1", WM8753_PWR3, 8, 0, &wm8753_lout1_vol_controls, 1),
SND_SOC_DPM_VOLUME("Left Out 2", WM8753_PWR3, 6, 0, &wm8753_lout2_vol_controls, 1),
SND_SOC_DPM_DAC("Left DAC", "Left HiFi Playback", WM8753_PWR1, 3, 0, NULL, 0),
SND_SOC_DPM_OUTPUT("LOUT1"),
SND_SOC_DPM_OUTPUT("LOUT2"),
SND_SOC_DPM_MIXER("Right Mixer", WM8753_PWR4, 1, 0, &wm8753_right_mixer_controls[0], 
	ARRAY_SIZE(wm8753_right_mixer_controls)),
SND_SOC_DPM_VOLUME("Right Out 1", WM8753_PWR3, 7, 0, &wm8753_rout1_vol_controls, 1),
SND_SOC_DPM_VOLUME("Right Out 2", WM8753_PWR3, 5, 0, &wm8753_rout2_vol_controls, 1),
SND_SOC_DPM_DAC("Right DAC", "Right HiFi Playback", WM8753_PWR1, 2, 0, NULL, 0),
SND_SOC_DPM_OUTPUT("ROUT1"),
SND_SOC_DPM_OUTPUT("ROUT2"),
SND_SOC_DPM_MIXER("Mono Mixer", WM8753_PWR4, 2, 0, &wm8753_mono_mixer_controls[0], 
	ARRAY_SIZE(wm8753_mono_mixer_controls)),
SND_SOC_DPM_VOLUME("Mono Out 1", WM8753_PWR3, 2, 0, &wm8753_mout1_vol_controls, 1),
SND_SOC_DPM_VOLUME("Mono Out 2", WM8753_PWR3, 1, 0, NULL, 0),
SND_SOC_DPM_DAC("Voice DAC", "Voice Playback", WM8753_PWR1, 4, 0, NULL, 0),
SND_SOC_DPM_OUTPUT("MONO1"),
SND_SOC_DPM_MUX("Mono 2 Mux", -1, 0, 0, &wm8753_mono2_controls),
SND_SOC_DPM_OUTPUT("MONO2"),
SND_SOC_DPM_MIXER("Out3 Left + Right", -1, 0, 0, NULL, 0),
SND_SOC_DPM_MUX("Out3 Mux", -1, 0, 0, &wm8753_out3_controls),
SND_SOC_DPM_VOLUME("Out 3", WM8753_PWR3, 4, 0, NULL, 0),
SND_SOC_DPM_OUTPUT("OUT3"),
SND_SOC_DPM_MUX("Out4 Mux", -1, 0, 0, &wm8753_out4_controls),
SND_SOC_DPM_VOLUME("Out 4", WM8753_PWR3, 3, 0, NULL, 0),
SND_SOC_DPM_OUTPUT("OUT4"),
SND_SOC_DPM_MIXER("Playback Mixer", WM8753_PWR4, 3, 0, &wm8753_record_mixer_controls[0], 
	ARRAY_SIZE(wm8753_record_mixer_controls)),
SND_SOC_DPM_ADC("Left ADC", "Left Voice Capture", WM8753_PWR2, 3, 0, NULL, 0),
SND_SOC_DPM_ADC("Right ADC", "Right Voice Capture", WM8753_PWR2, 2, 0, NULL, 0),
SND_SOC_DPM_MUX("Capture Left Mixer", -1, 0, 0, &wm8753_adc_mono_controls),
SND_SOC_DPM_MUX("Capture Right Mixer", -1, 0, 0, &wm8753_adc_mono_controls),
SND_SOC_DPM_MUX("Capture Left Mux", -1, 0, 0, &wm8753_adc_left_controls),
SND_SOC_DPM_MUX("Capture Right Mux", -1, 0, 0, &wm8753_adc_right_controls),
SND_SOC_DPM_MUX("Mic Sidetone Mux", -1, 0, 0, &wm8753_mic_mux_controls),
SND_SOC_DPM_VOLUME("Left Capture Volume", WM8753_PWR2, 5, 0, NULL, 0),
SND_SOC_DPM_VOLUME("Right Capture Volume", WM8753_PWR2, 4, 0, NULL, 0),
SND_SOC_DPM_MIXER("ALC Mixer", WM8753_PWR2, 6, 0, &wm8753_alc_mixer_controls[0], 
	ARRAY_SIZE(wm8753_alc_mixer_controls)),
SND_SOC_DPM_MUX("Line Left Mux", -1, 0, 0, &wm8753_line_left_controls),
SND_SOC_DPM_MUX("Line Right Mux", -1, 0, 0, &wm8753_line_right_controls),
SND_SOC_DPM_MUX("Line Mono Mux", -1, 0, 0, &wm8753_line_mono_controls),
SND_SOC_DPM_MUX("Line Mixer", -1, 0, 0, &wm8753_line_mux_mix_controls),
SND_SOC_DPM_MUX("Rx Mixer", -1, 0, 0, &wm8753_rx_mux_mix_controls),
SND_SOC_DPM_VOLUME("Mic 1 Volume", WM8753_PWR2, 8, 0, NULL, 0),
SND_SOC_DPM_VOLUME("Mic 2 Volume", WM8753_PWR2, 7, 0, NULL, 0),
SND_SOC_DPM_MUX("Mic Selection Mux", -1, 0, 0, &wm8753_mic_sel_mux_controls),
SND_SOC_DPM_INPUT("LINE1"),
SND_SOC_DPM_INPUT("LINE2"),
SND_SOC_DPM_INPUT("RXP"),
SND_SOC_DPM_INPUT("RXN"),
SND_SOC_DPM_INPUT("ACIN"),
SND_SOC_DPM_INPUT("ACOP"),
SND_SOC_DPM_INPUT("MIC1N"),
SND_SOC_DPM_INPUT("MIC1"),
SND_SOC_DPM_INPUT("MIC2N"),
SND_SOC_DPM_INPUT("MIC2"),
};

static const char* intercon[][3] = {
	/* left mixer */
	{"Left Mixer", "Left Playback Switch", "Left DAC"},
	{"Left Mixer", "Voice Playback Switch", "Voice DAC"},
	{"Left Mixer", "Sidetone Playback Switch", "Mic Sidetone Mux"},
	{"Left Mixer", "Bypass Playback Switch", "Line Left Mux"},
	
	/* right mixer */
	{"Right Mixer", "Right Playback Switch", "Right DAC"},
	{"Right Mixer", "Voice Playback Switch", "Voice DAC"},
	{"Right Mixer", "Sidetone Playback Switch", "Mic Sidetone Mux"},
	{"Right Mixer", "Bypass Playback Switch", "Line Right Mux"},
	
	/* mono mixer */
	{"Mono Mixer", "Voice Playback Switch", "Voice DAC"},
	{"Mono Mixer", "Left Playback Switch", "Left DAC"},
	{"Mono Mixer", "Right Playback Switch", "Right DAC"},
	{"Mono Mixer", "Sidetone Playback Switch", "Mic Sidetone Mux"},
	{"Mono Mixer", "Bypass Playback Switch", "Line Mono Mux"},
	
	/* left out */
	{"Left Out 1", NULL, "Left Mixer"},
	{"Left Out 2", NULL, "Left Mixer"},
	{"LOUT1", NULL, "Left Out 1"},
	{"LOUT2", NULL, "Left Out 2"},

	/* right out */
	{"Right Out 1", NULL, "Right Mixer"},
	{"Right Out 2", NULL, "Right Mixer"},
	{"ROUT1", NULL, "Right Out 1"},
	{"ROUT2", NULL, "Right Out 2"},
	
	/* mono 1 out */
	{"Mono Out 1", NULL, "Mono Mixer"},
	{"MONO1", NULL, "Mono Out 1"},
	
	/* mono 2 out */
	{"Mono 2 Mux", "Left + Right", "Out3 Left + Right"},
	{"Mono 2 Mux", "Inverted Mono 1", "MONO1"},
	{"Mono 2 Mux", "Left", "Left Mixer"},
	{"Mono 2 Mux", "Right", "Right Mixer"},
	{"Mono Out 2", NULL, "Mono 2 Mux"},
	{"MONO2", NULL, "Mono Out 2"},
	
	/* out 3 */
	{"Out3 Left + Right", NULL, "Left Mixer"},
	{"Out3 Left + Right", NULL, "Right Mixer"},
	{"Out3 Mux", "VREF", "VREF"},
	{"Out3 Mux", "Left + Right", "Out3 Left + Right"},
	{"Out3 Mux", "ROUT2", "ROUT2"},
	{"Out 3", NULL, "Out3 Mux"},
	{"OUT3", NULL, "Out 3"},
	
	/* out 4 */
	{"Out4 Mux", "VREF", "VREF"},
	{"Out4 Mux", "Capture ST", "Capture ST Mixer"},
	{"Out4 Mux", "LOUT2", "LOUT2"},
	{"Out 4", NULL, "Out4 Mux"},
	{"OUT4", NULL, "Out 4"},
	
	/* record mixer  */
	{"Playback Mixer", "Left Capture Switch", "Left Mixer"},
	{"Playback Mixer", "Voice Capture Switch", "Mono Mixer"},
	{"Playback Mixer", "Right Capture Switch", "Right Mixer"},
	
	/* Mic/SideTone Mux */
	{"Mic Sidetone Mux", "Left PGA", "Left Capture Volume"},
	{"Mic Sidetone Mux", "Right PGA", "Right Capture Volume"},
	{"Mic Sidetone Mux", "Mic 1", "Mic 1 Volume"},
	{"Mic Sidetone Mux", "Mic 2", "Mic 2 Volume"},
	
	/* Capture Left Mux */
	{"Capture Left Mux", "PGA", "Left Capture Volume"},
	{"Capture Left Mux", "Line or RXP-RXN", "Line Left Mux"},
	{"Capture Left Mux", "Line", "LINE1"},
	
	/* Capture Right Mux */
	{"Capture Right Mux", "PGA", "Right Capture Volume"},
	{"Capture Right Mux", "Line or RXP-RXN", "Line Right Mux"},
	{"Capture Right Mux", "Sidetone", "Capture ST Mixer"},
	
	/* Mono Capture mixer-mux */
	{"Capture Right Mixer", "Stereo", "Capture Right Mux"},
	{"Capture Left Mixer", "Analogue Mix Left", "Capture Left Mux"},
	{"Capture Left Mixer", "Analogue Mix Left", "Capture Right Mux"},
	{"Capture Right Mixer", "Analogue Mix Right", "Capture Left Mux"},
	{"Capture Right Mixer", "Analogue Mix Right", "Capture Right Mux"},
	{"Capture Left Mixer", "Digital Mono Mix", "Capture Left Mux"},
	{"Capture Left Mixer", "Digital Mono Mix", "Capture Right Mux"},
	{"Capture Right Mixer", "Digital Mono Mix", "Capture Left Mux"},
	{"Capture Right Mixer", "Digital Mono Mix", "Capture Right Mux"},
	
	/* ADC */
	{"Left ADC", NULL, "Capture Left Mixer"},
	{"Right ADC", NULL, "Capture Right Mixer"},

	/* Left Capture Volume */
	{"Left Capture Volume", NULL, "ACIN"},
	
	/* Right Capture Volume */
	{"Right Capture Volume", NULL, "Mic 2 Volume"},

	/* ALC Mixer */
	{"ALC Mixer", "Line Capture Switch", "Line Mixer"},
	{"ALC Mixer", "Mic2 Capture Switch", "Mic 2 Volume"},
	{"ALC Mixer", "Mic1 Capture Switch", "Mic 1 Volume"},
	{"ALC Mixer", "Rx Capture Switch", "Rx Mixer"},
	
	/* Line Left Mux */
	{"Line Left Mux", "Line 1", "LINE1"},
	{"Line Left Mux", "Rx Mix", "Rx Mixer"},
	
	/* Line Right Mux */
	{"Line Right Mux", "Line 2", "LINE2"},
	{"Line Right Mux", "Rx Mix", "Rx Mixer"},
	
	/* Line Mono Mux */
	{"Line Mono Mux", "Line Mix", "Line Mixer"},
	{"Line Mono Mux", "Rx Mix", "Rx Mixer"},
	
	/* Line Mixer/Mux */
	{"Line Mixer", "Line 1 + 2", "LINE1"},
	{"Line Mixer", "Line 1 - 2", "LINE1"},
	{"Line Mixer", "Line 1 + 2", "LINE2"},
	{"Line Mixer", "Line 1 - 2", "LINE2"},
	{"Line Mixer", "Line 1", "LINE1"},
	{"Line Mixer", "Line 2", "LINE2"},
	
	/* Rx Mixer/Mux */
	{"Rx Mixer", "RXP - RXN", "RXP"},
	{"Rx Mixer", "RXP + RXN", "RXP"},
	{"Rx Mixer", "RXP - RXN", "RXN"},
	{"Rx Mixer", "RXP + RXN", "RXN"},
	{"Rx Mixer", "RXP", "RXP"},
	{"Rx Mixer", "RXN", "RXN"},
	
	/* Mic 1 Volume */
	{"Mic 1 Volume", NULL, "MIC1N"},
	{"Mic 1 Volume", NULL, "Mic Selection Mux"},
	
	/* Mic 2 Volume */
	{"Mic 2 Volume", NULL, "MIC2N"},
	{"Mic 2 Volume", NULL, "MIC2"},
	
	/* Mic Selector Mux */
	{"Mic Selection Mux", "Mic 1", "MIC1"},
	{"Mic Selection Mux", "Mic 2", "MIC2N"},
	{"Mic Selection Mux", "Mic 3", "MIC2"},
	
	/* ACOP */
	{"ACOP", NULL, "ALC Mixer"},
	
	/* terminator */
	{NULL, NULL, NULL},
};

static int wm8753_add_widgets(struct snd_soc_codec *codec)
{
	int i;
	
	for(i = 0; i < ARRAY_SIZE(wm8753_dpm_widgets); i++) {
		snd_soc_dpm_new_control(codec, &wm8753_dpm_widgets[i]);
	}
	
	/* set up audio path interconnects */
	for(i = 0; intercon[i][0] != NULL; i++) {
		snd_soc_dpm_connect_input(codec, intercon[i][0], intercon[i][1], intercon[i][2]);
	}
	
	snd_soc_dpm_sync(codec);
	return 0;
}

/*
 * set the PLL2 to generate 12.288MHz and then div by 6 to generate
 * DAC/ADC clocks from 12MHz MCLK 
 */
static u32 wm8753_config_pll(struct snd_soc_codec *codec, int pll, u32 speed)
{
	u16 clk;

    if (pll == 1) {
		clk = wm8753_read_reg_cache(codec, WM8753_CLOCK) & 0x0010;
        if (!speed) {
            /* disable PLL1  */
            wm8753_write(codec, WM8753_PLL1CTL1, 0x0026);
			wm8753_write(codec, WM8753_CLOCK, clk);
        } else {
            u16 value = 0;
            int i = 0;   
       
            /* if we cant match, then use good values for N and K */
            for (;i < ARRAY_SIZE(pll_div); i++) {
                if (pll_div[i].hz == speed)
                    break;
            }
			
			/* do we need to enable the PLL ? */
			if (!pll_div[i].n) {
				wm8753_write(codec, WM8753_CLOCK, 0x0);
				return 0;
			}
    
            /* set up N and K PLL divisor ratios */
            /* bits 8:5 = PLL_N, bits 3:0 = PLL_K[21:18] */
            value = (pll_div[i].n << 5) + ((pll_div[i].k & 0x3c0000) >> 18);  
            wm8753_write(codec, WM8753_PLL1CTL2, value);
     
            /* bits 8:0 = PLL_K[17:9] */
            value = (pll_div[i].k & 0x03fe00) >> 9;
            wm8753_write(codec, WM8753_PLL1CTL3, value);
    
            /* bits 8:0 = PLL_K[8:0] */
            value = pll_div[i].k & 0x0001ff;
            wm8753_write(codec, WM8753_PLL1CTL4, value);
     
            /* set PLL1 as input and enable */
            wm8753_write(codec, WM8753_PLL1CTL1, 0x0027);
			wm8753_write(codec, WM8753_CLOCK, clk | 0x0010);
        }
    } else {
		clk = wm8753_read_reg_cache(codec, WM8753_CLOCK) & 0x0008;
        if (!speed) {
            /* disable PLL2  */
            wm8753_write(codec, WM8753_PLL2CTL1, 0x0026);
			wm8753_write(codec, WM8753_CLOCK, clk);
        } else {
            u16 value = 0;
            int i = 0;   
       
            /* if we cant match, then use good values for N and K */
            for (;i < ARRAY_SIZE(pll_div); i++) {
                if (pll_div[i].hz == speed)
                    break;
            }
			
    		/* do we need to enable the PLL ? */
			if (!pll_div[i].n) {
				wm8753_write(codec, WM8753_CLOCK, 0x0);
				return 0;
			}
			
            /* set up N and K PLL divisor ratios */
            /* bits 8:5 = PLL_N, bits 3:0 = PLL_K[21:18] */
            value = (pll_div[i].n << 5) + ((pll_div[i].k & 0x3c0000) >> 18);  
            wm8753_write(codec, WM8753_PLL2CTL2, value);
     
            /* bits 8:0 = PLL_K[17:9] */
            value = (pll_div[i].k & 0x03fe00) >> 9;
            wm8753_write(codec, WM8753_PLL2CTL3, value);
    
            /* bits 8:0 = PLL_K[8:0] */
            value = pll_div[i].k & 0x0001ff;
            wm8753_write(codec, WM8753_PLL2CTL4, value);
     
            /* set PLL1 as input and enable */
            wm8753_write(codec, WM8753_PLL2CTL1, 0x0027);
			wm8753_write(codec, WM8753_CLOCK, clk | 0x0008);
        }
    }
	return speed;
}

static int wm8753_pcm_voice_prepare(snd_pcm_substream_t *substream)
{
	struct snd_soc_pcm_codec *pcm_c = substream->private_data;
	struct snd_soc_codec *codec = pcm_c->codec;
	u16 voice = 0, ioctl, clock, srate;
	
	/* enable PLL */
	wm8753_config_pll(codec, 2, pcm_c->hw_runtime.priv1);
	
	/* set master/slave audio interface */
	ioctl = wm8753_read_reg_cache(codec, WM8753_IOCTL) & 0x01f1;
	switch(pcm_c->hw_runtime.hformat & SND_SOC_CLOCK_MASK) {
		case SND_SOC_HWFMT_CBM_CFM:
			voice |= 0x0040;
			ioctl |= 0x0002;
			break;
		case SND_SOC_HWFMT_CBM_CFS:
			voice |= 0x0040;
			break;
	}
	
	/* interface format */
	switch(pcm_c->hw_runtime.hformat & SND_SOC_FORMAT_MASK) {
		case SND_SOC_HWFMT_I2S:
			voice |= 0x0002;
			break;
		case SND_SOC_HWFMT_RIGHT_J:
			break;
		case SND_SOC_HWFMT_LEFT_J:
			voice |= 0x0001;
			break;
		case SND_SOC_HWFMT_DSP_A:
			voice |= 0x0003;
			break;
		case SND_SOC_HWFMT_DSP_B:
			voice |= 0x0013;
			break;
	}
	
	/* bit size */
	switch(pcm_c->hw_runtime.hbits) {
		case SND_SOC_HWBITS(16):
			break;
		case SND_SOC_HWBITS(20):
			voice |= 0x0004;
			break;
		case SND_SOC_HWBITS(24):
			voice |= 0x0008;
			break;
		case SND_SOC_HWBITS(32):
			voice |= 0x000c;
			break;
	}
	
	/* set rate */
	clock = wm8753_read_reg_cache(codec, WM8753_CLOCK) & 0x003f;
	clock |= (pcm_c->hw_runtime.priv2 & 0xff00) >> 2;
	wm8753_write(codec, WM8753_CLOCK, clock);

	srate = wm8753_read_reg_cache(codec, WM8753_SRATE1) & 0x017f;
	srate |= (pcm_c->hw_runtime. priv2 & 0xff) << 7;
	wm8753_write(codec, WM8753_SRATE1, srate);
	
	/* clock inversion */
	switch(pcm_c->hw_runtime.hformat & SND_SOC_INV_MASK) {
		case SND_SOC_HWFMT_IB_IF:
			voice |= 0x0090;
			break;
		case SND_SOC_HWFMT_IB_NF:
			voice |= 0x0080;
			break;
		case SND_SOC_HWFMT_NB_IF:
			voice |= 0x0010;
			break;
	}
	wm8753_write(codec, WM8753_IOCTL, ioctl);
	wm8753_write(codec, WM8753_PCM, voice);
	return 0;
}

static int wm8753_pcm_hifi_prepare(snd_pcm_substream_t *substream)
{
	struct snd_soc_pcm_codec *pcm_c = substream->private_data;
	struct snd_soc_codec *codec = pcm_c->codec;
	u16 hifi = 0, ioctl, srate;
	
	/* enable PLL */
	wm8753_config_pll(codec, 1, pcm_c->hw_runtime.priv1);
	
	/* set master/slave audio interface */
	ioctl = wm8753_read_reg_cache(codec, WM8753_IOCTL) & 0x00f2;
	switch(pcm_c->hw_runtime.hformat & SND_SOC_CLOCK_MASK) {
		case SND_SOC_HWFMT_CBM_CFM:
			hifi |= 0x0040;
			ioctl |= 0x0001;
			break;
		case SND_SOC_HWFMT_CBM_CFS:
			hifi |= 0x0040;
			break;
	}
	
	/* interface format */
	switch(pcm_c->hw_runtime.hformat & SND_SOC_FORMAT_MASK) {
		case SND_SOC_HWFMT_I2S:
			hifi |= 0x0002;
			break;
		case SND_SOC_HWFMT_RIGHT_J:
			break;
		case SND_SOC_HWFMT_LEFT_J:
			hifi |= 0x0001;
			break;
		case SND_SOC_HWFMT_DSP_A:
			hifi |= 0x0003;
			break;
		case SND_SOC_HWFMT_DSP_B:
			hifi |= 0x0013;
			break;
	}
	
	/* bit size */
	switch(pcm_c->hw_runtime.hbits) {
		case SND_SOC_HWBITS(16):
			break;
		case SND_SOC_HWBITS(20):
			hifi |= 0x0004;
			break;
		case SND_SOC_HWBITS(24):
			hifi |= 0x0008;
			break;
		case SND_SOC_HWBITS(32):
			hifi |= 0x000c;
			break;
	}
	
	/* set rate */
	srate = wm8753_read_reg_cache(codec, WM8753_SRATE2) & 0x01c7;
	switch ((pcm_c->hw_runtime.priv2 & 0xff00) >> 8){
		case 1:
			break;
		case 2:
			srate |= (0x1 << 3);
			break;
		case 4:
			srate |= (0x2 << 3);
			break;
		case 8:
			srate |= (0x3 << 3);
			break;
		case 16:
			srate |= (0x4 << 3);
			break;
	}		
	wm8753_write(codec, WM8753_SRATE2, srate);
	srate = (pcm_c->hw_runtime. priv2 & 0xff) << 1;
	wm8753_write(codec, WM8753_SRATE1, srate);
	
	/* clock inversion */
	switch(pcm_c->hw_runtime.hformat & SND_SOC_INV_MASK) {
		case SND_SOC_HWFMT_IB_IF:
			hifi |= 0x0090;
			break;
		case SND_SOC_HWFMT_IB_NF:
			hifi |= 0x0080;
			break;
		case SND_SOC_HWFMT_NB_IF:
			hifi |= 0x0010;
			break;
	}
	wm8753_write(codec, WM8753_IOCTL, ioctl | 0x0008);
	wm8753_write(codec, WM8753_HIFI, hifi);
	return 0;
}

static void wm8753_device_release(struct device * dev)
{
}

static int wm8753_dpm_event(struct snd_soc_codec *codec, int event)
{
	u16 pwr_reg = wm8753_read_reg_cache(codec, WM8753_PWR1) & 0xfe3f;

	switch (event) {
		case SNDRV_CTL_POWER_D0: /* full On */
			/* set vmid to 50k and unmute dac */
			wm8753_write(codec, WM8753_PWR1, pwr_reg | 0x00c0);
			break;
		case SNDRV_CTL_POWER_D1: /* partial On */
		case SNDRV_CTL_POWER_D2: /* partial On */
			/* set vmid to 5k for quick power up */
			wm8753_write(codec, WM8753_PWR1, pwr_reg | 0x01c0);
			break;
		case SNDRV_CTL_POWER_D3hot: /* Off, with power */
			/* mute dac and set vmid to 500k, enable VREF */
			wm8753_write(codec, WM8753_PWR1, pwr_reg | 0x0140);
			break;
		case SNDRV_CTL_POWER_D3cold: /* Off, without power */
			wm8753_write(codec, WM8753_PWR1, 0x0001);
			break;
	}
	
	return 0;
}

static struct snd_soc_pcm_codec wm8753_pcm_clients[] = {
{	.name = "WM8753 Voice",
	.playback = {
		.sname = "Voice Playback",
		.rates = SNDRV_PCM_RATE_8000 | SNDRV_PCM_RATE_16000 |
		SNDRV_PCM_RATE_32000 | SNDRV_PCM_RATE_44100 | SNDRV_PCM_RATE_48000,
		.formats = SNDRV_PCM_FMTBIT_S16_LE,
		.channels_min = 1,
		.channels_max = 1,},
	.capture = {
		.sname = "Voice Capture",
		.rates = SNDRV_PCM_RATE_8000 | SNDRV_PCM_RATE_16000 | 
		SNDRV_PCM_RATE_32000 | SNDRV_PCM_RATE_48000 | SNDRV_PCM_RATE_44100 |
		SNDRV_PCM_RATE_96000,
		.formats = SNDRV_PCM_FMTBIT_S16_LE,
		.channels_min = 1,
		.channels_max = 2,},
	.nplayback = 1,
	.ncapture = 1,
	.ops = {
		.prepare = wm8753_pcm_voice_prepare,},
	.hw = {
		.num_hmodes = ARRAY_SIZE(wm8753_voice),
		.hmodes = &wm8753_voice[0],
		},},
{	.name = "WM8753 HiFi",
	.playback = {
		.sname = "HiFi Playback",
		.rates = SNDRV_PCM_RATE_8000 | SNDRV_PCM_RATE_16000 |
		SNDRV_PCM_RATE_32000 | SNDRV_PCM_RATE_48000 | SNDRV_PCM_RATE_44100 |
		SNDRV_PCM_RATE_96000,
		.formats = SNDRV_PCM_FMTBIT_S16_LE,
		.channels_min = 1,
		.channels_max = 2,},
	.nplayback = 1,
	.ncapture = 1,
	.ops = {
		.prepare = wm8753_pcm_hifi_prepare,},
	.hw = {
		.num_hmodes = ARRAY_SIZE(wm8753_hifi),
		.hmodes = &wm8753_hifi[0],},},
};

/*
 * initialise the WM8753 driver
 * register the mixer and dsp interfaces with the kernel 
 */
static int wm8753_probe(struct snd_soc_codec *codec)
{
	int reg;

	wm8753_reset(codec);
	
	/* charge output caps */
	wm8753_dpm_event(codec, SNDRV_CTL_POWER_D2);
	
	/* set the update bits */
	reg = wm8753_read_reg_cache(codec, WM8753_LDAC);
	wm8753_write(codec, WM8753_LDAC, reg | 0x0100);
	reg = wm8753_read_reg_cache(codec, WM8753_RDAC);
	wm8753_write(codec, WM8753_RDAC, reg | 0x0100);
	reg = wm8753_read_reg_cache(codec, WM8753_LOUT1V);
	wm8753_write(codec, WM8753_LOUT1V, reg | 0x0100);
	reg = wm8753_read_reg_cache(codec, WM8753_ROUT1V);
	wm8753_write(codec, WM8753_ROUT1V, reg | 0x0100);
	reg = wm8753_read_reg_cache(codec, WM8753_LOUT2V);
	wm8753_write(codec, WM8753_LOUT2V, reg | 0x0100);
	reg = wm8753_read_reg_cache(codec, WM8753_ROUT2V);
	wm8753_write(codec, WM8753_ROUT2V, reg | 0x0100);
	reg = wm8753_read_reg_cache(codec, WM8753_LINVOL);
	wm8753_write(codec, WM8753_LINVOL, reg | 0x0100);
	reg = wm8753_read_reg_cache(codec, WM8753_RINVOL);
	wm8753_write(codec, WM8753_RINVOL, reg | 0x0100);

	wm8753_add_controls(codec);
	wm8753_add_widgets(codec);
	
	/* wait for caps to finish charging - liam make this thread */
	ssleep(1);
	wm8753_dpm_event(codec, SNDRV_CTL_POWER_D3hot);

	return 0;
}

/* 
 * unregister interfaces and clean up
 */
static void wm8753_remove(struct snd_soc_codec *codec)
{
	/* power down chip */
	wm8753_dpm_event(codec, SNDRV_CTL_POWER_D3cold);
}

/*
 * Attach WM8753 2 wire client 
 */
static int wm8753_i2c_probe(struct i2c_adapter *adap, int addr, int kind)
{		
	int ret = 0;
	struct snd_soc_codec *codec;
	struct i2c_client *i2c;

	client_template.adapter = adap;
	client_template.addr = addr;
	
	if ((codec = kzalloc(sizeof(struct snd_soc_codec), GFP_KERNEL)) == NULL)
		return -ENOMEM;
 
	if ((i2c = kzalloc(sizeof(struct i2c_client), GFP_KERNEL)) == NULL){
		kfree(codec);
		return -ENOMEM;
	}
	memcpy(i2c, &client_template, sizeof(struct i2c_client));
	codec->control_data = i2c;
	i2c->dev.platform_data = codec;

	if((ret = i2c_attach_client(i2c)) < 0) {
		err("failed to attach codec at addr %x\n", addr);
		kfree(i2c);
		kfree(codec);
		return ret;
	}

	codec->name = "WM8753";
	codec->longname = "WM8753 Portable Codec";
	codec->owner = THIS_MODULE;
	codec->pdev.name = "alsa-wm8753";
	codec->pdev.id = -1;
	codec->pdev.dev.release = wm8753_device_release;
	codec->probe = wm8753_probe;
	codec->remove = wm8753_remove;
	codec->read = wm8753_read_reg_cache;
	codec->write = wm8753_write;
	codec->dpm_event = wm8753_dpm_event;
	codec->pcms = wm8753_pcm_clients;
	codec->npcms = 2;
	if ((codec->reg_cache = 
			kzalloc(sizeof(u16) * ARRAY_SIZE(wm8753_reg), GFP_KERNEL)) == NULL)
		return -ENOMEM;
	memcpy(codec->reg_cache, wm8753_reg, sizeof(u16) * ARRAY_SIZE(wm8753_reg));

	if((ret = snd_soc_register_codec(codec)) < 0) {
		err("can't register codec");
		i2c_del_driver(&wm8753_i2c_driver);
		kfree(i2c);
		kfree(codec->reg_cache);
		kfree(codec);
	}

	return ret;
}

static int wm8753_i2c_detach(struct i2c_client *client)
{
	struct snd_soc_codec *codec = 
		(struct snd_soc_codec*)client->dev.platform_data;

	i2c_detach_client(client);
	snd_soc_unregister_codec(codec);
	kfree(codec->reg_cache);
	kfree(codec);
	return 0;
}

static int wm8753_i2c_attach(struct i2c_adapter *adap)
{
	return i2c_probe(adap, &addr_data, wm8753_i2c_probe);
}

/* WM8753 2 Wire layer */ 
static struct i2c_driver wm8753_i2c_driver = {
	.name =           "i2c wm8753 driver",
	.id =             I2C_DRIVERID_WM8753,
	.flags =          I2C_DF_NOTIFY,
	.attach_adapter = wm8753_i2c_attach,
	.detach_client =  wm8753_i2c_detach,
	.command =        NULL,
};

static struct i2c_client client_template = {
	.name =   "WM8753",
	.flags =  I2C_CLIENT_ALLOW_USE,
	.driver = &wm8753_i2c_driver,
};

static int __init wm8753_init(void)
{	
	int ret = 0;
	
	info("WM8753 Audio Codec %s", WM8753_VERSION);

	if ((ret = i2c_add_driver(&wm8753_i2c_driver)) != 0)
		err("can't add i2c driver");
	
	return ret;
}

static void __exit wm8753_exit(void)
{
	i2c_del_driver(&wm8753_i2c_driver);
}

module_init(wm8753_init);
module_exit(wm8753_exit);

MODULE_DESCRIPTION("Soc WM8753 driver");
MODULE_AUTHOR("Liam Girdwood");
MODULE_LICENSE("GPL");
