/******************************************************************
 *
 *  File:   fiona-audio-pm.c
 *
 *  Author: Nick Vaccaro <nvaccaro@lab126.com>
 *
 *  Date:   08/14/06
 *
 *  Copyright 2006, Lab126, Inc.  All rights reserved.
 *
 *  Description:
 *
 *      Audio driver support for the Fiona Power Management System
 *
 ******************************************************************/
#include <linux/autoconf.h>
#include <linux/kernel.h>
#include <asm/arch/system.h>
#include <asm/arch/fiona.h>
#include <asm/arch/fiona_eventmap.h>
#include <asm/arch/fpow.h>	// FPOW-related constants

//extern fiona_hw g_fiona_hw;

// Static global holding current FPOW power state of the driver
FPOW_POWER_MODE	g_audio_power_mode = FPOW_MODE_INVALID; 

// Static global holding fpow device entry returned from fpow_register()
struct fpow_component *g_audio_fpow_component_ptr = NULL;

extern int snd_soc_pcm_fpow_set_mode(void *codec, int fpow_mode);


/*********************************************************
 * Routine:     audio_fpow_setmode
 *
 * Purpose:     Handles Fiona Power Management callbacks
 *              for doing Audio power management.
 *
 * Paramters:   private - value passed at registration - ignored
 *              state   - ignored
 *              mode    - Power mode to switch to
 *                          Currently supported Modes:
 *                              FPOW_MODE_ON
 *                              FPOW_MODE_SLEEP
 *                              FPOW_MODE_OFF
 *
 * Returns:     
 *  FPOW_NO_ERR                     - Transitioned to mode with no errors.
 *  -FPOW_NOT_ENOUGH_POWER_ERR      - not enough power
 *  -FPOW_COMPONENT_TRANSITION_ERR  - Failed to transition to mode
 *  -EINVAL                         - Invalid mode
 *
 * Assumptions: none
 *********************************************************/

static FPOW_ERR 
audio_fpow_setmode(void *private, u32 state, u32 mode)
{
	int	retVal = FPOW_NO_ERR;

    powdebug("audio_fpow_setmode(%s)\n",fpow_mode_to_str(mode));

	// If we're already in the proper mode, no need to do anything else
	if (g_audio_power_mode == mode)
		goto exit;

	// Switch to the appropriate power mode
	switch (mode)  {
		case FPOW_MODE_ON:
		case FPOW_MODE_SLEEP:
		case FPOW_MODE_OFF:
			retVal = snd_soc_pcm_fpow_set_mode(private, mode);
			if (retVal == 0) {
				g_audio_power_mode = mode;

				retVal = FPOW_NO_ERR;
			}
			break;

		default:
			retVal = -EINVAL;
			break;
	}

exit:
	powdebug("AUDIO: driver now in %s mode\n",fpow_mode_to_str(g_audio_power_mode));
    return(retVal);
}

static int
audio_fpow_getmode(void *private) 
{
    powdebug("audio_fpow_getmode(0x%x) returning %s\n",(int)private,fpow_mode_to_str(g_audio_power_mode));
	return (g_audio_power_mode);
}

int 
audio_register_with_fpow(struct fpow_component **fpow_dev, void *private)
{
    struct fpow_registration_rec   reg;
    int rc = 0;

	powdebug("audio_register_with_fpow(0x%x, 0x%x)\n",(int)fpow_dev, (int)private);
    memclr(&reg,sizeof(reg));

    strcpy(reg.name,"audio");
    reg.device_class = FPOW_CLASS_AUDIO;
    reg.supported_modes = ( FPOW_MODE_ON_SUPPORTED | \
                            FPOW_MODE_SLEEP_SUPPORTED | \
                            FPOW_MODE_OFF_SUPPORTED);
    reg.private = private;
    reg.getmode = audio_fpow_getmode;
    reg.setmode = audio_fpow_setmode;
    *fpow_dev = fpow_register(&reg);
    if (*fpow_dev == NULL) {
        printk("WAN: Failed to register with fpow !!\n");
        rc = -1;
    }
    return(rc);
}

void
audio_unregister_with_fpow(struct fpow_component **fpow_dev)
{
	fpow_unregister(*fpow_dev);
    *fpow_dev = NULL;
}
